/****************************************************************************
   Copyright & License:
   ====================
   
   Copyright 2009 - 2016 gAGE/UPC & ESA
   
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 ***************************************************************************/

/****************************************************************************
 * Copyright: gAGE/UPC & ESA
 * Project: EDUNAV GNSS Lab Tool 
 * Author: Pere Ramos-Bosch (group of Astronomy and GEomatics - gAGE/UPC)
 * Curator: Adria Rovira-Garcia ( gAGE/UPC )
 * Developers: Deimos Ibanez Segura ( gAGE/UPC )
 *             Jesus Romero Sanchez ( gAGE/UPC )
 *          glab.gage @ upc.edu
 * File: filter.c
 * Code Management Tool File Version: 3.1  Revision: 1
 * Date: 2016/10/21
 ***************************************************************************/

/****************************************************************************
 * MODULE DESCRIPTION
 *
 * Name: filter
 * Language: C
 *
 * Purpose:
 *  The purpose of this module is to contain all the necessary functions to
 *  do a Kalman filtering process.
 *
 * Function:
 *  This class is in charge of declaring providing all the necessary functions
 *  to do a Kalman filtering process. It has a high level function (Kalman) in
 *  which, given all necessary inputs, provides the estimated values for the
 *  parameters 
 *
 * Dependencies:
 *  dataHandling, model
 *
 * Files modified:
 *  None
 *
 * Files read:
 *  None
 *
 * Resources usage:
 *  See Design Document
 ****************************************************************************/

/****************************
 *       RELEASE_HISTORY
 * -------------
 *          gLAB v1.3.1
 * Release: 2010/01/22
 * Change Log: First public version
 * -------------
 *          gLAB v1.3.2
 * Release: 2010/03/23
 * Change Log: PREFIT message now prints the weight to be applied in the filter
 *             FILTER message now prints the total ZTD and the estimated ZTD (substracting the modeled value)
 *             PREFIT message now prints non-GPS satellites
 * -------------
 *          gLAB v1.4.0
 * Release: 2010/06/21
 * Change Log: Included usable measurement index to select/deselect individual measurements
 * -------------
 *          gLAB v1.4.1
 * Release: 2010/07/09
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.2
 * Release: 2010/07/31
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.3
 * Release: 2010/08/31
 * Change Log: Improved DCB Management.
 * -------------
 *          gLAB v1.4.4
 * Release: 2010/09/22
 * Change Log: Filter adjustments to new DCB management.
 * -------------
 *          gLAB v1.4.5
 * Release: 2010/10/31
 * Change Log: Dilution Of Precision extension: (GDOP, PDOP, TDOP, VDOP, HDOP).
 * -----------
 *          gLAB v2.0.0
 * Release: 2012/12/31
 * Change Log: gLAB version released together with the Educational Book:
 *             "GNSS Data Processing" (Vol. 1 and Vol. 2). ESA TM-23.
 *             Authors: J. Sanz Subirana, J.M. Juan Zornoza and M. Hernandez-Pajares
 *             ISBN: 978-92-9221-885-0 (two volumes)
 *             ISSN: 1013-7076
 *             December 2012.
 * -----------
 *          gLAB v2.2.0
 * Release: 2014/09/22
 * Change Log: Modified fillUsableSatellites function in order to discard satellites without IONEX or F-PPP model data
 *			(in case those models are used).
 * -----------
 *          gLAB v2.2.1
 * Release: 2014/12/23
 * Change Log: Fixed initialization to given value per parameter of covariance of coordinates Y and Z (before only coordinate X was initialized, Y and Z were 0)
 * -----------
 *          gLAB v2.2.2
 * Release: 2015/03/02
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.3
 * Release: 2015/05/28
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.4
 * Release: 2015/07/01
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.5
 * Release: 2015/10/20
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.6
 * Release: 2015/11/25
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.7
 * Release: 2016/03/01
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.8
 * Release: 2016/04/12
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v3.0.0
 * Release: 2016/09/05 
 * Change Log: Added step detector function.
 *	       Added Stanford-ESA computation.
 *	       Modified structure definitions for handling Cycle Slip variables.
 *	       Modified printSatSel calls.
 * -----------
 *          gLAB v3.1.0
 * Release: 2016/09/26
 * Change Log: Modified Kalman function to allow model recomputation in NPA mode
 *               if after calling fillUsableSatellites and we were in PA mode, there 
 *               were not enough satellites for computing navigation solution (only in SBAS).
 *             Eliminated "preauxEpoch" struct from Kalman function, improving performance
 *               (specially when computing trajectories).
 * -----------
 *          gLAB v3.1.1
 * Release: 2016/10/21
 * Change Log: 'Elevation too low' SATSEL message is not shown if ephemerides are not available.
 *             Added quantization check in Stanford-ESA plots generation for points greater than the
 *                maximum X and Y. For example, (60,100) and (100,60) with a max X and Y of 50, 
 *                would both result in the plot in the point (50,50) or (Xmax,Ymax), which made
 *                it impossible to distinguish the MI from the non-MI. Now, MI points will result
 *                in (Xmax,Ymax) while non MI will result in (Xmax-1,Ymax)
 * -----------
 *       END_RELEASE_HISTORY
 *****************************/

/* External classes */
#include "filter.h"
extern int	linesstored[MAX_SATELLITES_IN_VIEW];
extern char	printbuffer[MAX_SATELLITES_IN_VIEW][MAX_LINES_BUFFERED][MAX_INPUT_LINE]; 

/**************************************
 * Declarations of internal operations
 **************************************/

/*****************************************************************************
 * Name        : m2v
 * Description : Transforms the indices from matrix to vector
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  i                          I  N/A  Matrix row
 * int  j                          I  N/A  Matrix column
 * Returned value (int)            O  N/A  Vector index
 *****************************************************************************/
int m2v (int i, int j) {
  return i>j ? i*(i+1)/2+j : j*(j+1)/2+i;
}

/*****************************************************************************
 * Name        : m2v_opt
 * Description : Optimized version of m2v (much faster). It suposes i>j.
 *               It is specially useful when inverting matrixes.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  i                          I  N/A  Matrix row
 * int  j                          I  N/A  Matrix column
 * Returned value (int)            O  N/A  Vector index
 *****************************************************************************/
static inline int m2v_opt (int i, int j) {
  return i*(i+1)/2+j;
}

/*****************************************************************************
 * Name        : cholinv
 * Description : Cholesky matrix inversion
 *               Version inspired from "Numerical Recipes in C" P.97-98
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *mat                    IO N/A  Matrix in vector form (overwritten)
 * int  n                          I  N/A  Matrix size
 * Returned value (int)            O  N/A  Status of the function
 *                                         0  => Cholesky worked properly
 *                                         -1 => Cholesky inversion failed (singular
 *                                               matrix)
 *****************************************************************************/
int cholinv (double *mat, int n) {
  int i,j,k;
  double sum;

  for (i=0;i<n;i++) {
	for (j=i;j<n;j++) {
	  for (sum=mat[m2v(i,j)],k=i-1;k>=0;k--) {sum -= mat[m2v(i,k)]*mat[m2v(j,k)]; } 
	  if (i==j) {
		if (sum <= 0) return -1;  // Cholesky decomposition failed
		mat[m2v(i,i)]=sqrt(sum);
	  } else mat[m2v(j,i)]=sum/mat[m2v(i,i)];
	}
  }

// Beginning Upper Inversion

  for (i=0;i<n;i++) {
	mat[m2v(i,i)]=1/mat[m2v(i,i)];
	for (j=i+1;j<n;j++) {
	  sum=0;
	  for (k=i;k<j;k++) sum -= mat[m2v(j,k)]*mat[m2v(k,i)];
	  mat[m2v(j,i)]=sum/mat[m2v(j,j)];
	}
  }

// Multiplying U^-1 * UT^-1

  for (i=0;i<n;i++) {
	for (j=i;j<n;j++) {
	  sum=0;
	  for (k=j;k<n;k++) sum += mat[m2v(i,k)]*mat[m2v(k,j)];
	  mat[m2v(i,j)]=sum;
	}
  }

  return 0;
}

/*****************************************************************************
 * Name        : cholinv_opt
 * Description : Optimized version of cholinv. It uses m2v_opt
 *               Version inspired from "Numerical Recipes in C" P.97-98
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *mat                    IO N/A  Matrix in vector form (overwritten)
 * int  n                          I  N/A  Matrix size
 * Returned value (int)            O  N/A  Status of the function
 *                                         0  => Cholesky worked properly
 *                                         -1 => Cholesky inversion failed (singular
 *                                               matrix)
 *****************************************************************************/
int cholinv_opt (double *mat, int n) {
  int i,j,k;
  double sum;
  double *aux;
 
  // Reserving space 
  aux = malloc((m2v_opt(n-1,n-1)+1)*sizeof(double));
  memset(aux,0,(m2v_opt(n-1,n-1)+1)*sizeof(double));

  for (i=0;i<n;i++) {
	for (j=i;j<n;j++) {
	  for (sum=mat[m2v_opt(j,i)],k=i-1;k>=0;k--) sum -= mat[m2v_opt(i,k)]*mat[m2v_opt(j,k)];
	  if (i==j) {
		if (sum <= 0) {
			free(aux);
			return -1;  // Cholesky decomposition failed
		}
		mat[m2v_opt(i,i)]=sqrt(sum);
	  } else mat[m2v_opt(j,i)]=sum/mat[m2v_opt(i,i)];
	}
  }

// Upper Inversion and multiplication of U^-1 * UT^-1

  for (i=0;i<n;i++) {
	mat[m2v_opt(i,i)]=1/mat[m2v_opt(i,i)];
	for (k=0;k<=i;k++) aux[m2v_opt(i,k)]+=mat[m2v_opt(i,i)]*mat[m2v_opt(i,k)];
	for (j=i+1;j<n;j++) {
	  sum=0;
	  for (k=i;k<j;k++) sum -= mat[m2v_opt(j,k)]*mat[m2v_opt(k,i)];
	  mat[m2v_opt(j,i)]=sum/mat[m2v_opt(j,j)];
	  for (k=0;k<=i;k++) aux[m2v_opt(i,k)]+=mat[m2v_opt(j,i)]*mat[m2v_opt(j,k)];
	}
  }

  memcpy(mat,aux,(m2v_opt(n-1,n-1)+1)*sizeof(double));

  free (aux);
  return 0;
}

/*****************************************************************************
 * Name        : mxv
 * Description : Multiplies a matrix in vectorial form and a vector
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *mat                    I  N/A  Matrix in vector form
 * double  *vec                    IO N/A  Vector (overwritten)
 * int  n                          I  N/A  Matrix and vector size
 *****************************************************************************/
void mxv (double *mat, double *vec, int n) {
	int		i,j;
	double	*aux;

	aux=(double *)(malloc(sizeof(double)*n));

	for (i=0;i<n;i++) {
		aux[i]=0;
		for (j=0;j<n;j++) {
			aux[i] += mat[m2v(i,j)] * vec[j];
		}
	}
	for (i=0;i<n;i++) vec[i]=aux[i];
	free(aux);
}

/*****************************************************************************
 * Name        : StepDetector
 * Description : Check for jumps greater than 700 meters. The step detector
 *               is useful for detecting jumps in navigation message. 
 *               The threshold for comparing is 73.8 meters (see MOPS-C, 
 *                Appendix I, page I-3). To be used only with C1C code.
 *               The step detector computes the difference of prefits between 
 *               the current and previous epoch (only with satellites that have been 
 *               used in the navigation solution in both epochs) with reference satellites
 *               (the two ones with highest elevation in current epoch). Then computes
 *               the difference between prefit difference with both satellite. If
 *               both differences are over the threshold, then there has been a step.
 *               With this method a big step will not affect other satellites, but it
 *               the drawback that it assumes that only one satellite will fail or that
 *               more than one satellite may fail but not the reference ones.
 *               The problem with any step detector is that it will only trigger when there
 *               is a jump when changing the ephemerides to ones which have incorrect values, 
 *               but in the next epochs where previous and current epochs use the same bad
 *               ephemerides, the step detector will not trigger.
 *               
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  Reference epoch
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void StepDetector (TEpoch *epoch, TOptions *options) {

	int				i,j,k;
	int				ind,ret;
	int				numsats=0;
	int				C1Cused=0;
	int				maxelevationpos=-1,secondmaxelevationpos=-1,thirdmaxelevationpos=-1;
	int				SatUsed[MAX_GNSS][MAX_SBAS_PRN];
	static int		PrevSatUsed[MAX_GNSS][MAX_SBAS_PRN];

	double			Threshold=73.8;
	double			measurement,smoothWith,model,prefit;
	double			diff1,diff2,Rescurrfirst,Resprevfirst,Rescurrsecond,Resprevsecond;
	double			Prefits[MAX_GNSS][MAX_SBAS_PRN];
	static double	PrevPrefits[MAX_GNSS][MAX_SBAS_PRN];

	//Put this option to 1 so all prints are buffered
	options->printInBuffer=1;
	
	//Check that C1C measurement is used in the filter. Otherwise the step detector has no use
	for(i=0;i<options->totalFilterMeasurements;i++) {
		if(options->measurement[i]==C1C) {
			C1Cused=1;
			break;
		}
	}

	if(C1Cused==0) return;

	//Initialize SatUsed
	for(i=0;i<MAX_GNSS;i++) {
		for(j=0;j<MAX_SBAS_PRN;j++) {
			SatUsed[i][j]=0;
			Prefits[i][j]=0;
		}
	}

	ind = epoch->measOrder[GPS].meas2Ind[C1C];
	k = satIs(epoch,epoch->sat[i].PRN,epoch->sat[i].GNSS);

	//Retrieve all the prefits available
	for(i=0;i<epoch->numSatellites;i++) {
		if(epoch->sat[i].available==0) {
			SatUsed[epoch->sat[i].GNSS][epoch->sat[i].PRN]=0;
			continue;
		}
		ret = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,epoch->measOrder[epoch->sat[i].GNSS].ind2Meas[ind],&measurement,&model);
		if (ret!=1) {
			SatUsed[epoch->sat[i].GNSS][epoch->sat[i].PRN]=0;
			continue;
		}

		// Smoothing
		if (options->smoothEpochs!=0 && options->smoothMeas[0]!=NA) {
			ret = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,options->smoothMeas[0],&smoothWith,NULL);
			if (ret!=1) {
				return;
			}
			Prefits[epoch->sat[i].GNSS][epoch->sat[i].PRN] = epoch->cycleslip.smoothedMeas[k][0] + smoothWith - model;
		} else {
			Prefits[epoch->sat[i].GNSS][epoch->sat[i].PRN] = measurement - model;
		}

		if(PrevSatUsed[epoch->sat[i].GNSS][epoch->sat[i].PRN]==1) {
			numsats++;
		//Look for the three satellites with higher elevation
		if(maxelevationpos==-1) {
			maxelevationpos=i;
		} else if (secondmaxelevationpos==-1) {
			secondmaxelevationpos=i;
		} else if (thirdmaxelevationpos==-1) {
			thirdmaxelevationpos=i;
		} else {
			if(epoch->sat[i].elevation>epoch->sat[maxelevationpos].elevation) {
				maxelevationpos=i;
			} else if (epoch->sat[i].elevation>epoch->sat[secondmaxelevationpos].elevation) {
				secondmaxelevationpos=i;
			} else if (epoch->sat[i].elevation>epoch->sat[thirdmaxelevationpos].elevation) {
				thirdmaxelevationpos=i;
			}
		}
			}
			SatUsed[epoch->sat[i].GNSS][epoch->sat[i].PRN]=1;
	}

	if(numsats>=3) {
		for(i=0;i<epoch->numSatellites;i++) {
			if(SatUsed[epoch->sat[i].GNSS][epoch->sat[i].PRN]==0) continue;
			if(PrevSatUsed[epoch->sat[i].GNSS][epoch->sat[i].PRN]==0) continue;
			if(i==maxelevationpos) {
				//Current satellite is the one with highest elevation.
				//We have to compare with satellites with second and third highest elevation
				//First compute the prefit difference between satellites for current and previous epoch
				Rescurrfirst=Prefits[epoch->sat[i].GNSS][epoch->sat[i].PRN]-Prefits[epoch->sat[secondmaxelevationpos].GNSS][epoch->sat[secondmaxelevationpos].PRN];
				Resprevfirst=PrevPrefits[epoch->sat[i].GNSS][epoch->sat[i].PRN]-PrevPrefits[epoch->sat[secondmaxelevationpos].GNSS][epoch->sat[secondmaxelevationpos].PRN];
				Rescurrsecond=Prefits[epoch->sat[i].GNSS][epoch->sat[i].PRN]-Prefits[epoch->sat[thirdmaxelevationpos].GNSS][epoch->sat[thirdmaxelevationpos].PRN];
				Resprevsecond=PrevPrefits[epoch->sat[i].GNSS][epoch->sat[i].PRN]-PrevPrefits[epoch->sat[thirdmaxelevationpos].GNSS][epoch->sat[thirdmaxelevationpos].PRN];
			} else if (i==secondmaxelevationpos) {
				//Current satellite is the one with second highest elevation.
				//We have to compare with satellites with first and third highest elevation
				//First compute the prefit difference between satellites for current and previous epoch
				Rescurrfirst=Prefits[epoch->sat[i].GNSS][epoch->sat[i].PRN]-Prefits[epoch->sat[maxelevationpos].GNSS][epoch->sat[maxelevationpos].PRN];
				Resprevfirst=PrevPrefits[epoch->sat[i].GNSS][epoch->sat[i].PRN]-PrevPrefits[epoch->sat[maxelevationpos].GNSS][epoch->sat[maxelevationpos].PRN];
				Rescurrsecond=Prefits[epoch->sat[i].GNSS][epoch->sat[i].PRN]-Prefits[epoch->sat[thirdmaxelevationpos].GNSS][epoch->sat[thirdmaxelevationpos].PRN];
				Resprevsecond=PrevPrefits[epoch->sat[i].GNSS][epoch->sat[i].PRN]-PrevPrefits[epoch->sat[thirdmaxelevationpos].GNSS][epoch->sat[thirdmaxelevationpos].PRN];
			} else {
				//We have to compare with satellites with first and second highest elevation
				//First compute the prefit difference between satellites for current and previous epoch
				Rescurrfirst=Prefits[epoch->sat[i].GNSS][epoch->sat[i].PRN]-Prefits[epoch->sat[maxelevationpos].GNSS][epoch->sat[maxelevationpos].PRN];
				Resprevfirst=PrevPrefits[epoch->sat[i].GNSS][epoch->sat[i].PRN]-PrevPrefits[epoch->sat[maxelevationpos].GNSS][epoch->sat[maxelevationpos].PRN];
				Rescurrsecond=Prefits[epoch->sat[i].GNSS][epoch->sat[i].PRN]-Prefits[epoch->sat[secondmaxelevationpos].GNSS][epoch->sat[secondmaxelevationpos].PRN];
				Resprevsecond=PrevPrefits[epoch->sat[i].GNSS][epoch->sat[i].PRN]-PrevPrefits[epoch->sat[secondmaxelevationpos].GNSS][epoch->sat[secondmaxelevationpos].PRN];
			}

			//Compute the difference between prefit differences
			diff1=fabs(Rescurrfirst-Resprevfirst);
			diff2=fabs(Rescurrsecond-Resprevsecond);

			//Check that differences do not exceed the threshold
			if(diff1>Threshold && diff2>Threshold) {
				linesstored[i]=0;
				sprintf(messagestr,"STEP detected (jump of %7.3f)",max(diff1,diff2));
				printSatSel(epoch,0,messagestr,i,options);
				epoch->sat[i].available=0;
				epoch->usableSatellites--;
			}
		}

	} else {
		//Step detector not available
	}

	memcpy(PrevSatUsed,SatUsed,sizeof(SatUsed));
	memcpy(PrevPrefits,Prefits,sizeof(Prefits));


/*
 * ALTERNATIVE METHOD FOR STEP DETECTOR: 
 * Description : Check for jumps greater than 700 meters. The step detector
 *               is useful for detecting jumps in navigation message. 
 *               The threshold for comparing is 73.8 meters (see MOPS-C, 
 *                Appendix I, page I-3).
 *               The step detector compares the value of the current and previous epoch
 *               (only with satellites that have been used in the navigation solution in both epochs)
 *               of the prefit minus the receiver clock, which is estimated from the mean 
 *               of all prefits. This mean has to be computed using only satellites available
 *               in the current and previous epoch. The problem with this approach is that as the
 *               receiver clock is computed with the mean of the prefits, if a satellite has a very
 *               high jump, a jump will be detected in all satellites 

	int 		i,j,k;
	int		ind,ret;
	int		numsats=0;
	int		SatUsed[MAX_GNSS][MAX_SBAS_PRN];
	static int	PrevSatUsed[MAX_GNSS][MAX_SBAS_PRN];

	double		Threshold=73.8;
	double		dtCurrent=0,dtPrev=0;
	double		measurement,smoothWith,model,prefit;
	double		diff,Rescurr,Resprev;
	double		Prefits[MAX_GNSS][MAX_SBAS_PRN];
	static double	PrevPrefits[MAX_GNSS][MAX_SBAS_PRN];

	//Put this option to 1 so all prints are buffered
	options->printInBuffer=1;

	//Initialize SatUsed
	for(i=0;i<MAX_GNSS;i++) {
		for(j=0;j<MAX_SBAS_PRN;j++) {
			SatUsed[i][j]=0;
			Prefits[i][j]=0;
		}
	}

	ind = epoch->measOrder[GPS].meas2Ind[C1C];
	k = satIs(epoch,epoch->sat[i].PRN,epoch->sat[i].GNSS);

	//Retrieve all the prefits available
	for(i=0;i<epoch->numSatellites;i++) {
		if(epoch->sat[i].available==0) {
			SatUsed[epoch->sat[i].GNSS][epoch->sat[i].PRN]=0;
			continue;
		}
		ret = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,epoch->measOrder[epoch->sat[i].GNSS].ind2Meas[ind],&measurement,&model);
		if (ret!=1) {
			SatUsed[epoch->sat[i].GNSS][epoch->sat[i].PRN]=0;
			continue;
		}

		// Smoothing
		if (options->smoothEpochs!=0 && options->smoothMeas[0]!=NA) {
			ret = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,options->smoothMeas[0],&smoothWith,NULL);
			if (ret!=1) {
				return;
			}
			Prefits[epoch->sat[i].GNSS][epoch->sat[i].PRN] = epoch->cycleslip.smoothedMeas[k][0] + smoothWith - model;
		} else {
			Prefits[epoch->sat[i].GNSS][epoch->sat[i].PRN] = measurement - model;
		}

		if(PrevSatUsed[epoch->sat[i].GNSS][epoch->sat[i].PRN]==1) {
			dtCurrent+=Prefits[epoch->sat[i].GNSS][epoch->sat[i].PRN];
			dtPrev+=PrevPrefits[epoch->sat[i].GNSS][epoch->sat[i].PRN];
			numsats++;
		}
		SatUsed[epoch->sat[i].GNSS][epoch->sat[i].PRN]=1;
	}

	if(numsats>=3) {
		dtCurrent/=numsats;
		dtPrev/=numsats;
		//Compute for each satellite the prefit minus the receiver clock estimate and compare it with previous one (if available)
		for(i=0;i<epoch->numSatellites;i++) {
			if(SatUsed[epoch->sat[i].GNSS][epoch->sat[i].PRN]==0) continue;
			//Substract the receiver clock estimate from the prefits
			if(PrevSatUsed[epoch->sat[i].GNSS][epoch->sat[i].PRN]==1) {
				Rescurr=Prefits[epoch->sat[i].GNSS][epoch->sat[i].PRN]-dtCurrent;
				Resprev=PrevPrefits[epoch->sat[i].GNSS][epoch->sat[i].PRN]-dtPrev;
				diff=fabs(Rescurr-Resprev);
				if(diff>Threshold) {
					linesstored[i]=0;
					sprintf(messagestr,"STEP detected (jump of %7.3f)",diff);
					printSatSel(epoch,0,messagestr,i,options);
					epoch->sat[i].available=0;
					epoch->usableSatellites--;
				}
			}
		}
	} else {
		//Step detector not available
	}

	memcpy(PrevSatUsed,SatUsed,sizeof(SatUsed));
	memcpy(PrevPrefits,Prefits,sizeof(Prefits));
*/



}

/*****************************************************************************
 * Name        : fillUsableSatellites
 * Description : Check for the usable satellites in the specified epoch
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  Reference epoch
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void fillUsableSatellites (TEpoch *epoch, TOptions *options) {
	int		i,j,k;
	int		check;
	double	sunPos[3];
	double	auxT;
	int		onlyHasPCorPCC;
	enum MeasurementType	meas;

	//Put this option to 1 so all prints are buffered
	options->printInBuffer=1;
	
	epoch->usableSatellites = 0;

	if (options->discardEclipsedSatellites) {
		findSun(&epoch->t,sunPos);
	}

	for (i=0;i<epoch->numSatellites;i++) {
		check = 1;
		k = satIs(epoch,epoch->sat[i].PRN,epoch->sat[i].GNSS);

		// Arc length sanity check
		if ( epoch->cycleslip.arcLength[k] <= options->minArcLength ) {
			sprintf(messagestr,"Arc too short (%d)",epoch->cycleslip.arcLength[k]);
			printSatSel(epoch,0,messagestr,i,options);
			check = 0;
		}

		// N-consecutive sanity check
		if ( epoch->cycleslip.Nconsecutive[k] != 0 ) {
			sprintf(messagestr,"Non N-consecutive epochs");
			printSatSel(epoch,0,messagestr,i,options);
			// No check -> only flag
		}

		// Cycle slip sanity check
		if ( epoch->cycleslip.CS[k] == 1 ) {
			sprintf(messagestr,"Cycle-slip");
			printSatSel(epoch,0,messagestr,i,options);
			// No check -> only flag
		}

		// Satellite availability sanity check
		if ( epoch->sat[i].available == 0 && epoch->sat[i].preSBASavailable == 0 ) {
			check = 0;
		}

		// Satellite Pre-Check sanity check
		if ( epoch->cycleslip.preCheck[k] == 1 && !epoch->cycleslip.consistency[k] ) {
			sprintf(messagestr,"Pre-Check; measurements missing for CS detection");
			printSatSel(epoch,0,messagestr,i,options);
			// No check -> only flag
		}

		// Satellite Consistency of Measurements sanity check
		if ( epoch->cycleslip.consistency[k] == 1 ) {
			sprintf(messagestr,"Unconsistency of measurements");
			printSatSel(epoch,0,messagestr,i,options);
			// No check -> only flag
		}

		// Outlier sanity check
		if ( epoch->cycleslip.outlier[k] == 1 ) {
			sprintf(messagestr,"Outlier");
			printSatSel(epoch,0,messagestr,i,options);
			check = 0;
		}

		// Data availability check
		for (j=0;j<options->totalFilterMeasurements;j++) {
			// Check if there are valid measurements for navigating
			if (getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,options->measurement[j],NULL,NULL)!=1) {
				sprintf(messagestr,"Measurement unavailable (%s)",meastype2measstr(options->measurement[j]));
				printSatSel(epoch,0,messagestr,i,options);
				check = 0; 
			}
		}
		
		// Satellite selection check
		if ( !options->includeSatellite[epoch->sat[i].GNSS][epoch->sat[i].PRN] ) {
			printSatSel(epoch,0,"Deselection forced",i,options);
			check = 0;
		}

		// SNR check
		if ( options->SNRfilter==1 && epoch->sat[i].hasSNR==0 ) {
			sprintf(messagestr,"SNR too low (%d)",epoch->sat[i].lowSNR);
			printSatSel(epoch,0,messagestr,i,options);
			check = 0;
		}

		// Steady-state operation check
		if ( options->excludeSmoothingConvergence > 0 && (( epoch->cycleslip.arcLength[k] - options->minArcLength ) < options->excludeSmoothingConvergence) ) {
			// Satellite not reached steady-state operation
			sprintf(messagestr,"Steady-state operation not reached. %d seconds passed, still %d seconds to go",(int)max(1.0,epoch->cycleslip.arcLength[k]-options->minArcLength),options->excludeSmoothingConvergence-((int)max(1,epoch->cycleslip.arcLength[k]-options->minArcLength)));
			printSatSel(epoch,0,messagestr,i,options);
			check = 0;
		}

		// SBAS correction availability
		if ( options->SBAScorrections==1 && options->onlySBASiono==0 ) {
			if (epoch->sat[i].hasSBAScor == 0) {
				printSatSel(epoch,0,"SBAS corrections unavailable",i,options);
				check = 0;
			}
		}

		// Ephemeris availability check
		if ( epoch->sat[i].hasOrbitsAndClocks == 0 ) {
			printSatSel(epoch,0,"Ephemerides unavailable",i,options);
			check = 0;
		}
		
		// DCBs availability check
		if ( epoch->sat[i].hasDCBs == 0 ) {
			if (epoch->sat[i].hasDCBsofC1P1 == 1) {
				onlyHasPCorPCC = 1;
				for (j=0;j<options->totalFilterMeasurements;j++) {
					meas = options->measurement[j];
					if (whatIs(meas)==Pseudorange && meas!=PC && meas!=PCC) onlyHasPCorPCC = 0;
				}
			}
			if (!(epoch->sat[i].hasDCBs==1 || (epoch->sat[i].hasDCBsofC1P1==1 && onlyHasPCorPCC==1))) {
				printSatSel(epoch,0,"DCBs unavailable (strict TGD treatment)",i,options);
				check = 0;
			}
		}
		
		// Elevation check (if approximate position is good enough, ie <100Km)
		if ( (epoch->sat[i].elevation < options->elevationMask) && (epoch->receiver.aproxPositionError < 1e5) && (epoch->sat[i].hasOrbitsAndClocks == 1) ) { // 100Km
			sprintf(messagestr,"Elevation too low (%7.2f)",epoch->sat[i].elevation/d2r);
			printSatSel(epoch,0,messagestr,i,options);
			check = 0;
		}
		
		// Eclipsed satellite check
		if ( options->discardEclipsedSatellites ) {
			if (isEclipsed(&epoch->t,&epoch->sat[i],&epoch->satInfo[epoch->sat[i].GNSS][epoch->sat[i].PRN],sunPos,&auxT)) {
				if (auxT==0) {
					printSatSel(epoch,0,"Satellite under eclipse",i,options);
				} else {
					sprintf(messagestr,"Satellite was under eclipse %4.0f seconds ago",auxT);
					printSatSel(epoch,0,messagestr,i,options);
				}
				check = 0;
			}
		}

		// Ionex check (in case a IONEX file is used)
		if ( options->ionoModel == IONEX && epoch->sat[i].validIonex == 0 ) {
			sprintf(messagestr,"Missing IONEX data");
			printSatSel(epoch,0,messagestr,i,options);
			check = 0;
		}

		// FPPP check (in case a FPPP file is used)
		if ( options->ionoModel == FPPP && epoch->sat[i].validIonex == 0 ) {
			sprintf(messagestr,"Missing FPPP data");
			printSatSel(epoch,0,messagestr,i,options);
			check = 0;
		}

		// DGNSS check (in case a DGNSS file is used)
		if ( options->isDGNSS == 1 && epoch->sat[i].hasDGNSScor == 0 ) {
			sprintf(messagestr,"No DGNSS corrections");
			printSatSel(epoch,0,messagestr,i,options);
			check = 0;
		}

		// SBAS iono availability (if we are not in SBAS mode processing)
		if ( options->onlySBASiono == 1 ) {
			if (epoch->sat[i].hasSBAScor == 0) {
				printSatSel(epoch,0,"SBAS iono unavailable",i,options);
				check = 0;
			}
		}

		// GNSS System check (only prepared for GPS processing)
		if ( epoch->sat[i].GNSS != GPS ) {
			sprintf(messagestr,"Invalid GNSS System: %s",gnsstype2gnssstr(epoch->sat[i].GNSS));
			printSatSel(epoch,0,messagestr,i,options);
			check = 0;
		}
		

		// If everything is available satellite can be used
		if ( check ) {
			epoch->sat[i].available = 1;
			printSatSel(epoch,1,"",i,options);
			epoch->usableSatellites++;
		} else {
			epoch->sat[i].available = 0;
		}
	}
}

/*****************************************************************************
 * Name        : isEpochComputable
 * Description : Check (taking into account the usable satellites) if the 
 *               specified epoch can be computed
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  Reference epoch
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => Epoch not computable
 *                                         1 => Epoch computable
 *****************************************************************************/
int isEpochComputable (TEpoch *epoch) {
	if (epoch->usableSatellites>=4) return 1;
	else return 0;
}

/*****************************************************************************
 * Name        : calculateUnknowns
 * Description : Computes, from the options specified, how many unknowns the 
 *               filter has.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  obs                        I  N/A  Usable observations
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (int)            O  N/A  Number of unknowns in the filter
 *****************************************************************************/
int calculateUnknowns (int obs, TOptions *options) {
	int	value = 4;

	if (options->usePhase==1) value+=obs;
	if (options->estimateTroposphere==1) value++;
	return value;
}

/*****************************************************************************
 * Name        : initUnkinfo
 * Description : Initialises the information for the unknowns of the filter
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TUnkinfo  *unkinfo              O  N/A  TUnkinfo structure to store unknows
 * int  nunk                       I  N/A  Number of unknowns
 * int  observations               I  N/A  Number of observations
 * double  t                       I  s    Time since last filter run
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void initUnkinfo (TEpoch *epoch, TUnkinfo *unkinfo, int nunk, int observations, double t, TOptions *options){
	int	i;

	unkinfo->nunk=nunk;
	unkinfo->nunkvector=m2v(nunk-1,nunk-1)+1;
	unkinfo->par2unk[DR_UNK]=0;
	unkinfo->par2unk[DT_UNK]=3;
	i=4;
	if (options->estimateTroposphere==1) {
		unkinfo->par2unk[TROP_UNK]=i;
		i++;
	}
	if (options->usePhase==1) {
		unkinfo->par2unk[BIAS_UNK]=i;
		i+=observations;
	}
	
	// Forming PHI and QNOISE MATRIX (ONLY DIAGONAL)
	for (i=0;i<3;i++) {
		unkinfo->phi[unkinfo->par2unk[DR_UNK]+i] = options->filterParams[PHI_PAR][DR_UNK];
		unkinfo->qnoise[unkinfo->par2unk[DR_UNK]+i] = options->filterParams[Q_PAR][DR_UNK];
	}
	unkinfo->phi[unkinfo->par2unk[DT_UNK]] = options->filterParams[PHI_PAR][DT_UNK];
	unkinfo->qnoise[unkinfo->par2unk[DT_UNK]] = options->filterParams[Q_PAR][DT_UNK];

	if (options->estimateTroposphere==1) {
		unkinfo->phi[unkinfo->par2unk[TROP_UNK]] = options->filterParams[PHI_PAR][TROP_UNK];
		unkinfo->qnoise[unkinfo->par2unk[TROP_UNK]] = options->filterParams[Q_PAR][TROP_UNK]*(t/3600.);
	}
	if (options->usePhase==1) {
		for (i=0;i<observations;i++) {
			unkinfo->phi[unkinfo->par2unk[BIAS_UNK]+i] = options->filterParams[PHI_PAR][BIAS_UNK];
			unkinfo->qnoise[unkinfo->par2unk[BIAS_UNK]+i] = options->filterParams[Q_PAR][BIAS_UNK];
		}
	}
}

/*****************************************************************************
 * Name        : prepareCorrelation
 * Description : Updates the correlations for the solution computation
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TFilterSolution  *solution      IO N/A  Structure containing the filter 
 *                                         information of the previous epoch
 * int  *PRNlist                   I  N/A  List of satellites of the previous epoch
 * TUnkinfo  *unkinfo              I  N/A  TUnkinfo structure with information on 
 *                                         the current epoch unknowns
 * TUnkinfo  *prevUnkinfo          I  N/A  TUnkinfo structure with information on 
 *                                         the previous epoch unknowns
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void prepareCorrelation (TEpoch *epoch, TFilterSolution *solution, int *PRNlist, TUnkinfo *unkinfo, TUnkinfo *prevUnkinfo, TOptions *options) {
	int		i,j,alti,altj,k;
	double	updatedCorrelations[MAX_VECTOR_UNK];
	double	updatedParameterValues[MAX_UNK];
	int		new2old[MAX_SATELLITES_IN_VIEW];
	int		iniBiasUnk,endBiasUnk;
	int		new;
	int		found;
	int		ind;
	int		res;
	
	// Check for FIRST time
	if (solution->prevNumSatellites==0) {
		if (options->estimateTroposphere==1) {
			solution->correlations[m2v(unkinfo->par2unk[TROP_UNK],unkinfo->par2unk[TROP_UNK])] = 1/(options->filterParams[P0_PAR][TROP_UNK]);
		}
		//Loop to initialize covariance for all coordinates
		for (i=0;i<3;i++) {
			solution->correlations[m2v(unkinfo->par2unk[DR_UNK]+i,unkinfo->par2unk[DR_UNK]+i)] = 1/(options->filterParams[P0_PAR][DR_UNK]);
		}
		solution->correlations[m2v(unkinfo->par2unk[DT_UNK],unkinfo->par2unk[DT_UNK])] = 1/(options->filterParams[P0_PAR][DT_UNK]);
	}

	// Vector new2old will have the dictionary from new PRN positions to previous PRN positions
	for (i=0;i<epoch->numSatellites;i++) {
		new2old[i]=-1;
		k = satIs(epoch,epoch->sat[i].PRN,epoch->sat[i].GNSS);
		if ( epoch->sat[i].GNSS == GPS ) {
			for (j=0;j<solution->prevNumSatellites;j++) {
				if ( PRNlist[j] == epoch->sat[i].PRN && epoch->sat[i].available &&
					 epoch->cycleslip.outlier[k] == 0 && epoch->cycleslip.hasBeenCycleslip[k] == 0 ) {
					new2old[i]=j;
				}
			}
		}
		if ( epoch->cycleslip.hasBeenCycleslip[k] != 0 ) {
			epoch->cycleslip.hasBeenCycleslip[k] = 0; // clear cycle slip flag
		}
	}
	if (options->usePhase==1) {
		iniBiasUnk=unkinfo->par2unk[BIAS_UNK];
		endBiasUnk=unkinfo->par2unk[BIAS_UNK]+epoch->numSatellites-1;
	} else {
		iniBiasUnk=1;
		endBiasUnk=0;
	}
	
	// Updating correlations (including Phi and Q)
	if (solution->prevNumSatellites!=0) {
		for (i=0;i<prevUnkinfo->nunk;i++) {
			for (j=i;j<prevUnkinfo->nunk;j++) {
				solution->correlations[m2v(i,j)] *= prevUnkinfo->phi[i]*prevUnkinfo->phi[j];
			}
			solution->correlations[m2v(i,i)] += prevUnkinfo->qnoise[i];
		}
	}
	
	// Updating parameters
	for (i=0;i<prevUnkinfo->nunk;i++) {
		solution->parameters[i] *= prevUnkinfo->phi[i];
	}
	
	if (solution->prevNumSatellites!=0) { // If not in the first epoch
		// Update position parameters
		for(i=0,j=unkinfo->par2unk[DR_UNK];i<3;i++,j++) {
			solution->parameters[j] = solution->x[j] - epoch->receiver.aproxPosition[i];
		}
		// Update tropospheric parameter
		if (options->estimateTroposphere) {
			j = unkinfo->par2unk[TROP_UNK];
			solution->parameters[j] =  solution->x[j] - epoch->ZTD;
		}
	}
	
	// Satellite ambiguities that have disappeared or that have had a cycle-slip will be decorrelated from the
	// rest of ambiguities by putting an infinite in its self-correlation
	if (solution->prevNumSatellites!=0 && options->usePhase) {
		for (j=0;j<solution->prevNumSatellites;j++) {
			found = 0;
			for (i=0;i<epoch->numSatellites;i++) {
				if (new2old[i]==j) {
					found = 1;
					break;
				}
			}
			if (!found) {
				ind = prevUnkinfo->par2unk[BIAS_UNK]+j;
				solution->correlations[m2v(ind,ind)] = (options->filterParams[P0_PAR][BIAS_UNK]);
			}
		}
	}
	
	// Inverting correlation matrix from the former epoch
	res = cholinv_opt(solution->correlations,prevUnkinfo->nunk);
	if (res==-1) {
		sprintf(messagestr,"Problem in CHOLESKI2 res=%d\n",res);
		printError(messagestr,options);
		return;
	}
	
	// Creating new correlation matrix
	for (i=0;i<unkinfo->nunk;i++) {
		if (i>=iniBiasUnk && i<=endBiasUnk) {
			if (new2old[i-iniBiasUnk]!=-1) {
				updatedParameterValues[i] = solution->parameters[new2old[i-iniBiasUnk]+iniBiasUnk];
			} else {
				updatedParameterValues[i] = 0;
			}
		} else {
			updatedParameterValues[i] = solution->parameters[i];
		}
		for (j=i;j<unkinfo->nunk;j++) {
			new=0;
			if ((i>=iniBiasUnk && i<=endBiasUnk) || (j>=iniBiasUnk && j<=endBiasUnk)) {
				if (i>=iniBiasUnk && i<=endBiasUnk) {
					if (new2old[i-iniBiasUnk]==-1) {new=1;
					} else { alti = prevUnkinfo->par2unk[BIAS_UNK]+new2old[i-iniBiasUnk]; }
				} else { alti = i;}
				if (j>=iniBiasUnk && j<=endBiasUnk) {
					if (new2old[j-iniBiasUnk]==-1) { new=1;
					} else { altj = prevUnkinfo->par2unk[BIAS_UNK]+new2old[j-iniBiasUnk];}
				} else { altj = j;}
				if (new) {
					if (i==j) updatedCorrelations[m2v(i,j)] = 1/(options->filterParams[P0_PAR][BIAS_UNK]);
					else updatedCorrelations[m2v(i,j)] = 0;
				} else {
					updatedCorrelations[m2v(i,j)] = solution->correlations[m2v(alti,altj)];
				}
			} else {
				updatedCorrelations[m2v(i,j)] = solution->correlations[m2v(i,j)];
			}
		}
	}

	memcpy(solution->correlations,updatedCorrelations,sizeof(double)*unkinfo->nunkvector);
	memcpy(solution->parameters,updatedParameterValues,sizeof(double)*unkinfo->nunk);
}

/*****************************************************************************
 * Name        : atwa_atwy_insertline
 * Description : Insert an equation (prefilter values and design vector) into 
 *               the AtWA matrix and AtWY vector
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *atwa                   IO N/A  At*W*A matrix
 * double  *atwy                   IO N/A  At*W*Y vector
 * double  *unk                    I  N/A  Design vector
 * double  prefit                  I  N/A  Prefit value
 * int  nunk                       I  N/A  Number of unknowns
 * double  sigma2                  I  N/A  Sigma^2 associated with the prefit
 *****************************************************************************/
void atwa_atwy_insertline (double *atwa, double *atwy, double *unk, double prefit, int nunk, double sigma2) {
	int j,k;

	for (j=0;j<nunk;j++) {
		atwy[j] = atwy[j] + unk[j]*prefit/sigma2;
		for (k=0;k<=j;k++) {
			atwa[m2v(k,j)] = atwa[m2v(k,j)] + unk[j]*unk[k]/sigma2;
		}
	}
}

/*****************************************************************************
 * Name        : atwa_atwy_insertLineWithCorrelation
 * Description : Similar to atwa_atwy_insert for apriori values (with a full 
 *               correlations vector)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *atwa                   IO N/A  At*W*A matrix
 * double  *atwy                   IO N/A  At*W*Y vector
 * double  *corr                   I  N/A  Correlations vector
 * double  apriorivalue            I  N/A  Apriori value (equivalent to prefit)
 * int  nunk                       I  N/A  Number of unknowns
 * double  multiplier              I  N/A  Multiplier in the Fi matrix (equivalent
 *                                         to the design vector)
 * int  iniunk                     I  N/A  Initial unk to start updating
 *****************************************************************************/
void atwa_atwy_insertLineWithCorrelation (double *atwa, double *atwy, double *corr, double apriorivalue, int nunk, double multiplier, int iniunk) {
	int i;

	for (i=0;i<nunk;i++) {
		atwy[i] += multiplier*apriorivalue*corr[i];
	}

	for (i=iniunk;i<nunk;i++) {
		atwa[m2v(iniunk,i)] += multiplier*multiplier*corr[i];
	}
}

/*****************************************************************************
 * Name        : designSystem
 * Description : Designs the system for a specific measurement marked in the filter.
 *               Computes At*W*A and At*W*Y for the specified epoch 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * double  *atwa                   IO N/A  At*W*A matrix
 * double  *atwy                   IO N/A  At*W*Y vector
 * TUnkinfo  *unkinfo              I  N/A  Structure containing the information of
 *                                         the unknowns 
 * MeasurementType  measType	   I  N/A  Measurement to be included in the filter
 * int  filterInd                  I  N/A  Index of the measurement of the filter
 * double *prefits		   O  m    Vector with the prefits
 * double *weights		   O  m    Vector with the measurement weights
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void designSystem (TEpoch *epoch, double *atwa, double *atwy, TUnkinfo *unkinfo, enum MeasurementType measType, int filterInd, double **G, double *prefits, double *weights, TOptions *options) {
	int			i,j,k,l=0;
	double		unk[MAX_UNK],prefit,measurement,model;
	int			ns=0;
	TSatellite	*sat;
	double		weight;
	double		stdDev;
	int			ret;
	int			sats[MAX_SATELLITES_IN_VIEW];
	double		smoothWith;
	int			dontuse;

	epoch->numSatSolutionSBAS=0;
	epoch->numSatnotUsedSBAS=0;

	for (i=0;i<epoch->numSatellites;i++) {
		dontuse = 0;
		sat = &epoch->sat[i];
		k = satIs(epoch,epoch->sat[i].PRN,epoch->sat[i].GNSS);
		
		// If not GPS, the satellite will not be included in the filter
		if (sat->GNSS != GPS) dontuse=1;


		ret = getMeasModelValue(epoch,sat->GNSS,sat->PRN,measType,&measurement,&model);
		if (ret!=1) {
			sat->available=0;
			model = 0;
			measurement = 0;
		}
		// Smoothing
		if (options->smoothEpochs!=0 && options->smoothMeas[filterInd]!=NA) {
			ret = getMeasModelValue(epoch,sat->GNSS,sat->PRN,options->smoothMeas[filterInd],&smoothWith,NULL);
			if (ret!=1) {
				sat->available=0;
				model = 0;
				smoothWith = 0;
			}
			prefit = epoch->cycleslip.smoothedMeas[k][filterInd] + smoothWith - model;
		} else {
			prefit = measurement - model;
		}

		// Compute weights and available satellites
		if (sat->available) {
			stdDev = sat->measurementWeights[filterInd];
			sats[ns] = sat->PRN;
			ns++;
			//Count the number in satellites suitable for SBAS (currently only GPS) and used in solution computation
			if (sat->GNSS == GPS) {
				epoch->listPRNSatSolutionSBAS[epoch->numSatSolutionSBAS]=sat->PRN;
				epoch->listConsSatSolutionSBAS[epoch->numSatSolutionSBAS]=gnsstype2char(sat->GNSS);
				epoch->numSatSolutionSBAS+=1;
			}
		} else {
			// Satellite not counted for number of satellites
			stdDev = SIGMA_INF;
			//Count the number in satellites suitable for SBAS (currently only GPS) but not used in solution computation
			if (sat->GNSS == GPS) {
				epoch->listPRNSatnotUsedSBAS[epoch->numSatnotUsedSBAS]=sat->PRN;
				epoch->listConsSatnotUsedSBAS[epoch->numSatnotUsedSBAS]=gnsstype2char(sat->GNSS);
				epoch->numSatnotUsedSBAS+=1;
			}
		}
		weight = stdDev * stdDev;
	
		// Stop if it is declared as Don't Use
		if (dontuse) continue;

		
		// Print prefit residuals
		printPrefit(&epoch->t,sat,measType,prefit,measurement,model,stdDev,options);


		// unk vector is the a row of the design matrix
		for (j=0;j<unkinfo->nunk;j++) unk[j]=0;

		// Position unknowns
		unk[unkinfo->par2unk[DR_UNK]] = -sat->LoS[0];
		unk[unkinfo->par2unk[DR_UNK]+1] = -sat->LoS[1];
		unk[unkinfo->par2unk[DR_UNK]+2] = -sat->LoS[2];

		if (options->stanfordesa==1 && sat->available) {
			G[l][0] = cos ( epoch->sat[i].elevation ) * sin ( epoch->sat[i].azimuth ) ;
			G[l][1] = cos ( epoch->sat[i].elevation ) * cos ( epoch->sat[i].azimuth ) ;
			G[l][2] = sin ( epoch->sat[i].elevation ) ;
			G[l][3] = 1 ;

			prefits[l]=prefit;
			weights[l]=weight;
			l++;
		}

		// Time unknown
		unk[unkinfo->par2unk[DT_UNK]] = 1;

		// Troposphere unknown
		if (options->estimateTroposphere)
			unk[unkinfo->par2unk[TROP_UNK]] = sat->tropWetMap;
			
		// Carrier-phase ambiguities unknowns - if applicable
		if (whatIs(measType)==CarrierPhase) {
			unk[unkinfo->par2unk[BIAS_UNK]+i] = 1;
		}
		
		// Insert the data in atwa and atwy. Only include if available or it is a carrierphase
		if (sat->available || whatIs(measType)==CarrierPhase) {
			atwa_atwy_insertline(atwa,atwy,unk,prefit,unkinfo->nunk,weight);
		}
		
	}

	printEpochsat(&epoch->t, measType, ns, sats, options);
}




/*****************************************************************************
 * Name        : designSystemIniValues
 * Description : Updates the filter with the information of the former epoch 
 *               (initial values for the current one)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * double  *atwa                   IO N/A  At*W*A matrix
 * double  *atwy                   IO N/A  At*W*Y vector
 * TUnkinfo  *unkinfo              I  N/A  Structure containing the information of
 *                                         the unknowns 
 * TFilterSolution  *solution      I  N/A  Structure with the information of the 
 *                                         solution of the previous epoch
 *****************************************************************************/
void designSystemIniValues (TEpoch *epoch, double *atwa, double *atwy, TUnkinfo *unkinfo, TFilterSolution *solution) {
	int			i,j;
	double		corr[MAX_UNK];

	for (i=0;i<unkinfo->nunk;i++) {
		for (j=0;j<unkinfo->nunk;j++) {
			corr[j]=solution->correlations[m2v(i,j)];
		}
		atwa_atwy_insertLineWithCorrelation(atwa,atwy,corr,solution->parameters[i],unkinfo->nunk,1,i);
	}
}

/*****************************************************************************
 * Name        : calculatePostfits
 * Description : Calculate and print the postfits after the filtering process
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * double  *solution               I  N/A  Solution of the filter
 * TUnkinfo  *unkinfo              I  N/A  Structure containing the information of
 *                                         the unknowns
 * MeasurementType  measType       I  N/A  Measurement type of the postfit
 * int  filterInd                  I  N/A  Index of the measurement of the filter
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (double)         O  m    Standard deviation of the postfits
 *****************************************************************************/
double calculatePostfits (TEpoch *epoch, double *solution, TUnkinfo *unkinfo, enum MeasurementType measType, int filterInd, TOptions *options) {
	double		stddev2postfit=0;
	double		postfit;
	double		measurement,model;
	double		estimatedParameters;
	double		geometricCorrection;
	double		smoothWith;
	TSatellite	*sat;
	int			i,j,k;
	int			ret;

	for (i=0;i<epoch->numSatellites;i++) {
		sat = &epoch->sat[i];
		j = satIs(epoch,epoch->sat[i].PRN,epoch->sat[i].GNSS);
		if (sat->available) {
			estimatedParameters = solution[unkinfo->par2unk[DT_UNK]];
			geometricCorrection = 0;
			for (k=0;k<3;k++) {
				geometricCorrection += -sat->LoS[k]*solution[unkinfo->par2unk[DR_UNK]+k];
			}
			estimatedParameters += geometricCorrection;
			if(options->estimateTroposphere) {
				estimatedParameters += sat->tropWetMap * solution[unkinfo->par2unk[TROP_UNK]];
			}
			if(whatIs(measType)==CarrierPhase) {
				estimatedParameters += solution[unkinfo->par2unk[BIAS_UNK]+i];
			}
			
			getMeasModelValue(epoch,sat->GNSS,sat->PRN,measType,&measurement,&model);
			
			// Smoothing
			if (options->smoothEpochs!=0 && options->smoothMeas[filterInd]!=NA) {
				ret = getMeasModelValue(epoch,sat->GNSS,sat->PRN,options->smoothMeas[filterInd],&smoothWith,NULL);
				if (ret!=1) {
					sat->available = 0;
					model = 0;
					smoothWith = 0;
					epoch->cycleslip.smoothedMeas[j][filterInd] = 0.0;
				}
				postfit = epoch->cycleslip.smoothedMeas[j][filterInd] + smoothWith - (model + estimatedParameters);
			} else {
				postfit = measurement - (model + estimatedParameters);
			}
			
			if (whatIs(measType)==CarrierPhase) {
				printPostfit(&epoch->t, sat, measType, postfit, measurement, model, estimatedParameters, solution[unkinfo->par2unk[BIAS_UNK]+i], options);
			} else {
				printPostfit(&epoch->t, sat, measType, postfit, measurement, model, estimatedParameters, 0, options);
			}

			stddev2postfit += postfit*postfit;
		}
	}
	stddev2postfit /= epoch->usableSatellites;
	return stddev2postfit;
}

/*****************************************************************************
 * Name        : computeSolution
 * Description : Obtain the filter solution for the specified epoch
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * double  *stddev2postfit         O  m    Standard deviation of the postfits
 * TFilterSolution  *solution      IO N/A  Structure with the information of the 
 *                                         solution of the previous epoch (it is 
 *                                         updated with the solution of the present epoch)
 * double  *newcorrelations        O  N/A  Correlations after the filter run
 * double  *newparameters          O  N/A  Parameters after the filter run
 * TUnkinfo  *unkinfo              I  N/A  Structure containing the information of
 *                                         the unknowns
 * TStdESA *StdESA                 O  N/A  TStdESA struct                                        
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => Problem in the matrix inversion
 *                                         1 => Solution found
 *****************************************************************************/
int computeSolution (TEpoch *epoch,double *stddev2postfit, TFilterSolution *solution, double *newcorrelations, double *newparameterValues, TUnkinfo *unkinfo, TStdESA *StdESA, TOptions *options) {
	double 		*atwa;
	double 		*atwy;
	double 		*aux;
	double		*prefits;
	double		*weights;
	double		**G;
	int		*k;
	int		*kmask;
	int 		i,j;
	int		res;

	atwy=malloc(sizeof(double)*unkinfo->nunk);
	atwa=malloc(sizeof(double)*unkinfo->nunkvector);
	prefits=malloc(sizeof(double)*epoch->numSatellites);
	weights=malloc(sizeof(double)*epoch->numSatellites);
	G=malloc(sizeof(double *)*epoch->numSatellites);
	for(i=0;i<epoch->numSatellites;i++) {
		G[i]=malloc(sizeof(double)*4);
	}


	// Initializing At*W*A and At*W*Y
	for (i=0;i<unkinfo->nunk;i++) {
		atwy[i]=0;
	}
	for (i=0;i<unkinfo->nunkvector;i++) {
		atwa[i]=0;
	}
	
	// Forming At*W*A and At*W*Y
	for (i=0;i<options->totalFilterMeasurements;i++) {
		designSystem(epoch,atwa,atwy,unkinfo,options->measurement[i],i,G,prefits,weights,options);
	}

	designSystemIniValues(epoch,atwa,atwy,unkinfo,solution);

	res = cholinv_opt(atwa,unkinfo->nunk);

	if (res==-1) {
		free(atwa);
		free(atwy);
		sprintf(messagestr,"Problem in CHOLESKI1 res=%d\n",res);
		printError(messagestr,options);
		return 0;
	}
	
	// Saving parameters formal errors
//	for (i=0;i<unkinfo->nunk;i++) {
//		solution->formalError[i] = sqrt(atwa[m2v(i,i)]);
//	}
	
	mxv(atwa,atwy,unkinfo->nunk);


	// Updating correlations
	if (solution->correlations!=NULL) {
		for (i=0;i<unkinfo->nunk;i++) {
			for (j=0;j<=i;j++) {
				newcorrelations[m2v(i,j)] = atwa[m2v(i,j)];
			}
		}
		
		memcpy (newparameterValues,atwy,sizeof(double)*unkinfo->nunk);
	}

	// Building solution vector
	for (i=0;i<unkinfo->nunk;i++) {
		solution->x[i] = atwy[i];
	}
	for (i=0;i<3;i++) {
		solution->x[unkinfo->par2unk[DR_UNK]+i] = epoch->receiver.aproxPosition[i] + atwy[unkinfo->par2unk[DR_UNK]+i];
	}

	if (options->estimateTroposphere) solution->x[unkinfo->par2unk[TROP_UNK]] += epoch->ZTD;

	// High 'dr' checking => Initial aproximate position far from estimated position
	aux = &atwy[unkinfo->par2unk[DR_UNK]];
//	if (aux[0]*aux[0]+aux[1]*aux[1]+aux[2]*aux[2] > options->maxdr2) {
//		for (i=0;i<3;i++) epoch->receiver.aproxPosition[i] = solution->x[unkinfo->par2unk[DR_UNK]+i];
//			printf("MAXDR FOUND: %f>%f\n",aux[0]*aux[0]+aux[1]*aux[1]+aux[2]*aux[2],options->maxdr2);
//		computeSolution (epoch,stddev2postfit,solution,newcorrelations,newparameterValues,unkinfo,options);
//	}

	// Postfit calculation
	for (i=0;i<options->totalFilterMeasurements;i++) {
		*stddev2postfit = calculatePostfits(epoch,atwy,unkinfo,options->measurement[i],i,options);
	}
	// Formal Error Calculation
	solution->dop=0;

	for (i=0;i<3;i++) {
		solution->dop += atwa[m2v(unkinfo->par2unk[DR_UNK]+i,unkinfo->par2unk[DR_UNK]+i)];
	}
	solution->dop = sqrt(solution->dop);

	// Horizontal and Vertical Dilution Of Precision Computations
	calculateDOP(epoch,solution,options);

	//Compute Stanford-ESA plots (only for SBAS)
	if(options->SBAScorrections==1 && options->onlySBASiono==0) {
		if(options->stanfordesa==1) {
			k=malloc(sizeof(int)*(epoch->numSatSolutionSBAS+1));
			kmask=malloc(sizeof(int)*(epoch->numSatSolutionSBAS+1));
			k[0]=0;
			for(i=0;i<=epoch->numSatSolutionSBAS;i++) {
				kmask[i]=1;
			}
			StanfordESAComputationRecursive(k,kmask,epoch->numSatSolutionSBAS,0,G,prefits,weights,unkinfo,StdESA,options);
			free(k);
			free(kmask);
		}
	}


	// Freeing memory
	free(atwa);
	free(atwy);
	free(weights);
	free(prefits);
	for(i=0;i<epoch->numSatellites;i++) {
		free(G[i]);
	}
	free(G);

	return 1;
}

/*****************************************************************************
 * Name        : calculateDOP
 * Description : Computes the Dilution of Precision of the current epoch  
 *               obtain the solution (only for the used satellites)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TFilterSolution  *solution      I  N/A  Structure with the information of the 
 *                                         solution of the previous epoch (it is 
 *                                         updated with the solution of the present epoch)
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void calculateDOP (TEpoch *epoch, TFilterSolution *solution, TOptions *options) {

	int		i,j;
	double		dmajor,weight;
		double		RowMatrixG_NEU[4], RowMatrixG_XYZ[4];
		double		VectorGtG_NEU[10], VectorGtG_XYZ[10], FalsePrefit[4];	
	double		VectorGtWG_NEU[10]; 


	//Initialize Variables
	solution->GDOP = 0.0; solution->PDOP = 0.0; solution->TDOP = 0.0; solution->VPL = 0.0;
	solution->HDOP = 0.0; solution->VDOP = 0.0; solution->HPL = 0.0;

	for (j=0;j<4;j++)   { RowMatrixG_NEU[j] = 0.0;   RowMatrixG_XYZ[j] = 0.0; FalsePrefit[j] = 0.0; }
	for (j=0;j<10;j++)  {  VectorGtG_NEU[j] = 0.0;   VectorGtG_XYZ[j] = 0.0; VectorGtWG_NEU[j]=0.0; }

	// Build Geometry Matrix and its Transpose, then store its multiplication in a vector
	for (i=0;i<epoch->numSatellites;i++) 
		if (epoch->sat[i].available) {
			RowMatrixG_NEU[0] = cos ( epoch->sat[i].elevation ) * sin ( epoch->sat[i].azimuth ) ;
			RowMatrixG_NEU[1] = cos ( epoch->sat[i].elevation ) * cos ( epoch->sat[i].azimuth ) ;
			RowMatrixG_NEU[2] = sin ( epoch->sat[i].elevation ) ;
			RowMatrixG_NEU[3] = 1 ;

			RowMatrixG_XYZ[0] = epoch->sat[i].LoS[0] ;
			RowMatrixG_XYZ[1] = epoch->sat[i].LoS[1] ;
			RowMatrixG_XYZ[2] = epoch->sat[i].LoS[2] ;
			RowMatrixG_XYZ[3] = 1 ;

			atwa_atwy_insertline (VectorGtG_XYZ, FalsePrefit, RowMatrixG_XYZ, 0.0, 4, 1.0);
			atwa_atwy_insertline (VectorGtG_NEU, FalsePrefit, RowMatrixG_NEU, 0.0, 4, 1.0);

			weight=epoch->sat[i].measurementWeights[0]*epoch->sat[i].measurementWeights[0];
			atwa_atwy_insertline (VectorGtWG_NEU,FalsePrefit,RowMatrixG_NEU,0.0,4,weight);

		}

	// Invert the Gt*G Matrixes:
	cholinv_opt(VectorGtG_NEU,4);
	cholinv_opt(VectorGtG_XYZ,4);


	// Compute DOPs
	solution->HDOP = sqrt ( VectorGtG_NEU[0] + VectorGtG_NEU[2] );
	solution->VDOP = sqrt ( VectorGtG_NEU[5] );
	solution->GDOP = sqrt ( VectorGtG_XYZ[0] + VectorGtG_XYZ[2] + VectorGtG_XYZ[5] + VectorGtG_XYZ[9] );
	solution->PDOP = sqrt ( VectorGtG_XYZ[0] + VectorGtG_XYZ[2] + VectorGtG_XYZ[5] );
	solution->TDOP = sqrt ( VectorGtG_XYZ[9] );

	//Compute protection level for SBAS only if SBAS is selected
	if(options->SBAScorrections==1 && options->onlySBASiono==0) {
		cholinv_opt(VectorGtWG_NEU,4);
		dmajor=sqrt(0.5*(VectorGtWG_NEU[2]+VectorGtWG_NEU[0])+sqrt(0.25*pow(VectorGtWG_NEU[2]-VectorGtWG_NEU[0],2)+VectorGtWG_NEU[1]*VectorGtWG_NEU[1]));
		if(options->precisionapproach==1) {
			solution->HPL=KHPA*dmajor;
		} else {
			solution->HPL=KHNPA*dmajor;
		}
		solution->VPL=KV*sqrt(VectorGtWG_NEU[5]);

	}

}

/*****************************************************************************
 * Name        : StanfordESAComputationRecursive
 * Description : Recursive function to compute all the possible navigation
 *               solution with all geometries for Stanford-ESA plot
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TFilterSolution  *solution      I  N/A  Structure with the information of the 
 *                                         solution of the previous epoch (it is 
 *                                         updated with the solution of the present epoch)
 * TStdESA *StdESA                 O  N/A  TStdESA struct                                        
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void StanfordESAComputationRecursive (int *k, int *kmask, int numsat, int depth, double **G, double *prefits, double *weights, TUnkinfo *unkinfo, TStdESA *StdESA, TOptions  *options) {

	int				i;
	int				res1;
	unsigned int	posx,posy;
	double			dmajor,HPL,VPL,Herror,Verror;
	double			*Vector_GtWG;
	double			*Vector_GtY;
	char			strmask[100],auxstr[100];

	//Mount matrix data
	Vector_GtWG=malloc(sizeof(double)*unkinfo->nunkvector);
	Vector_GtY=malloc(sizeof(double)*unkinfo->nunk);

	// Initializing At*W*A and At*W*Y
	for (i=0;i<unkinfo->nunk;i++) {
		Vector_GtY[i]=0;
	}
	for (i=0;i<unkinfo->nunkvector;i++) {
		Vector_GtWG[i]=0;
	}
	

	for(i=0;i<numsat;i++) {
		if(kmask[i]==1) {
			atwa_atwy_insertline(Vector_GtWG,Vector_GtY,G[i],prefits[i],unkinfo->nunk,weights[i]);
		}
	}	

	res1=cholinv_opt(Vector_GtWG,unkinfo->nunk);

	if(res1==0) {
		dmajor=sqrt(0.5*(Vector_GtWG[2]+Vector_GtWG[0])+sqrt(0.25*pow(Vector_GtWG[2]-Vector_GtWG[0],2)+Vector_GtWG[1]*Vector_GtWG[1]));
		if(options->precisionapproach==1) {
			HPL=KHPA*dmajor;
		} else {
			HPL=KHNPA*dmajor;
		}
		VPL=KV*sqrt(Vector_GtWG[5]);

		
		mxv(Vector_GtWG,Vector_GtY,unkinfo->nunk);
	
		Herror=sqrt(Vector_GtY[0]*Vector_GtY[0]+Vector_GtY[1]*Vector_GtY[1]);
		Verror=fabs(Vector_GtY[2]);

	} else {
		//Singular matrix not positive defined
		strmask[0]='\0';
		for(i=0;i<numsat;i++) {
			sprintf(auxstr,"%s%1d",strmask,kmask[i]);
			sprintf(strmask,"%s",auxstr);
		}
		sprintf(messagestr,"Geometry with singular matrix when using %d satellites, with satellite mask %s in Stanford-ESA computation",numsat-depth,strmask);
		printInfo(messagestr,options);
	}

	if((numsat-depth)>4) {
		for(i=k[depth];i<=(numsat-depth-1);i++) {
			k[depth+1]=i;
			kmask[i+depth]=0;
			StanfordESAComputationRecursive(k,kmask,numsat,depth+1,G,prefits,weights,unkinfo,StdESA,options);
			kmask[i+depth]=1;
		}
	}

	if(res1==0) {
		//Locate pixel position in grid and add 1 to the number of ocurrencies in that pixel
		//Horizontal error grid
		posx=(unsigned int)(Herror/StdESA->xstep);
		posy=(unsigned int)(HPL/StdESA->ystep);
		if(posx>=(StdESA->numpixelhor-1)) {
			if((HPL-Herror)>0) {
				//It is not an MI. Put in position StdESA->numpixelhor-2 to avoid
				//confusing it with an MI point
				posx=StdESA->numpixelhor-2;
			} else {
				posx=StdESA->numpixelhor-1;
			}
		}
		if(posy>=StdESA->numpixelver) posy=StdESA->numpixelver-1;
		StdESA->counthor[posy][posx]+=1;
		//Vertical error grid
		posx=(unsigned int)(Verror/StdESA->xstep);
		posy=(unsigned int)(VPL/StdESA->ystep);
		if(posx>=(StdESA->numpixelhor-1)) {
			if((VPL-Verror)>0) {
				//It is not an MI. Put in position StdESA->numpixelhor-2 to avoid
				//confusing it with an MI point
				posx=StdESA->numpixelhor-2;
			} else {
				posx=StdESA->numpixelhor-1;
			}
		}
		if(posy>=StdESA->numpixelver) posy=StdESA->numpixelver-1;
		StdESA->countver[posy][posx]+=1;
	}


	//Free memory
	free(Vector_GtWG);
	free(Vector_GtY);
}


/*****************************************************************************
 * Name        : updatePRNlist
 * Description : Update the PRNlist vector for the list of satellites visible 
 *               in the current epoch
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * int  *PRNlist                   O  N/A  List of satellites
 * TFilterSolution  *solution      I  N/A  Structure with the information of the 
 *                                         solution
 *****************************************************************************/
void updatePRNlist (TEpoch *epoch, int *PRNlist, TFilterSolution *solution) {
	int i;

	solution->prevNumSatellites = 0;
	for (i=0;i<epoch->numSatellites;i++) {
		if (epoch->sat[i].GNSS == GPS) {
			solution->prevNumSatellites++;
			PRNlist[i]=epoch->sat[i].PRN;
		} else {
			PRNlist[i]=0;
		}
	}
}

/*****************************************************************************
 * Name        : Kalman
 * Description : Process the filter for the specified epoch
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TFilterSolution  *solution      IO N/A  Structure with the information of the 
 *                                         previous solution (updated)
 * int  *PRNlist                   IO N/A  List of satellites of the previous epoch
 *                                         (updated)
 * TUnkinfo  *prevUnkinfo          IO N/A  Structure containing the information
 *                                         of the unknowns of the previous epoch
 *                                         (updated)
 * TStdESA *StdESA                 O  N/A  TStdESA struct                                        
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => Epoch not computable
 *                                         1 => Epoch computed
 *                                         2 => Initial position wrong, model and Kalman must be relaunched
 *                                         3 => (SBAS mode only) Not enough satellites for positioning after 
 *                                              deselecting satellites with fillUsableSatellites (due to no SBAS criteria), 
 *                                              try computing again model in NPA mode or (if we are in PA and we can change mode)
 *****************************************************************************/
int Kalman (TEpoch *epoch, TFilterSolution *solution, int *PRNlist, TUnkinfo *prevUnkinfo, TStdESA *StdESA, TOptions *options) {
	double			newcorrelations[MAX_VECTOR_UNK];
	double			stddev2postfit;
	double			newparameterValues[MAX_UNK];
	TUnkinfo		unkinfo;
	TFilterSolution	aux;
	static int		first = 1;
	static TTime	prevT;
	double 			distance;
	int				i;
	int			numsatellites;

	
	if (first) {
		memcpy(&prevT,&epoch->t,sizeof(TTime)); 
	}

	fillUsableSatellites(epoch,options);
	//If we are in SBAS processing, we have less than 4 satellites, 
	//check if we can switch mode, we are in PA, and if adding the unselected satellites due to SBAS we can reach 4 satellites
	if(epoch->usableSatellites<4) {
		if ( options->SBAScorrections == 1 && options->onlySBASiono == 0 ) {
			if (options->switchmode==1 && options->precisionapproach==1 && (epoch->usableSatellites+epoch->numsatdiscardedSBAS>=4) ) {
				// Empty print buffers for SATSEL messages
				options->printInBuffer=3;
				printBuffers(epoch,options);
				return 3;
			}
		}
	}

	// Step detector for SBAS
	if ( options->stepdetector == 1 ) {
		StepDetector(epoch,options);
	}
	// Empty print buffers for SATSEL messages
	options->printInBuffer=2;
	printBuffers(epoch,options);

	if (isEpochComputable(epoch)) {
		memcpy(&aux,solution,sizeof(TFilterSolution));	
		initUnkinfo(epoch,&unkinfo,calculateUnknowns(epoch->numSatellites,options),epoch->numSatellites,fabs(tdiff(&epoch->t,&prevT)),options);
		prepareCorrelation(epoch,solution,PRNlist,&unkinfo,prevUnkinfo,options);
		if (!computeSolution(epoch,&stddev2postfit,solution,newcorrelations,newparameterValues,&unkinfo,StdESA,options)) {
			memcpy(solution,&aux,sizeof(TFilterSolution));
			return 0;
		}
		
		if (options->receiverPositionSource >= rpCALCULATE) {
			distance = sqrt((solution->x[0]-epoch->receiver.aproxPosition[0])*(solution->x[0]-epoch->receiver.aproxPosition[0]) + 
							(solution->x[1]-epoch->receiver.aproxPosition[1])*(solution->x[1]-epoch->receiver.aproxPosition[1]) + 
							(solution->x[2]-epoch->receiver.aproxPosition[2])*(solution->x[2]-epoch->receiver.aproxPosition[2]));
			for (i=0;i<3;i++) {
				epoch->receiver.aproxPosition[i] = solution->x[i];
			}
			epoch->receiver.aproxPositionError = solution->dop;
			sprintf(messagestr,"Apriori position updated to: %13.4f %13.4f %13.4f   Distance: %13.4f",epoch->receiver.aproxPosition[0],epoch->receiver.aproxPosition[1],epoch->receiver.aproxPosition[2],distance);
			printInfo(messagestr,options);
			if (distance > 100) {
				memcpy(solution,&aux,sizeof(TFilterSolution));
				return 2;
			}
			epoch->receiver.receiverSolConverged=1;
		}

		printFilter(&epoch->t,unkinfo.nunk,solution,options);

		// Preparing values for next epoch
		memcpy(prevUnkinfo,&unkinfo,sizeof(TUnkinfo));
		memcpy(solution->correlations,newcorrelations,sizeof(double)*unkinfo.nunkvector);
		memcpy(solution->parameters,newparameterValues,sizeof(double)*unkinfo.nunk);

		updatePRNlist(epoch,PRNlist,solution);
		
		// Clean first flag
		first = 0;
		
		// Save last computed time
		memcpy(&prevT,&epoch->t,sizeof(TTime)); 
		return 1;
	} else {
		return 0;
	}
}
