/****************************************************************************
   Copyright & License:
   ====================
   
   Copyright 2009 - 2016 gAGE/UPC & ESA
   
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 ***************************************************************************/

/****************************************************************************
 * Copyright: gAGE/UPC & ESA
 * Project: EDUNAV GNSS Lab Tool 
 * Author: Pere Ramos-Bosch (group of Astronomy and GEomatics - gAGE/UPC)
 * Curator: Adria Rovira-Garcia ( gAGE/UPC )
 * Developers: Deimos Ibáñez Segura ( gAGE/UPC )
 *             Jesus Romero Sanchez ( gAGE/UPC )
 *          glab.gage @ upc.edu
 * File: output.c
 * Code Management Tool File Version: 3.1  Revision: 1
 * Date: 2016/10/21
 ***************************************************************************/

/****************************************************************************
 * MODULE DESCRIPTION
 *
 * Name: output
 * Language: C
 *
 * Purpose:
 *  The purpose of this module is to act as output for the solution of the filter
 *
 * Function:
 *  This class is in charge of outputting the results of the Kalman filter done 
 *  by the filter module
 *
 * Dependencies:
 *  dataHandling, filter, model
 *
 * Files modified:
 *  None
 *
 * Files read:
 *  None
 *
 * Resources usage:
 *  See Design Document
 ****************************************************************************/

/****************************
 *       RELEASE_HISTORY
 * -------------
 *          gLAB v1.3.1
 * Release: 2010/01/22
 * Change Log: First public version
 * -------------
 *          gLAB v1.3.2
 * Release: 2010/02/02
 * Change Log: Imported SATDIFF, SATSTAT and SATSTATTOT printing from gLAB.c
 *             Added satellite SATPVT message
 *             Change in SATDIFF and CS messages to show the GNSS system that 
 *                  the satellite belongs to
 *             It is now possible to print in the MODEL and SATDIFF messages 
 *                  the inertial velocity of the satellite
 *             PREFIT message now includes the weight that will be applied to 
 *                  the filter
 *             ZTD and estimated residual tropospheric delay is now printed 
 *                  in the FILTER message
 *             Forward/Backward processing included
 *             SATSTAT message will now provide the GNSS system it refers to
 * -------------
 *          gLAB v1.4.0
 * Release: 2010/06/21
 * Change Log: Included prealignCP option to set/unset carrier-phase prealignment
 *             C1/P1 flexibility option is added to allow flexible handling of C1 and P1
 *             Receiver apriori position INFO message now provides the origin of the
 *                  position
 *             Changed printInput and printMeas function to show all the measurements 
 *                  independently that they have been set as available or unavailable
 *             Corrected a bug when computing the position in trajectories
 *             Cycle-slip detector changed to specify when it does not have enough measurements
 * -------------
 *          gLAB v1.4.0
 * Release: 2010/08/19
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.1
 * Release: 2010/07/09
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.2
 * Release: 2010/07/31
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.3
 * Release: 2010/08/31
 * Change Log: Satellite elevation and azimuth addition in the print MODEL message.
 * -------------
 *          gLAB v1.4.4
 * Release: 2010/09/22
 * Change Log: Included new DCB management.
 * -------------
 *          gLAB v1.4.5
 * Release: 2010/10/31
 * Change Log: Tropospheric correction option is extended to select independently the nominal and mapping used.
 *             Dilution Of Precision extension: (GDOP, PDOP, TDOP, VDOP, HDOP) new variables.
 *             Some minor floating length adjustments. 
 *             Filter 'inf' value for clock and coordinates noise process and initial covariance substituted
 *                  to real floating quantities.
 * -----------
 * Release: 2012/12/31
 * Change Log: gLAB version released together with the Educational Book:
 *             "GNSS Data Processing" (Vol. 1 and Vol. 2). ESA TM-23.
 *             Authors: J. Sanz Subirana, J.M. Juan Zornoza and M. Hernandez-Pajares
 *             ISBN: 978-92-9221-885-0 (two volumes)
 *             ISSN: 1013-7076
 *             December 2012.
 *
 *             SATSEL debugging message included in the output messages.
 * -----------
 *          gLAB v2.2.0
 * Release: 2014/09/22
 * Change Log: Added cycle slip L1C1 initial value in  message "INFO PREPROCESSING CycleSlip L1C1".
 *	       Added the following INFO message "INFO MODELING Satellite mass centre to Antenna Phase Centre (APC) correction: ON/OFF"
 *	       Added text messages when using BeiDou, NeQuick, F-PPP for ionospheric models, or ESA's tropospheric model or
 *			IONEX, F-PPP DCB values.
 * -----------
 *          gLAB v2.2.1
 * Release: 2014/12/23
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.2
 * Release: 2015/03/02
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.3
 * Release: 2015/05/28
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.4
 * Release: 2015/07/01
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.5
 * Release: 2015/10/20
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.6
 * Release: 2015/11/25
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.7
 * Release: 2016/03/01
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.8
 * Release: 2016/04/12
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v3.0.0
 * Release: 2016/09/05 
 * Change Log: Added SBAS processing.
 *             Added printUserError function for printing user-defined error.
 *             Modifications regarding PRINT INFO (PREPROCESSING) and PRINT CS.
 *             Modified printSatSel function (now prints epoch and constellation)
 * -----------
 *          gLAB v3.1.0
 * Release: 2016/09/26
 * Change Log: SATSEL messages may have an '*' if the model will be recalculated before computing the 
 *                     navigation solution (only in SBAS mode).
 *             Modified "Receiver apriori position" INFO message for parameters 'calculateRINEX' and 'calculateUSER'.
 * -----------
 *          gLAB v3.1.1
 * Release: 2016/10/21
 * Change Log: Added Number of samples to L1C1 and BW cycle-slip info messages.
 *             Fixed a '*' that was in incorrect position in MODEL messages in SBAS mode.
 *             The '[NPA only]' tag from SBASUNSEL message number 36 has been erased.
 * -----------
 *       END_RELEASE_HISTORY
 *****************************/

/* External classes */
#include "output.h"
#include <ctype.h>

//Global variables for printing buffers
int		linesstored[MAX_SATELLITES_IN_VIEW];
char	printbuffer[MAX_SATELLITES_IN_VIEW][MAX_LINES_BUFFERED][MAX_INPUT_LINE];

/**************************************
 * Declarations of internal operations
 **************************************/

/*****************************************************************************
 * Name        : printParameters
 * Description : Print Input parameters
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printParameters (TOptions *options, TEpoch *epoch, TGNSSproducts *products) {
	int		i,j;
	int		aux;
	char	auxstr[300];
	
	if (!options->printInfo) return;
	
	if (epoch->receiver.name[0] == '\0') {
		strcpy(auxstr,"Unknown");
	} else {
		strcpy(auxstr,epoch->receiver.name);
	}

	if(options->SBAScorrections==1 && options->onlySBASiono==0) {
		printf("INFO SBAS processing mode activated\n");
	}

	printf("INFO INPUT Station marker: %s\n",auxstr);
	if (epoch->receiver.antenna.type[0] == '\0') {
		strcpy(auxstr,"Unknown");
	} else {
		strcpy(auxstr,epoch->receiver.antenna.type);
	}
	printf("INFO INPUT Antenna type:   %s\n",auxstr);
	if (epoch->receiver.type[0] == '\0') {
		printf("INFO INPUT Receiver type:  0-Unknown\n");
	} else {
		if (epoch->receiver.recType == rtCROSS) strcpy(auxstr,"1-Cross-correlated");
		else if (epoch->receiver.recType == rtNOP1) strcpy(auxstr,"2-No P1");
		else if (epoch->receiver.recType == rtNOMINAL) strcpy(auxstr,"3-Consistent measurements");
		else strcpy(auxstr,"Unknown");
		printf("INFO INPUT Receiver type:  %s [%s]\n",epoch->receiver.type,auxstr);
	}
	printf("INFO PREPROCESSING Prealign carrier phase measurements: %3s\n",options->prealignCP?"ON":"OFF");
	printf("INFO PREPROCESSING Decimation %d\n",options->decimate);
	aux = 0;
	for (i=0;i<MAX_GNSS;i++) {
		for (j=0;j<MAX_SATELLITES_PER_GNSS;j++) {
			if (options->includeSatellite[i][j]==0) {
				if (!aux) {
					printf("INFO PREPROCESSING Excluding satellites:");
				}
				printf(" %c%02d",gnsstype2char(i),j);
				aux = 1;
			}
		}
	}
	if (aux) printf("\n");
	printf("INFO PREPROCESSING Usable frequencies [GPS]:");
	if (options->usableFreq[GPS][1]) printf(" F%1d",1);
	if (options->usableFreq[GPS][2]) printf(" F%1d",2);
	if (options->usableFreq[GPS][5]) printf(" F%1d",5);
	
	printf("\n");
	printf("INFO PREPROCESSING Elevation mask: %.2f\n",options->elevationMask/d2r);
	printf("INFO PREPROCESSING Discard satellites under eclipse conditions: %3s\n",options->discardEclipsedSatellites?"YES":"NO");
	printf("INFO PREPROCESSING Receiver apriori position: %14.4f %14.4f %14.4f  [",epoch->receiver.aproxPosition[0],epoch->receiver.aproxPosition[1],epoch->receiver.aproxPosition[2]);
	if (options->receiverPositionSource == rpRINEX) printf("RINEX");
	else if (options->receiverPositionSource == rpSINEX) printf("SINEX");
	else if (options->receiverPositionSource == rpCALCULATE) printf("TO_BE_CALCULATED");
	else if (options->receiverPositionSource == rpCALCULATERINEX) printf("TO_BE_CALCULATED, RINEX as initial position");
	else if (options->receiverPositionSource == rpCALCULATEUSER) printf("TO_BE_CALCULATED, User input as initial position");
	else if (options->receiverPositionSource == rpSET) printf("SET");
	printf("]\n");
	printf("INFO PREPROCESSING Cycle Slip Data Gap: %.1f\n",options->csDataGap);
	if (options->csLLI) {
		printf("INFO PREPROCESSING Cycle Slip LLI:   ON\n");
	} else {
		printf("INFO PREPROCESSING Cycle Slip LLI:   OFF\n");
	}
//	if (options->csNcon) {
//		printf("INFO PREPROCESSING Cycle Slip N-con: ON  [Min:%d]\n",-1*options->csNconMin-1);
//	} else {
//		printf("INFO PREPROCESSING Cycle Slip N-con: OFF\n");
//	}
	if (options->csL1C1) {
		printf("INFO PREPROCESSING Cycle Slip L1C1:  ON  [Init:%.3lf  Window:%d  K-factor:%.3lf  Samples:%d]\n",options->csL1C1init,options->csL1C1window,options->csL1C1kfactor,options->csL1C1samples);
	} else {
		printf("INFO PREPROCESSING Cycle Slip L1C1:  OFF\n");
	}
	if (options->csBW) {
		printf("INFO PREPROCESSING Cycle Slip BW:    ON  [SigMin:%.3lf  Window:%.d  K-factor:%.3lf  Samples:%d]\n",options->csBWmin,options->csBWwindow,options->csBWkfactor,options->csBWsamples);
	} else {
		printf("INFO PREPROCESSING Cycle Slip BW:    OFF\n");
	}
	if (options->csLI) {
		printf("INFO PREPROCESSING Cycle Slip LI:    ON  [MaxJump:%.3lf  Max:%.3lf  To:%.3lf  Samples:%d]\n",options->csLImaxjump,options->csLImax,options->csLIt,options->csLIsamples);
	} else {
		printf("INFO PREPROCESSING Cycle Slip LI:    OFF\n");
	}
	if (epoch->receiver.recType == rtNOP1) {
		printf("INFO PREPROCESSING Receiver type does not provide P1... I will use C1 instead of P1\n"); 
	}


	printf("INFO MODELING Satellite clock offset correction: %3s\n",options->satelliteClockCorrection?"ON":"OFF");
	printf("INFO MODELING Consider satellite movement during signal flight time: %3s\n",options->timeTrans?"ON":"OFF");
	printf("INFO MODELING Consider Earth rotation during signal flight time: %3s\n",options->earthRotation?"ON":"OFF");
	for (i=1;i<3;i++)
		printf("INFO MODELING Receiver Antenna Phase Center Offset (PCO): F%1d %8.5f %8.5f %8.5f\n",i,epoch->receiver.antenna.PCO[i][0],epoch->receiver.antenna.PCO[i][1],epoch->receiver.antenna.PCO[i][2]);
	printf("INFO MODELING Satellite mass centre to Antenna Phase Centre (APC) correction: %3s\n",options->satellitePhaseCenter?"ON":"OFF");
	printf("INFO MODELING Receiver Antenna Reference Point (ARP): %7.4f %7.4f %7.4f\n",epoch->receiver.ARP[0],epoch->receiver.ARP[1],epoch->receiver.ARP[2]);
	printf("INFO MODELING Relativistic clock correction: %3s\n",options->relativisticCorrection?"ON":"OFF");
	if (options->ionoModel == NoIono) strcpy(auxstr,"OFF");
	else if (options->ionoModel == Klobuchar) strcpy(auxstr,"Klobuchar");
	else if (options->ionoModel == BeiDou) strcpy(auxstr,"BeiDou");
	else if (options->ionoModel == neQuick) strcpy(auxstr,"NeQuick");
	else if (options->ionoModel == IONEX) strcpy(auxstr,"IONEX");
	else if (options->ionoModel == FPPP) strcpy(auxstr,"FPPP");
	else if (options->ionoModel == SBASiono) strcpy(auxstr,"SBAS");
	printf("INFO MODELING Ionosphere model: %s\n",auxstr);
	printf("INFO MODELING Troposphere model: %3s\n",options->troposphericCorrection?"ON":"OFF");
	if (options->troposphericCorrection  == 1) { 
		if (options->tropNominal == SimpleNominal) printf("INFO MODELING Troposphere model: Simple Nominal\n");
		else if (options->tropNominal == UNB3Nominal) printf("INFO MODELING Troposphere model: UNB-3 Nominal\n");
		else if (options->tropNominal == TropoESA) printf("INFO MODELING Troposphere model: ESA Nominal\n");
		if (options->tropMapping == SimpleMapping) printf("INFO MODELING Troposphere model: Simple Mapping\n");
		else if (options->tropMapping == NiellMapping) printf("INFO MODELING Troposphere model: Niell Mapping\n");
	}
	if (options->p1c1dcbModel == p1c1NONE) strcpy(auxstr,"OFF");
	else if (options->p1c1dcbModel == p1c1FLEXIBLE) strcpy(auxstr,"Flexible");
	else if (options->p1c1dcbModel == p1c1STRICT) strcpy(auxstr,"Strict");
	printf("INFO MODELING P1-C1 DCB model: %s\n",auxstr);
	if (options->p1p2dcbModel == p1p2NONE) strcpy(auxstr,"OFF");
	else if (options->p1p2dcbModel == p1p2RINEX) strcpy(auxstr,"RINEX");
	else if (options->p1p2dcbModel == p1p2DCB) strcpy(auxstr,"DCB");
	else if (options->p1p2dcbModel == p1p2IONEX) strcpy(auxstr,"IONEX");
	else if (options->p1p2dcbModel == p1p2FPPP) strcpy(auxstr,"FPPP");
	printf("INFO MODELING P1-P2 DCB model: %s\n",auxstr);
	printf("INFO MODELING Wind up correction: %3s\n",options->windUpCorrection?"ON":"OFF");
	printf("INFO MODELING Solid tides correction: %3s\n",options->solidTidesCorrection?"ON":"OFF");
	printf("INFO MODELING Relativistic path range correction: %3s\n",options->gravitationalCorrection?"ON":"OFF");
	if (products->type == SP3) {
		printf("INFO MODELING Orbit interpolation degree: %2d\n",options->orbitInterpolationDegree);
		printf("INFO MODELING Clock interpolation degree: %2d\n",options->clockInterpolationDegree);
	}
	printf("INFO MODELING Use satellite 'SV Health' flag of navigation message: %s\n",options->satelliteHealth?"ON":"OFF");
	for (i=0;i<options->totalFilterMeasurements;i++) {
		printf("INFO FILTER Measurement: %1d %3s  StdDev:%.2f",i+1,meastype2measstr(options->measurement[i]),options->fixedMeasurementWeight[i]);
		if (options->elevationMeasurementWeight[i][0]!=0) {
			printf("+%.2f*e^(elev/%.2f)",options->elevationMeasurementWeight[i][0],options->elevationMeasurementWeight[i][1]/d2r);
		}
		if (options->smoothMeas[i]!=NA) 
			printf("  SmoothWith:%3s [n=%d]",meastype2measstr(options->smoothMeas[i]),options->smoothEpochs);
		printf("\n");
	}
	printf("INFO FILTER Number of seconds of continuous code smoothing before steady-state operation: %d\n",options->excludeSmoothingConvergence);
	printf("INFO FILTER Carrierphase is used: %3s\n",options->usePhase?"YES":"NO");
	printf("INFO FILTER Estimate troposphere: %3s\n",options->estimateTroposphere?"ON":"OFF");
	printf("INFO FILTER %s Processing\n",(options->filterIterations==1)?"Forward":"Backward"); 
	for (i=0;i<MAX_PAR;i++) {
		if (i==DR_UNK) strcpy (auxstr,"Position:");
		else if (i==DT_UNK) strcpy (auxstr,"Clock:");
		else if (i==TROP_UNK) strcpy (auxstr,"Troposphere:");
		else if (i==BIAS_UNK) strcpy (auxstr,"Ambiguity:"); 
		printf("INFO FILTER Parameters [Phi,Q,P0] %-14s",auxstr);
		for (j=0;j<3;j++) {
				printf(" %9.2e",options->filterParams[j][i]);
		}
		printf("\n");
	}
	printf("INFO OUTPUT Satellite Velocity:     %s\n",options->satVel==svINERTIAL?"Inertial":"ITRF");
	printf("INFO PRINT INFO:     	    %-3s\n",options->printInfo?"ON":"OFF");
	printf("INFO PRINT CS:     	    %-3s\n",options->printCycleslips?"ON":"OFF");
	printf("INFO PRINT USERADDEDERROR:  %-3s\n",options->printUserError?"ON":"OFF");
	printf("INFO PRINT INPUT:    	    %-3s\n",options->printInput?"ON":"OFF");
	printf("INFO PRINT MEAS :           %-3s\n",options->printMeas?"ON":"OFF");
	printf("INFO PRINT MODEL:   	    %-3s\n",options->printModel?"ON":"OFF");
	printf("INFO PRINT EPOCHSAT:  	    %-3s\n",options->printSatellites?"ON":"OFF");
	printf("INFO PRINT SATSEL:   	    %-3s\n",options->printSatSel?"ON":"OFF");
	printf("INFO PRINT PREFIT:    	    %-3s\n",options->printPrefit?"ON":"OFF");
	printf("INFO PRINT POSTFIT:	    %-3s\n",options->printPostfit?"ON":"OFF");
	printf("INFO PRINT FILTER:   	    %-3s\n",options->printFilterSolution?"ON":"OFF");
	printf("INFO PRINT OUTPUT: 	    %-3s\n",options->printOutput?"ON":"OFF");
	printf("INFO PRINT SATDIFF:	    %-3s\n",options->printSatDiff?"ON":"OFF");
	printf("INFO PRINT SATSTAT:   	    %-3s\n",options->printSatStat?"ON":"OFF");
	printf("INFO PRINT SATSTATTOT:	    %-3s\n",options->printSatStatTot?"ON":"OFF");
	printf("INFO PRINT SBASCORR:  	    %-3s\n",options->printSBASCORR?"ON":"OFF");
	printf("INFO PRINT SBASVAR:   	    %-3s\n",options->printSBASVAR?"ON":"OFF");
	printf("INFO PRINT SBASIONO:	    %-3s\n",options->printSBASIONO?"ON":"OFF");
	printf("INFO PRINT SBASOUT:   	    %-3s\n",options->printSBASOUT?"ON":"OFF");
	printf("INFO PRINT SBASUNSEL: 	    %-3s\n",options->printSBASUNSEL?"ON":"OFF");
	printf("INFO PRINT SBASUNUSED:	    %-3s\n",options->printSBASUNUSED?"ON":"OFF");
}

/*****************************************************************************
 * Name        : printError
 * Description : Print an ERROR message and exit with code -1
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *message                  I  N/A  Message
 * TOptions  *options              I  N/A  TOptions structure 
 *****************************************************************************/
void printError (char *message, TOptions *options) {
	printf("ERROR %s\n",message);
	exit(-1);
}
 
/*****************************************************************************
 * Name        : printInfo
 * Description : Print an INFO message
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *message                  I  N/A  Message
 * TOptions  *options              I  N/A  TOptions structure 
 *****************************************************************************/
void printInfo (char *message, TOptions *options) {
	if (options->printInfo)
		printf("INFO %s\n",message);
}

/*****************************************************************************
 * Name        : printInput
 * Description : Print the input data
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TOptions  *options              I  N/A  TOptions structure 
 *****************************************************************************/
void printInput (TEpoch *epoch, TOptions *options) {
	int		i,j,k;
	double	measurements[10];
	int		res[10];
	
	if (options->printInput) {
		for (i=0;i<epoch->numSatellites;i++) {
			k = satIs(epoch,epoch->sat[i].PRN,epoch->sat[i].GNSS);
			printf ("INPUT    %17s %3s %2d %6d",t2doystr(&epoch->t),gnsstype2gnssstr(epoch->sat[i].GNSS),epoch->sat[i].PRN,epoch->cycleslip.arcLength[k]);
			if (epoch->sat[i].GNSS == GPS) {
				res[0] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C1C,&measurements[0],NULL);
				res[1] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C1P,&measurements[1],NULL);
				res[2] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C2P,&measurements[2],NULL);
				res[3] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L1P,&measurements[3],NULL);
				res[4] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L2P,&measurements[4],NULL);
				for (j=0;j<5;j++) {
					if (measurements[j] ==-1 || res[j]==0) measurements[j] = 0;
					printf (" %14.4f",measurements[j]);
				}
			} else if (epoch->sat[i].GNSS == Galileo) {
				res[0] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C1A,&measurements[0],NULL);
				res[1] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C1B,&measurements[1],NULL);
				res[2] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C1C,&measurements[2],NULL);
				res[3] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C7Q,&measurements[3],NULL);
				res[4] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C8Q,&measurements[4],NULL);
				res[5] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L1A,&measurements[5],NULL);
				res[6] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L1B,&measurements[6],NULL);
				res[7] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L1C,&measurements[7],NULL);
				res[8] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L7Q,&measurements[8],NULL);
				res[9] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L8Q,&measurements[9],NULL);
				for (j=0;j<10;j++) {
					if (measurements[j] ==-1 || res[j]==0) measurements[j] = 0;
					printf (" %14.4f",measurements[j]);
				}
			} else if (epoch->sat[i].GNSS == GLONASS) {
				res[0] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C1C,&measurements[0],NULL);
				res[1] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C2C,&measurements[1],NULL);
				res[2] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L1P,&measurements[2],NULL);
				res[3] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L2P,&measurements[3],NULL);
				for (j=0;j<4;j++) {
					if (measurements[j] ==-1 || res[j]==0) measurements[j] = 0;
					printf (" %14.4f",measurements[j]);
				}
			} else if (epoch->sat[i].GNSS == GEO) {
				res[0] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C1C,&measurements[0],NULL);
				for (j=0;j<1;j++) {
					if (measurements[j] ==-1 || res[j]==0) measurements[j] = 0;
					printf (" %14.4f",measurements[j]);
				}
			}
			printf("\n");
		}
	}
}

/*****************************************************************************
 * Name        : printMeas
 * Description : Print information on the measurements of the input and azimuth
 *               and elevation of the satellite
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TOptions  *options              I  N/A  TOptions structure 
 *****************************************************************************/
void printMeas (TEpoch *epoch, TOptions *options) {
	int						i,j,k;
	int						n[MAX_GNSS];
	int						maxN = 0;
	char					measStr[MAX_GNSS][MAX_MEASUREMENTS_PER_SATELLITE*2*3+1];
	int						len;
	enum MeasurementType	measChain[MAX_GNSS][MAX_MEASUREMENTS_PER_SATELLITE*2];
	double					measVal;
	char					mS[2][4];
	char					format[20];
	int						res;
	
	if (options->printMeas) {
		for (i=0;i<MAX_GNSS;i++) {
			n[i] = 0;
			measStr[i][0] = '\0';
			len = 0;
			for (j=1;j<ENDMEAS;j+=4) {
				// j=1 is a pseudorange, and then pseudoranges come every 4 identifiers.
				// See the "enum MeasurementType" enumerator at dataHandling.h file for
				// more details 
				if (epoch->measOrder[i].meas2Ind[j]!=-1 || epoch->measOrder[i].meas2Ind[j+1]!=-1) {
					// See if the Pseudorange OR the associated Carrier Phase (j+1) are available
					measChain[i][n[i]] = j;
					measChain[i][n[i]+1] = j+1;
					n[i]+=2;
					strcpy(mS[0], meastype2measstr(j));
					strcpy(mS[1], meastype2measstr(j+1));
					if (len!=0) {
						len += sprintf(&measStr[i][len],":");
					}
					len += sprintf(&measStr[i][len],"%s:%s",mS[0],mS[1]);
				}
			}
			if (n[i]>maxN) maxN = n[i];
		}
		for (i=0;i<epoch->numSatellites;i++) {
			k = satIs(epoch,epoch->sat[i].PRN,epoch->sat[i].GNSS);
			if(epoch->cycleslip.consistency[k]) continue;
			printf ("MEAS     %17s %3s %2d %8.3f %8.3f %2d ",t2doystr(&epoch->t),gnsstype2gnssstr(epoch->sat[i].GNSS),epoch->sat[i].PRN,epoch->sat[i].elevation/d2r,epoch->sat[i].azimuth/d2r,n[epoch->sat[i].GNSS]);
			sprintf (format,"%%%ds",maxN*4);
			printf(format,measStr[epoch->sat[i].GNSS]);
			for (j=0;j<n[epoch->sat[i].GNSS];j++) {
				res = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,measChain[epoch->sat[i].GNSS][j],&measVal,NULL);
				if (measVal==-1 || res==0) measVal = 0;
				printf(" %14.4f",measVal);
			}
			printf("\n");
		}
	}
}

/*****************************************************************************
 * Name        : printCS
 * Description : Prints the cycle-slip information when one is found
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  Reference epoch
 * int  satIndex                   I  N/A  Satellite index
 * int  LiCheck                    I  N/A  Li Flag (1-> cycle-slip found in Li)
 * double  LiLiestDiff             I    m  Absolute difference between measured 
 *                                         and estimated Li values
 * double  LiThreshold             I    m  Threshold for LiLiestDiff
 * int  BwCheck                    I  N/A  Bw Flag (1-> cycle-slip found in Bw)
 * double  nBwdiff                 I  cyc  Absolute difference between measured 
 *                                         and estimated Bw values
 * double  nBwdiffThreshold        I  cyc  Threshold for nBwdiff
 * int  L1C1Check                  I  N/A  L1C1 Flag (1->cycle-slip found in L1C1)
 * double  L1C1diff                I    m  Absolute difference between measured 
 *                                         and estimated L1C1 values
 * double  L1C1Threshold           I    m  Threshold for L1C1diff
 * TOptions  *options              I  N/A  TOptions structure
 * int  DataGapCheck               I  N/A  DataGap Flag (1-> cycle-slip found in DataGap)
 * double maxDataGap               I    s  Maximum period of time allowed before declaring CS
 * double timeDiff                 I    s  Delta time from previous usable epoch
 * int  LLICheck                   I  N/A  LLI Flag (1-> cycle-slip found in LLI)
 * int  mode                       I  N/A  Mode of processing
 *                                         0 => Process rover
 *                                         1 => Process Base Station (for DGNSS)
 *****************************************************************************/
void printCS (TEpoch *epoch, int satIndex, int LiCheck, double LiLiestDiff, double LiThreshold, int BwCheck, double nBwdiff, double nBwdiffThreshold, int L1C1Check, double L1C1diff, double L1C1Threshold, TOptions *options, int DataGapCheck, double maxDataGap, double timeDiff, int LLICheck, int mode) {
	if (options->printCycleslips) {
		char cs[6];
		if ( mode != 0 ) strcpy(cs,"CS_BS");
		else strcpy(cs,"CS");
		if ( DataGapCheck ) {
			 printf("%-9s%17s %3s %2d  DATA_GAP = %lf THRESHOLD = %lf\n",cs,t2doystr(&epoch->t),gnsstype2gnssstr(epoch->sat[satIndex].GNSS),epoch->sat[satIndex].PRN,timeDiff,maxDataGap);
		} else if ( LLICheck ) {
			 printf("%-9s%17s %3s %2d  LLI\n",cs,t2doystr(&epoch->t),gnsstype2gnssstr(epoch->sat[satIndex].GNSS),epoch->sat[satIndex].PRN);
		} else {
			if ( options->csLI && LiCheck ) {
				printf("%-9s%17s %3s %2d  LI = %lf THRESHOLD = %lf\n",cs,t2doystr(&epoch->t),gnsstype2gnssstr(epoch->sat[satIndex].GNSS),epoch->sat[satIndex].PRN,LiLiestDiff,LiThreshold);
			}
			if ( options->csBW && BwCheck ) {
				printf("%-9s%17s %3s %2d  Bw = %lf THRESHOLD = %lf\n",cs,t2doystr(&epoch->t),gnsstype2gnssstr(epoch->sat[satIndex].GNSS),epoch->sat[satIndex].PRN,nBwdiff,nBwdiffThreshold);
			}
			if ( options->csL1C1 && L1C1Check ) {
				printf("%-9s%17s %3s %2d  L1C1 = %lf THRESHOLD = %lf\n",cs,t2doystr(&epoch->t),gnsstype2gnssstr(epoch->sat[satIndex].GNSS),epoch->sat[satIndex].PRN,L1C1diff,L1C1Threshold);
			}
		}
	}
}
/*****************************************************************************
 * Name        : printModel
 * Description : Prints each of the components of the model for a specific 
 *               satellite and measurement
 * Parameters  :
 * Name                                    |Da|Unit|Description
 * TEpoch  *epoch                           I  N/A  Reference epoch
 * int  satIndex                            I  N/A  Selected satellite
 * int  measIndex                           I  N/A  Selected measurement
 * double  flightTime                       I    s  Signal flight time
 * double  geometricDistance                I    m  Geometric distance between satellite and receiver
 * double  satelliteClockModel              I    m  Satellite clock error
 * double  satellitePhaseCenterProjection   I    m  Projection of the satellite phase center into the LoS
 * double  receiverPhaseCenterProjection    I    m  Projection of the receiver phase center into the LoS 
 * double  receiverARPProjection            I    m  Projection of the receiver ARP into the LoS
 * double  relativityModel                  I    m  Relativity model
 * double  windUpMetModel                   I    m  Wind up model (for carrier phase only)
 * double  troposphereModel                 I    m  Troposphere model
 * double  ionosphereModel                  I    m  Ionosphere model
 * double  gravitationalDelayModel          I    m  Gravitational delay model
 * double  TGDModel                         I    m  TGD model
 * double  solidTidesModel                  I    m  Projection of the effect of solid tides into the LoS
 * TOptions  *options                       I  N/A  TOptions structure
 *****************************************************************************/
void printModel (TEpoch *epoch, int satIndex, int measIndex, double flightTime, double geometricDistance, double satelliteClockModel, double satellitePhaseCenterProjection, double receiverPhaseCenterProjection, double receiverARPProjection, double relativityModel, double windUpMetModel, double troposphereModel, double ionosphereModel, double gravitationalDelayModel, double TGDModel, double solidTidesModel, TOptions *options) {
	double 	velocity[3];
	double	measurement;
	double	model;
	int		res,i;
	
	i = satIs(epoch,epoch->sat[satIndex].PRN,epoch->sat[satIndex].GNSS);

	if (options->satVel==svINERTIAL) {
		memcpy(velocity,epoch->sat[satIndex].velocity,3*sizeof(double));
	} else {
		memcpy(velocity,epoch->sat[satIndex].ITRFvel,3*sizeof(double));
	}	 
	if (options->printModel) {
		res = getMeasModelValue(epoch,epoch->sat[satIndex].GNSS,epoch->sat[satIndex].PRN,epoch->measOrder[epoch->sat[satIndex].GNSS].ind2Meas[measIndex],&measurement,&model);
		if (res==1) {
			sprintf(printbuffer[satIndex][linesstored[satIndex]],"MODEL    %17s %3s %2d %3s %11.5f %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %14.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %8.3f %8.3f\n",
					t2doystr(&epoch->t),
					gnsstype2gnssstr(epoch->sat[satIndex].GNSS),
					epoch->sat[satIndex].PRN,
					meastype2measstr(epoch->measOrder[epoch->sat[satIndex].GNSS].ind2Meas[measIndex]),
					flightTime,
					getMeasurementValue(epoch,epoch->sat[satIndex].GNSS,epoch->sat[satIndex].PRN,epoch->measOrder[epoch->sat[satIndex].GNSS].ind2Meas[measIndex]),
					getModelValue(epoch,epoch->sat[satIndex].GNSS,epoch->sat[satIndex].PRN,epoch->measOrder[epoch->sat[satIndex].GNSS].ind2Meas[measIndex]),
					epoch->sat[satIndex].position[0],
					epoch->sat[satIndex].position[1],
					epoch->sat[satIndex].position[2],
					velocity[0],
					velocity[1],
					velocity[2],
										//JAUME//
					geometricDistance,
										//GEOMETRIC RANGE BETWEEN RECEIVER AND ANTENNA PHASE CENTERS//
										//geometricDistance+satellitePhaseCenterProjection+receiverPhaseCenterProjection+receiverARPProjection+solidTidesModel,//
										//JAUME//
					satelliteClockModel,
					satellitePhaseCenterProjection,
					receiverPhaseCenterProjection,
					receiverARPProjection,
					relativityModel,
					//windUpRadModel,
					windUpMetModel,
					troposphereModel,
					ionosphereModel,
					gravitationalDelayModel,
					TGDModel,
					solidTidesModel,
										//JAUME
										epoch->sat[satIndex].elevation/d2r,
										epoch->sat[satIndex].azimuth/d2r);
										//JAUME

			//Check if satellite has reached steady-state operation. If not, put an '*' to indicate that it will not be used
			if(options->excludeSmoothingConvergence>0 && epoch->cycleslip.arcLength[i] < options->excludeSmoothingConvergence) {
				//Steady state not reached
				printbuffer[satIndex][linesstored[satIndex]][5]='*';
			}

			if(options->printInBuffer==1) {
				linesstored[satIndex]++;
			} else {
				printf("%s",printbuffer[satIndex][linesstored[satIndex]]);
				printbuffer[satIndex][linesstored[satIndex]][0]='\0';
			}
		}
	}
}

/*****************************************************************************
 * Name        : printSatSel
 * Description : Prints debug information of the reason why a satellite has 
 *               been discarded (or selected) for processing 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                           I  N/A  Reference epoch
 * int  selected                   I  N/A  Flag to set if the satellite was 
 *                                         selected (1) or discarded (0) 
 * int  satellite                  I  N/A  PRN of the satellite
 * char  *message                  I  N/A  Reason of the discard (if selected==0)
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printSatSel (TEpoch *epoch, int selected, char *message, int satIndex, TOptions *options) {
        if (options->printSatSel) {

                if (!selected) sprintf(printbuffer[satIndex][linesstored[satIndex]],"SATSEL  %17s %3s %2d discarded: %s\n",t2doystr(&epoch->t),gnsstype2gnssstr(epoch->sat[satIndex].GNSS),epoch->sat[satIndex].PRN,message);
                else sprintf(printbuffer[satIndex][linesstored[satIndex]],"SATSEL  %17s %3s %2d selected\n",t2doystr(&epoch->t),gnsstype2gnssstr(epoch->sat[satIndex].GNSS),epoch->sat[satIndex].PRN);

                if(options->printInBuffer==1) {
                                        linesstored[satIndex]++;
                        } else {
                        printf("%s",printbuffer[satIndex][linesstored[satIndex]]);
                        printbuffer[satIndex][linesstored[satIndex]][0]='\0';
                }
        }
}

/*****************************************************************************
 * Name        : printPrefit
 * Description : Prints the prefilter residuals for a specific satellite/measurement
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * TSatellite  *sat                I  N/A  Selected satellite
 * enum MeasurementType  measType  I  N/A  Selected measurement
 * double  prefit                  I    m  Computed prefit
 * double  measurement             I    m  Measurement value
 * double  model                   I    m  Modeled value
 * double  stdDev                  I    m  Assigned weight of the measurement 
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printPrefit (TTime *t, TSatellite *sat, enum MeasurementType measType, double prefit, double measurement, double model, double stdDev, TOptions *options) {	
	if (options->printPrefit) {
		printf("PREFIT%1s  %17s %3s %2d %3s %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %8.3f %8.3f",
				sat->available?"":"*",
				t2doystr(t),
				gnsstype2gnssstr(sat->GNSS),
				sat->PRN,
				meastype2measstr(measType),
				prefit,
				measurement,
				model,
				-1.0*sat->LoS[0],
				-1.0*sat->LoS[1],
				-1.0*sat->LoS[2],
				sqrt(sat->LoS[0]*sat->LoS[0] + sat->LoS[1]*sat->LoS[1] + sat->LoS[2]*sat->LoS[2]),				
				sat->elevation/d2r,
				sat->azimuth/d2r
				);


		if (stdDev==SIGMA_INF) {
			printf(" 9999.999\n");
		} else {
			printf(" %8.3f\n", stdDev);
		}
	}
}

/*****************************************************************************
 * Name        : printEpochsat
 * Description : Prints the satellites used for processing an epoch
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * enum MeasurementType  measType  I  N/A  Selected measurement
 * int  ns                         I  N/A  Number of satellites used
 * int  *sats                      I  N/A  Vector containing the satellites PRN
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printEpochsat (TTime *t, enum MeasurementType measType, int ns, int *sats, TOptions *options) {
	int i;
	
	if (options->printSatellites) {
		printf("EPOCHSAT %17s %3s %2d",t2doystr(t),meastype2measstr(measType),ns);
		for (i=0;i<ns;i++) {
			printf (" %2d",sats[i]);
		}
		printf("\n");
	}
}

/*****************************************************************************
 * Name        : printFilter
 * Description : Prints the parameters estimated from the filter
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * int  nunk                       I  N/A  Number of unknowns in the filter
 * TFilterSolution  *solution      I  N/A  Structure storing the estimated parameters
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printFilter (TTime *t, int nunk, TFilterSolution *solution, TOptions *options) {
	int i;

	if (options->printFilterSolution) {
		printf("FILTER   %17s",t2doystr(t));
		for (i=0;i<nunk;i++) {
			printf(" %14.4f",solution->x[i]);
		}
		printf("\n");
	}
}

/*****************************************************************************
 * Name        : printPostfit
 * Description : Prints the postfilter residuals for a specific satellite/measurement
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * TSatellite  *sat                I  N/A  Selected satellite
 * enum MeasurementType  measType  I  N/A  Selected measurement
 * double  postfit                 I  N/A  Computed postfit
 * double  measurement             I  N/A  Measurement value
 * double  model                   I  N/A  Modeled value
 * double  estimatedParameters     I  N/A  Estimated parameters from the filter
 * double  ambiguity               I  N/A  Estimated ambiguity value (only for carrier phases)
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printPostfit (TTime *t, TSatellite *sat, enum MeasurementType measType, double postfit, double measurement, double model, double estimatedParameters, double ambiguity, TOptions *options) {
	if (options->printPostfit) {
		if (whatIs(measType)==CarrierPhase) {
			printf("POSTFIT%1s %17s %3s %2d %3s %14.4f %14.4f %14.4f %8.3f %8.3f %14.4f\n",
					sat->available?"":"*",
					t2doystr(t),
					gnsstype2gnssstr(sat->GNSS),
					sat->PRN,
					meastype2measstr(measType),
					postfit,
					measurement,
					model + estimatedParameters,
					sat->elevation/d2r,
					sat->azimuth/d2r,
					ambiguity);
		} else if (whatIs(measType)==Pseudorange) {
			printf("POSTFIT%1s %17s %3s %2d %3s %14.4f %14.4f %14.4f %8.3f %8.3f\n",
					sat->available?"":"*",
					t2doystr(t),
					gnsstype2gnssstr(sat->GNSS),
					sat->PRN,
					meastype2measstr(measType),
					postfit,
					measurement,
					model + estimatedParameters,
					sat->elevation/d2r,
					sat->azimuth/d2r);
		}
	}
}
 
/*****************************************************************************
 * Name        : printOutput
 * Description : Prints the XYZ and NEU solutions for a station obtained by the
 *               filter, along with the dXYZ, dNEU in relation of the apriori
 *               position (if given), and formal errors of both
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TFilterSolution  *solution      I  N/A  Solution of the filter
 * TOptions  *options              I  N/A  TOptions structure
 * TUnkinfo  *unkinfo              I  N/A  Information on the unknowns of the filter
 *****************************************************************************/
void printOutput (TEpoch *epoch, TFilterSolution *solution, TOptions *options, TUnkinfo *unkinfo) {
	double	positionNEU[3];
	double	dNEU[3];
	double  dXYZ[3];
	double  feNEU[3];
	int		i,j,k;
	int 	indx;
	int		indtrop;
	
	// First index of XYZ
	indx = unkinfo->par2unk[DR_UNK];
	
	// Index of troposphere
	if (options->estimateTroposphere) {
		indtrop = unkinfo->par2unk[TROP_UNK];
	}
	
	// Obtaing NEU coordinates
	XYZ2NEU(&solution->x[indx],positionNEU);
	
	// Compute:
	//		dNEU differences in relation to apriori position
	//		Formal errors in NEU system. For this:
	//					feNEU = R^-1*feXYZ*R
	//					Being, R the rotation matrix (orientation matrix)
	//						   feXYZ the formal error matrix in XYZ components (the correlation matrix)
	//					As we are only interested in the diagonal of feNEU, only these elements are computed
	for (i=0;i<3;i++) {
		dNEU[i] = 0;
		feNEU[i] = 0;
		for (j=0;j<3;j++) {
			dNEU[i] += (solution->x[indx+j] - epoch->receiver.aproxPosition[j])*epoch->receiver.orientation[i][j];
			for (k=0;k<3;k++) {
				feNEU[i] += epoch->receiver.orientation[j][i] * solution->correlations[m2v(indx+j,indx+k)] * epoch->receiver.orientation[k][i];
			}
		}
		feNEU[i] = sqrt(feNEU[i]);
		dXYZ[i] = solution->x[indx+i] - epoch->receiver.aproxPosition[i];
	}
	
	if (options->receiverPositionSource >= rpCALCULATE) {
		for (i=0;i<3;i++) {
			dNEU[i] = 0;
			dXYZ[i] = 0;
		}
	}
	
	if (options->printOutput) {
		printf("OUTPUT %17s %9.4f %14.4f %14.4f %14.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %14.9f %14.9f %14.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %9.4f %9.4f %9.4f %9.4f %9.4f",
				t2doystr(&epoch->t),
				solution->dop,
				solution->x[indx+0],
				solution->x[indx+1],
				solution->x[indx+2],
				dXYZ[0],
				dXYZ[1],
				dXYZ[2],
				sqrt(solution->correlations[m2v(indx+0,indx+0)]),
				sqrt(solution->correlations[m2v(indx+1,indx+1)]),
				sqrt(solution->correlations[m2v(indx+2,indx+2)]),
				positionNEU[0]/d2r, // Latitude
				positionNEU[1]/d2r, // Longitude
				positionNEU[2],     // Height
				dNEU[0],			// dNorth
				dNEU[1],			// dEast
				dNEU[2],			// dUp
				feNEU[0],
				feNEU[1],
				feNEU[2],
				solution->GDOP,
				solution->PDOP,
				solution->TDOP,				
				solution->HDOP,
				solution->VDOP
				);
		if (options->estimateTroposphere) {
			printf(" %10.4f %10.4f %10.4f\n",
					solution->x[indtrop],
					solution->x[indtrop] - epoch->ZTD,
					sqrt(solution->correlations[m2v(indtrop,indtrop)])
					);
		}else {
			printf("\n");
		}
	}
}

/*****************************************************************************
 * Name        : printSatDiff
 * Description : Prints the difference between two different satellites at a 
 *               specific time. Provides the difference in SISRE, SISRE 
 *               orbit-only, 3D orbit, clock and radial, along and cross-track
 *               components
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * enum GNSSystem  GNSS            I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN satellite identifier
 * double  SISREDiff               I    m  Signal-in-Space Range Error difference
 * double  OrbDiff                 I    m  SISRE orbit-only difference
 * double  Orb3DDiff               I    m  3D orbit difference
 * double  ClkDiff                 I    m  Clock difference
 * double  posDiff[3]              I    m  Radial, along- and cross-track position difference 
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printSatDiff (TTime *t, enum GNSSystem GNSS, int PRN, double SISREDiff, double OrbDiff, double Orb3DDiff, double ClkDiff, double posDiff[3], TOptions *options) {
	if (options->printSatDiff) {
		printf("SATDIFF %17s %3s %2d %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f\n",
				t2doystr(t),
				gnsstype2gnssstr(GNSS),
				PRN,
				SISREDiff,
				OrbDiff,
				Orb3DDiff,
				ClkDiff,
				posDiff[2],
				posDiff[0],
				posDiff[1]
				);
	}
}

/*****************************************************************************
 * Name        : printSatStat
 * Description : Prints the statistics of differences between two different 
 *               satellites. Provides the difference in SISRE, SISRE orbit-only, 
 *               3D orbit, clock and radial, along and cross-track components
 * Parameters  :
 * Name                           |Da|Unit|Description
 * enum GNSSystem  GNSS            I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN satellite identifier
 * TStat  statSISRE                I  N/A  Signal-in-Space Range Error statistics
 * TStat  statOrb                  I  N/A  SISRE orbit-only statistics
 * TStat  stat3D                   I  N/A  3D orbit statistics
 * TStat  statClk                  I  N/A  Clock statistics
 * TStat  statRadOrb               I  N/A  Radial position statistics 
 * TStat  statAloOrb               I  N/A  Along-track position statistics 
 * TStat  statCroOrb               I  N/A  Cross-track position statistics 
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printSatStat (enum GNSSystem GNSS, int PRN, TStat statSISRE, TStat statOrb, TStat stat3D, TStat statClk, TStat statRadOrb, TStat statAloOrb, TStat statCroOrb, TOptions *options) {
	if (options->printSatStat && (statSISRE.n != 0)) {
		printf("SATSTAT %3s %2d %4d %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f\n",
				gnsstype2gnssstr(GNSS),
				PRN,
				statSISRE.n,
				statSISRE.mean,
				sqrt(statSISRE.mean2),
				sqrt(statSISRE.mean2 - statSISRE.mean*statSISRE.mean),
				statOrb.mean,
				sqrt(statOrb.mean2),
				sqrt(statOrb.mean2 - statOrb.mean*statOrb.mean),
				stat3D.mean,
				sqrt(stat3D.mean2),
				sqrt(stat3D.mean2 - stat3D.mean*stat3D.mean),
				statClk.mean,
				sqrt(statClk.mean2),
				sqrt(statClk.mean2 - statClk.mean*statClk.mean),
				statRadOrb.mean,
				sqrt(statRadOrb.mean2),
				sqrt(statRadOrb.mean2 - statRadOrb.mean*statRadOrb.mean),
				statAloOrb.mean,
				sqrt(statAloOrb.mean2),
				sqrt(statAloOrb.mean2 - statAloOrb.mean*statAloOrb.mean),
				statCroOrb.mean,
				sqrt(statCroOrb.mean2),
				sqrt(statCroOrb.mean2 - statCroOrb.mean*statCroOrb.mean)
				); 
	}
}

/*****************************************************************************
 * Name        : printSatStatTot
 * Description : Prints the statistics of differences between two products. 
 *               Provides the difference in SISRE, SISRE orbit-only, 3D orbit, 
 *               clock and radial, along and cross-track components
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TStat  statSISRE                I  N/A  Signal-in-Space Range Error statistics
 * TStat  statOrb                  I  N/A  SISRE orbit-only statistics
 * TStat  stat3D                   I  N/A  3D orbit statistics
 * TStat  statClk                  I  N/A  Clock statistics
 * TStat  statRadOrb               I  N/A  Radial position statistics 
 * TStat  statAloOrb               I  N/A  Along-track position statistics 
 * TStat  statCroOrb               I  N/A  Cross-track position statistics 
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printSatStatTot (TStat statSISRETot, TStat statOrbTot, TStat stat3DTot, TStat statClkTot, TStat statRadOrbTot, TStat statAloOrbTot, TStat statCroOrbTot, TOptions *options) {
	if (options->printSatStatTot && statSISRETot.n != 0) {
		printf("SATSTATTOT %2d %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f\n",
				statSISRETot.n,
				statSISRETot.mean,
				sqrt(statSISRETot.mean2),
				sqrt(statSISRETot.mean2 - statSISRETot.mean*statSISRETot.mean),
				statOrbTot.mean,
				sqrt(statOrbTot.mean2),
				sqrt(statOrbTot.mean2 - statOrbTot.mean*statOrbTot.mean),
				stat3DTot.mean,
				sqrt(stat3DTot.mean2),
				sqrt(stat3DTot.mean2 - stat3DTot.mean*stat3DTot.mean),
				statClkTot.mean,
				sqrt(statClkTot.mean2),
				sqrt(statClkTot.mean2 - statClkTot.mean*statClkTot.mean),
				statRadOrbTot.mean,
				sqrt(statRadOrbTot.mean2),
				sqrt(statRadOrbTot.mean2 - statRadOrbTot.mean*statRadOrbTot.mean),
				statAloOrbTot.mean,
				sqrt(statAloOrbTot.mean2),
				sqrt(statAloOrbTot.mean2 - statAloOrbTot.mean*statAloOrbTot.mean),
				statCroOrbTot.mean,
				sqrt(statCroOrbTot.mean2),
				sqrt(statCroOrbTot.mean2 - statCroOrbTot.mean*statCroOrbTot.mean)
				);
	}
}

/*****************************************************************************
 * Name        : printSatPvt
 * Description : Prints the statistics of differences between two products. 
 *               Provides the difference in SISRE, SISRE orbit-only, 3D orbit, 
 *               clock and radial, along and cross-track components
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * enum GNSSystem  GNSS            I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN satellite identifier
 * double  position[3]             I    m  XYZ position of the satellite
 * double  velocity[3]             I  m/s  VXVYVZ velocity of the satellite
 * double  clock                   I    m  Clock offset of the satellite
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printSatPvt (TTime *t, enum GNSSystem GNSS, int PRN, double position[3], double velocity[3], double clock, TOptions *options) {
	if (options->printSatPvt) {
		printf("SATPVT %17s %3s %2d %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %17.6f\n",
				t2doystr(t),
				gnsstype2gnssstr(GNSS),
				PRN,
				position[0],
				position[1],
				position[2],
				velocity[0],
				velocity[1],
				velocity[2],
				clock
				);
	}
}

/*****************************************************************************
 * Name        : writeSBASfiles
 * Description : Checks which SBAS data files have to be written and
 *               calls the function to write the current line to them
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  **fdlist                  I  N/A  Vector of file descriptors for writing data
 * FILE  *readfd		   I  N/A  File descriptor of the data that is being read
 * char **filelist		   I  N/A  List of filenames for writing data
 * char **line                     I  N/A  Data lines to be written
 * int numlines                    I  N/A  Number of lines given in string "line"
 * int header                      I  N/A  Indicates if the current lines
 *                                         are from a header
 *                                         0 => Not a header line
 *                                         1 => Header line
 * int sourcefile                  I  N/A  Indicates from what type file the
 *                                         SBAS data has been read.
 *                                         0 => RINEXB
 *                                         1 => EMS
 * int decoderesult		   I  N/A  Result of the message decodification
 * int decodedmessagetype          I  N/A  Message type decoded from hexadecimal message 					   
 * int *prevday			   O  N/A  Indicates if a RINEX H file from the
 *                                         previous day has been written
 * TSBASblock  *sbasblock          I  N/A  TSBASblock struct
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void writeSBASfiles (FILE  **fdlist, FILE  *readfd, char **filelist, char **line, int numlines, int header, int sourcefile, int decoderesult, int decodedmessagetype, int *prevday, TSBASblock  *sbasblock, TOptions  *options) {

	if(decoderesult<0) {
		//An incorrect message has been detected. It must no be written in RINEX B files or EMS files
		//For pegasus, the data will not be written but it will be logged in the "SBAS Message Transmission Sequence File"
		//Furthermore, a log will be created with the incorrect messages
					
		if(options->writePegasusfile==1) {
			writePegasusfiles(fdlist,readfd,filelist,line,numlines,sourcefile,header,decoderesult,decodedmessagetype,prevday,sbasblock,options);
		}

		writeLogfile(fdlist,readfd,filelist[fdLOGFILE],line,numlines,sourcefile,decoderesult,sbasblock,options);			
	} else {
		//Check if we have to write a new Rinex B files (taking out the messages with failed CRC)
		if(options->writeRinexBfile==1) {
			writeRinexBfile(fdlist,readfd,filelist[fdRINEXB],line,numlines,header,sourcefile,sbasblock,options);
		}

		//Check if we have to write a EMS file (taking out the messages with failed CRC)
		if(options->writeEMSfile==1) {
			writeEMSfile(fdlist,readfd,filelist[fdEMS],line,numlines,header,sourcefile,sbasblock,options);

		}

		//Check if we have to write a Pegasus file (taking out the messages with failed CRC)
		if(options->writePegasusfile==1) {
			writePegasusfiles(fdlist,readfd,filelist,line,numlines,sourcefile,header,decoderesult,decodedmessagetype,prevday,sbasblock,options);
		}
	}
}



/*****************************************************************************
 * Name        : writeLogfile
 * Description : Writes corrupted messages to log file in EMS format along with
 *		 the detected error at the begining of each message
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  **fdlist                  I  N/A  Vector of file descriptors for writing data
 * FILE  *readfd		   I  N/A  File descriptor of the data that is being read
 * char *filename                  I  N/A  Filename where the data is saved
 * char **line                     I  N/A  Data lines to be written
 * int numlines                    I  N/A  Number of lines given in string "line"
 * int sourcefile                  I  N/A  Indicates from what type file the
 *                                         SBAS data has been read.
 *                                         0 => RINEXB
 *                                         1 => EMS
 * int decoderesult                I  N/A  Indicates the result of message decoding
 * TSBASblock  *sbasblock        I  N/A  TSBASblock struct                                      
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void writeLogfile (FILE **fdlist, FILE *readfd, char *filename, char **line, int numlines, int sourcefile, int decoderesult, TSBASblock  *sbasblock, TOptions  *options) {
	int			i,j;
	int			ret;
	int			numberofcolumns;
	char		aux[200],hexstring[300],error[50];
	struct tm 	tm;
	int			year;
	double		doy,seconds;

	switch(decoderesult) {
		case(-1):
			strcpy(error,"CRC24Q_FAILURE              ");
			break;

		case(-2):
			strcpy(error,"INVALID_PREAMBLE            ");
			break;

		case(-3):
			strcpy(error,"MESSAGE_IDENTIFIER_MISMATCH ");
			break;

		case(-4):
			strcpy(error,"SATNUMBER_CORRECTED_EXCEEDED");
			break;

		case(-5):
			strcpy(error,"INVALID_TIME_OF_DAY         ");
			break;

		case(-6):
			strcpy(error,"INVALID_TIME_OF_WEEK        ");
			break;

		case(-7):
			strcpy(error,"INVALID_PRN_MASK_NUMBER     ");
			break;

		case(-8):
			strcpy(error,"UNKNOWN_MESSAGE_TYPE        ");
			break;

	}

	if(sourcefile==0) {
		//The origin file is a RINEX B, but time is saved in time of applicability, the same as EMS file
		t2tmnolocal(&sbasblock->t,&tm,&seconds);
		t2doy(&sbasblock->t,&year,&doy);

		//Save all the hexademical values in a single string
		for(i=1;i<numlines;i++) {
			//Compute the number of records (columns) in the current line (normally the last one is not full)
			if(i==numlines-1) {
				numberofcolumns=sbasblock->datalength%18;
			} else {
				numberofcolumns=18;
			}

			//Read lines
			for(j=0;j<numberofcolumns;j++) {
				getstr(&hexstring[18*2*(i-1)+2*j],line[i],7+3*j,2);
			}
		}
		ret=fprintf(fdlist[fdLOGFILE],"%s %03d %02d %02d %02d %02d %02d %02d %d %s\n",error,sbasblock->PRN,year,tm.tm_mon+1,tm.tm_mday,tm.tm_hour,tm.tm_min,tm.tm_sec,sbasblock->messagetype,hexstring);
		if(ret<0) {
			fclose(readfd);		//Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);  //Close all other files before exiting
			sprintf(aux,"When writing SBAS data to EMS file [%s]\n",filename);
			printError(aux,options);
		}
	} else {
		//The origin file is a EMS file
		ret=fprintf(fdlist[fdLOGFILE],"%s %s",error,line[0]);
		if(ret<0) {
			fclose(readfd);		//Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);  //Close all other files before exiting
			sprintf(aux,"When writing SBAS data to EMS file [%s]\n",filename);
			printError(aux,options);
		}
	}
}

/*****************************************************************************
 * Name        : writeRinexBfile
 * Description : Writes a Rinex B file from SBAS data. This file will not
 * 		 contain any corrupted message
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  **fdlist                  I  N/A  Vector of file descriptors for writing data
 * FILE  *readfd		   I  N/A  File descriptor of the data that is being read
 * char *filename 	           I  N/A  Filename where the data is saved
 * char **line 		           I  N/A  Data lines to be written
 * int numlines                    I  N/A  Number of lines given in string "line"
 * int header                      I  N/A  Indicates if the current lines
 * 					   are from a RINEXB header
 * 					   0 => Not a header line
 * 					   1 => Header line
 * int sourcefile                  I  N/A  Indicates from what type file the
 * 					   SBAS data has been read.
 * 					   0 => RINEXB
 * 					   1 => EMS
 * TSBASblock  *sbasblock          I  N/A  TSBASblock struct 					   
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void writeRinexBfile (FILE **fdlist, FILE *readfd, char *filename, char **line, int numlines, int header, int sourcefile, TSBASblock  *sbasblock, TOptions  *options) {
	int 		i;
	int 		ret;
	int 		len;
	char 		aux[200];
	char 		aux2[200];
	char 		aux3[10];
	char		monthname[10];
	static int 	headerwritten=0;
	struct 		tm tm;
	TTime 		t;
	int 		year;
	double 		doy,seconds;

	//EMS files do not contain a header. It has to be manually added to the file
	if(sourcefile==1 && headerwritten==0) {
		//Write first line of header (RINEX VERSION / TYPE)
		ret=fprintf(fdlist[fdRINEXB],"     2.10           B                                       RINEX VERSION / TYPE\n");
		if(ret<0) {
			fclose(readfd);		//Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);	//Close all other files before exiting
			sprintf(aux,"When writing SBAS data to RINEX B file [%s]\n",filename);
			printError(aux,options);
		}

		//Write second line of header (PGM / RUN BY / DATE)
		getcurrentsystemtime(&tm);	//Get current system time
		MonthName(tm.tm_mon+1,monthname);
		ret=fprintf(fdlist[fdRINEXB],"gLAB                gAGE (www.gage.es)  %02d-%.3s-%02d %02d:%02d     PGM / RUN BY / DATE\n",tm.tm_mday,monthname,tm.tm_year%100,tm.tm_hour,tm.tm_min);
		if(ret<0) {
			fclose(readfd);		//Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);	//Close all other files before exiting
			sprintf(aux,"When writing SBAS data to RINEX B file [%s]\n",filename);
			printError(aux,options);
		}

		//Write third line of header (END OF HEADER)
		ret=fprintf(fdlist[fdRINEXB],"                                                            END OF HEADER\n");
		if(ret<0) {
			fclose(readfd);		//Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);	//Close all other files before exiting
			sprintf(aux,"When writing SBAS data to RINEX B file [%s]\n",filename);
			printError(aux,options);
		}

		headerwritten=1;
	}

	//Header will only be true when the origin data file is a RINEX B
	if(header) {
		if(strncmp(&line[0][60],"PGM / RUN BY / DATE",19)==0) {
			//Change the author of the RINEX file
			getcurrentsystemtime(&tm);	//Get current system time
			MonthName(tm.tm_mon+1,monthname);
			ret=fprintf(fdlist[fdRINEXB],"gLAB                gAGE (www.gage.es)  %02d-%.3s-%02d %02d:%02d     PGM / RUN BY / DATE\n",tm.tm_mday,monthname,tm.tm_year%100,tm.tm_hour,tm.tm_min);
		} else {
			ret=fprintf(fdlist[fdRINEXB],"%s",line[0]);
		}
		if(ret<0) {
			fclose(readfd);		//Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);	//Close all other files before exiting
			sprintf(aux,"When writing SBAS data to RINEX B file [%s]\n",filename);
			printError(aux,options);
		}
		return;

	} else if (header==0 && headerwritten==0) {
		//If we don't have header line means that the header has ended
		headerwritten=1;
	}

	if(sourcefile==0) {
		//The origin file is a RINEX B
		for(i=0;i<numlines;i++) {
			ret=fprintf(fdlist[fdRINEXB],"%s",line[i]);
			if(ret<0) {
				fclose(readfd);		//Close RINEXB or EMS file where the data has been read
				closefiles(fdlist,18);	//Close all other files before exiting
				sprintf(aux,"When writing SBAS data to RINEX B file [%s]\n",filename);
				printError(aux,options);
			}
		}
	} else { //sourcefile==1
		//The origin file is a EMS. Time saved in time of applicability, and we have to write it in time of reception of first bit
		memcpy(&t,&sbasblock->t,sizeof(TTime));	//Make of copy of the time structure
		t.SoD--;					//Substract one second to time
		if(t.SoD<0) {
			//If it's negative then means that we are in the previous day
			t.MJDN--;
			t.SoD+=86400;
		}
		t2tmnolocal(&t,&tm,&seconds);
		t2doy(&t,&year,&doy);

		//Write first line of data
		ret=fprintf(fdlist[fdRINEXB],"%03d %02d %02d %02d %02d %02d%5.1f  L%1d   %3d   %3d   %3s\n",sbasblock->PRN,year%100,tm.tm_mon+1,tm.tm_mday,tm.tm_hour,tm.tm_min,seconds,sbasblock->frequency,sbasblock->datalength,sbasblock->receiverindex,sbasblock->sourceidentifyier);
		if(ret<0) {
			fclose(readfd);		//Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);	//Close all other files before exiting
			sprintf(aux,"When writing SBAS data to RINEX B file [%s]\n",filename);
			printError(aux,options);
		}

		//Write second line of data
		sprintf(aux," %2d   ",sbasblock->messagetype);
		for(i=0;i<18;i++) {
			if(sbasblock->messagetype<10) getstr(aux3,line[0],24+2*i,2);
			else getstr(aux3,line[0],25+2*i,2);
			sprintf(aux2,"%s %2s",aux,aux3);
			sprintf(aux,"%s",aux2);
		}
		
		ret=fprintf(fdlist[fdRINEXB],"%s\n",aux);
		if(ret<0) {
			fclose(readfd);		//Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);	//Close all other files before exiting
			sprintf(aux,"When writing SBAS data to RINEX B file [%s]\n",filename);
			printError(aux,options);
		}

		//Write third line of data
		sprintf(aux,"      ");
		for(i=0;i<14;i++) {
			if(sbasblock->messagetype<10) getstr(aux3,line[0],60+i*2,2);
			else getstr(aux3,line[0],61+i*2,2);
			sprintf(aux2,"%s %2s",aux,aux3);
			sprintf(aux,"%s",aux2);
		}
		
		ret=fprintf(fdlist[fdRINEXB],"%s\n",aux);
		if(ret<0) {
			fclose(readfd);		//Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);	//Close all other files before exiting
			sprintf(aux,"When writing SBAS data to RINEX B file [%s]\n",filename);
			printError(aux,options);
		}
	}
}

/*****************************************************************************
 * Name        : writeEMSfile
 * Description : Writes a EMS file from SBAS data. This file will not
 *               contain any corrupted message
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  **fdlist                  I  N/A  Vector of file descriptors for writing data
 * FILE  *readfd		   I  N/A  File descriptor of the data that is being read
 * char *filename                  I  N/A  Filename where the data is saved
 * char **line                     I  N/A  Data lines to be written
 * int numlines                    I  N/A  Number of lines given in string "line"
 * int header                      I  N/A  Indicates if the current lines
 *                                         are from a RINEXB header
 *                                         0 => Not a header line
 *                                         1 => Header line
 * int sourcefile                  I  N/A  Indicates from what type file the
 *                                         SBAS data has been read.
 *                                         0 => RINEXB
 *                                         1 => EMS
 * TSBASblock  *sbasblock          I  N/A  TSBASblock struct 					   
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void writeEMSfile (FILE **fdlist, FILE *readfd, char *filename, char **line, int numlines, int header, int sourcefile,  TSBASblock  *sbasblock, TOptions  *options) {
	int 	i,j;
	int 	ret;
	int 	numberofcolumns;
	char 	aux[200],hexstring[300];
	struct	tm tm;
	int 	year;
	double 	doy,seconds;

	//There are no header lines in EMS file
	if(header==1) return;

	if(sourcefile==0) {
		//The origin file is a RINEX B, but time is saved in time of applicability, the same as EMS file
		t2tmnolocal(&sbasblock->t,&tm,&seconds);
		t2doy(&sbasblock->t,&year,&doy);
		
		//Save all the hexademical values in a single string
		for(i=1;i<numlines;i++) {
			//Compute the number of records (columns) in the current line (normally the last one is not full)
			if(i==numlines-1) {
				numberofcolumns=sbasblock->datalength%18;
			} else {
				numberofcolumns=18;
			}

			//Read lines
			for(j=0;j<numberofcolumns;j++) {
				getstr(&hexstring[18*2*(i-1)+2*j],line[i],7+3*j,2);
			}
		}
		ret=fprintf(fdlist[fdEMS],"%03d %02d %02d %02d %02d %02d %02d %d %s\n",sbasblock->PRN,year,tm.tm_mon+1,tm.tm_mday,tm.tm_hour,tm.tm_min,tm.tm_sec,sbasblock->messagetype,hexstring);
		if(ret<0) {
			fclose(readfd);		//Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);	//Close all other files before exiting
			sprintf(aux,"When writing SBAS data to EMS file [%s]\n",filename);
			printError(aux,options);
		}
	} else {
		//The origin file is a EMS file
		ret=fprintf(fdlist[fdEMS],"%s",line[0]);
		if(ret<0) {
			fclose(readfd);		//Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);	//Close all other files before exiting
			sprintf(aux,"When writing SBAS data to EMS file [%s]\n",filename);
			printError(aux,options);
		}

	}

}

/*****************************************************************************
 * Name        : writeSBASdecodesummary
 * Description : Writes a EMS file from SBAS data. This file will not
 *               contain any corrupted message
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  **fdlist                  I  N/A  Vector of file descriptors for writing data
 * FILE  *readfd                   I  N/A  File descriptor of the data that is being read
 * char *filename                  I  N/A  Filename where the data is saved
 * int  numberofbadmessages	   I  N/A  Number of incorrect messages read
 * int  totalmessagesmissing	   I  N/A  Total number of messages missing (detected with preamble)
 * int  msgmissingdetected	   I  N/A  Number of detections of messages missing
 * int  *linesmissing		   I  N/A  Line numbers where messages are missing
 * int  *listmissing		   I  N/A  Vector with the messages lost for each line
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void writeSBASdecodesummary (FILE  **fdlist,FILE  *readfd, char *filename, int  numberofbadmessages, int  totalmessagesmissing, int  msgmissingdetected, int  *linesmissing, int  *listmissing, TOptions  *options) {

	int	i;
	int	ret;
	char	aux[300];

	if(numberofbadmessages==0 && totalmessagesmissing==0) {
		ret=fprintf(fdlist[fdLOGFILE],"All messages correctly decoded\n");
		if(ret<0) {
			fclose(readfd);		//Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);	//Close all other files before exiting
			sprintf(aux,"When writing SBAS data to EMS file [%s]\n",filename);
					printError(aux,options);
		}
		return;
	}

	if(numberofbadmessages>0) {
		if(numberofbadmessages==1) {
			ret=fprintf(fdlist[fdLOGFILE],"%d message was incorrect\n",numberofbadmessages);
		} else {
			ret=fprintf(fdlist[fdLOGFILE],"%d messages were incorrect\n",numberofbadmessages);
		}
		if(ret<0) {
			fclose(readfd);		//Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);	//Close all other files before exiting
			sprintf(aux,"When writing SBAS data to EMS file [%s]\n",filename);
					printError(aux,options);
		}
	}
	if(totalmessagesmissing>0) {
		if(totalmessagesmissing==1) {
			ret=fprintf(fdlist[fdLOGFILE],"%d message was missing at line %d\n",totalmessagesmissing,linesmissing[0]);
			if(ret<0) {
				fclose(readfd);		//Close RINEXB or EMS file where the data has been read
				closefiles(fdlist,18);	//Close all other files before exiting
				sprintf(aux,"When writing SBAS data to EMS file [%s]\n",filename);
						printError(aux,options);
			}
		} else if (msgmissingdetected==1 && totalmessagesmissing>1) {
			ret=fprintf(fdlist[fdLOGFILE],"%d messages were missing at line %d\n",totalmessagesmissing,linesmissing[0]);
			if(ret<0) {
				fclose(readfd);		//Close RINEXB or EMS file where the data has been read
				closefiles(fdlist,18);	//Close all other files before exiting
				sprintf(aux,"When writing SBAS data to EMS file [%s]\n",filename);
						printError(aux,options);
			}
		} else {
			for(i=0;i<msgmissingdetected;i++) {
				if(listmissing[i]==1) {
					ret=fprintf(fdlist[fdLOGFILE],"%d message was missing at line %d\n",listmissing[i],linesmissing[i]);
				} else {
					ret=fprintf(fdlist[fdLOGFILE],"%d messages were missing at line %d\n",listmissing[i],linesmissing[i]);
				}
				if(ret<0) {
					fclose(readfd);		//Close RINEXB or EMS file where the data has been read
					closefiles(fdlist,18);	//Close all other files before exiting
					sprintf(aux,"When writing SBAS data to EMS file [%s]\n",filename);
							printError(aux,options);
				}
			}
		}
	}
}


/*****************************************************************************
 * Name        : writePegasusheaders
 * Description : Writes all the headers for the Pegasus files
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  **fdlist                  I  N/A  Vector of file descriptors for writing data
 * FILE  *readfd		   I  N/A  File descriptor of the data that is being read
 * char **filelist		   I  N/A  List of filenames for writing data
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void writePegasusheaders (FILE **fdlist, FILE *readfd, char **filelist, TOptions  *options) {

	int		ret;
	char		aux[200];
	char		monthname[10];
	struct 		tm tm;
	TTime 		t;
	int 		year;
	double 		doy,seconds;

	//Write header of Message sequence
	if(options->pegasusaligned==1) {
		ret=fprintf(fdlist[fdMESSAGESEQUENCE],"\"RX_WEEK\"%c   \"RX_TOM\"%c   \"PRN\"%c  \"CRC_PASS\"%c  \"VALID\"%c   \"RMT\"%c   \"DMT\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	} else {
		ret=fprintf(fdlist[fdMESSAGESEQUENCE],"\"RX_WEEK\"%c\"RX_TOM\"%c\"PRN\"%c\"CRC_PASS\"%c\"VALID\"%c\"RMT\"%c\"DMT\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	}
	if(ret<0) {
		fclose(readfd);		//Close RINEXB or EMS file where the data has been read
		closefiles(fdlist,18);	//Close all other files before exiting
		sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdMESSAGESEQUENCE]);
		printError(aux,options);
	}

	//Write header of Alarm message (Message type 0)
	if(options->pegasusaligned==1) {
		ret=fprintf(fdlist[fdALARMMESSAGE],"\"RX_WEEK\"%c   \"RX_TOM\"%c   \"PRN\"\n",options->pegasusfs,options->pegasusfs);
	} else {
		ret=fprintf(fdlist[fdALARMMESSAGE],"\"RX_WEEK\"%c\"RX_TOM\"%c\"PRN\"\n",options->pegasusfs,options->pegasusfs);
	}
	if(ret<0) {
		fclose(readfd);		//Close RINEXB or EMS file where the data has been read
		closefiles(fdlist,18);	//Close all other files before exiting
		sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdALARMMESSAGE]);
		printError(aux,options);
	}

	//Write header of PRN Mask (Message type 1)
	if(options->pegasusaligned==1) {
		ret=fprintf(fdlist[fdPRNMASK],"\"RX_WEEK\"%c   \"RX_TOM\"%c   \"PRN\"%c   \"SVMASK\"%c                                                 \"IODP\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	} else {
		ret=fprintf(fdlist[fdPRNMASK],"\"RX_WEEK\"%c\"RX_TOM\"%c\"PRN\"%c\"SVMASK\"%c\"IODP\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	}
	if(ret<0) {
		fclose(readfd);		//Close RINEXB or EMS file where the data has been read
		closefiles(fdlist,18);	//Close all other files before exiting
		sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdPRNMASK]);
		printError(aux,options);
	}

	//Write header of Fast Corrections (Message type 2,3,4,5,24)
	if(options->pegasusaligned==1) {
		ret=fprintf(fdlist[fdFASTCORRECTIONS],"\"RX_WEEK\"%c   \"RX_TOM\"%c   \"PRN\"%c   \"TYPE\"%c   \"IODF\"%c   \"IODP\"%c   \"PRC01\"%c \"PRC02\"%c \"PRC03\"%c \"PRC04\"%c \"PRC05\"%c \"PRC06\"%c \"PRC07\"%c \"PRC08\"%c \"PRC09\"%c \"PRC10\"%c \"PRC11\"%c \"PRC12\"%c \"PRC13\"%c \"UDREI01\"%c \"UDREI02\"%c \"UDREI03\"%c \"UDREI04\"%c \"UDREI05\"%c \"UDREI06\"%c \"UDREI07\"%c \"UDREI08\"%c \"UDREI09\"%c \"UDREI10\"%c \"UDREI11\"%c \"UDREI12\"%c \"UDREI13\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	} else {
		ret=fprintf(fdlist[fdFASTCORRECTIONS],"\"RX_WEEK\"%c\"RX_TOM\"%c\"PRN\"%c\"TYPE\"%c\"IODF\"%c\"IODP\"%c\"PRC01\"%c\"PRC02\"%c\"PRC03\"%c\"PRC04\"%c\"PRC05\"%c\"PRC06\"%c\"PRC07\"%c\"PRC08\"%c\"PRC09\"%c\"PRC10\"%c\"PRC11\"%c\"PRC12\"%c\"PRC13\"%c\"UDREI01\"%c\"UDREI02\"%c\"UDREI03\"%c\"UDREI04\"%c\"UDREI05\"%c\"UDREI06\"%c\"UDREI07\"%c\"UDREI08\"%c\"UDREI09\"%c\"UDREI10\"%c\"UDREI11\"%c\"UDREI12\"%c\"UDREI13\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	}
	if(ret<0) {
		fclose(readfd);		//Close RINEXB or EMS file where the data has been read
		closefiles(fdlist,18);	//Close all other files before exiting
		sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdFASTCORRECTIONS]);
		printError(aux,options);
	}

	//Write header of Integrity (Message type 6)
	if(options->pegasusaligned==1) {
		ret=fprintf(fdlist[fdINTEGRITY],"\"RX_WEEK\"%c   \"RX_TOM\"%c   \"PRN\"%c \"IODF2\"%c \"IODF3\"%c \"IODF4\"%c \"IODF5\"%c \"UDREI01\"%c \"UDREI02\"%c \"UDREI03\"%c \"UDREI04\"%c \"UDREI05\"%c \"UDREI06\"%c \"UDREI07\"%c \"UDREI08\"%c \"UDREI09\"%c \"UDREI10\"%c \"UDREI11\"%c \"UDREI12\"%c \"UDREI13\"%c \"UDREI14\"%c \"UDREI15\"%c \"UDREI16\"%c \"UDREI17\"%c \"UDREI18\"%c \"UDREI19\"%c \"UDREI20\"%c \"UDREI21\"%c \"UDREI22\"%c \"UDREI23\"%c \"UDREI24\"%c \"UDREI25\"%c \"UDREI26\"%c \"UDREI27\"%c \"UDREI28\"%c \"UDREI29\"%c \"UDREI30\"%c \"UDREI31\"%c \"UDREI32\"%c \"UDREI33\"%c \"UDREI34\"%c \"UDREI35\"%c \"UDREI36\"%c \"UDREI37\"%c \"UDREI38\"%c \"UDREI39\"%c \"UDREI40\"%c \"UDREI41\"%c \"UDREI42\"%c \"UDREI43\"%c \"UDREI44\"%c \"UDREI45\"%c \"UDREI46\"%c \"UDREI47\"%c \"UDREI48\"%c \"UDREI49\"%c \"UDREI50\"%c \"UDREI51\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	} else {
		ret=fprintf(fdlist[fdINTEGRITY],"\"RX_WEEK\"%c\"RX_TOM\"%c\"PRN\"%c\"IODF2\"%c\"IODF3\"%c\"IODF4\"%c\"IODF5\"%c\"UDREI01\"%c\"UDREI02\"%c\"UDREI03\"%c\"UDREI04\"%c\"UDREI05\"%c\"UDREI06\"%c\"UDREI07\"%c\"UDREI08\"%c\"UDREI09\"%c\"UDREI10\"%c\"UDREI11\"%c\"UDREI12\"%c\"UDREI13\"%c\"UDREI14\"%c\"UDREI15\"%c\"UDREI16\"%c\"UDREI17\"%c\"UDREI18\"%c\"UDREI19\"%c\"UDREI20\"%c\"UDREI21\"%c\"UDREI22\"%c\"UDREI23\"%c\"UDREI24\"%c\"UDREI25\"%c\"UDREI26\"%c\"UDREI27\"%c\"UDREI28\"%c\"UDREI29\"%c\"UDREI30\"%c\"UDREI31\"%c\"UDREI32\"%c\"UDREI33\"%c\"UDREI34\"%c\"UDREI35\"%c\"UDREI36\"%c\"UDREI37\"%c\"UDREI38\"%c\"UDREI39\"%c\"UDREI40\"%c\"UDREI41\"%c\"UDREI42\"%c\"UDREI43\"%c\"UDREI44\"%c\"UDREI45\"%c\"UDREI46\"%c\"UDREI47\"%c\"UDREI48\"%c\"UDREI49\"%c\"UDREI50\"%c\"UDREI51\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	}
	if(ret<0) {
		fclose(readfd);		//Close RINEXB or EMS file where the data has been read
		closefiles(fdlist,18);	//Close all other files before exiting
		sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdINTEGRITY]);
		printError(aux,options);
	}


	//Write header of Fast Degradation (Message type 7)
	if(options->pegasusaligned==1) {
		ret=fprintf(fdlist[fdFASTDEGRADATION],"\"RX_WEEK\"%c   \"RX_TOM\"%c   \"PRN\"%c   \"TLAT\"%c   \"IODP\"%c   \"AI01\"%c \"AI02\"%c \"AI03\"%c \"AI04\"%c \"AI05\"%c \"AI06\"%c \"AI07\"%c \"AI08\"%c \"AI09\"%c \"AI10\"%c \"AI11\"%c \"AI12\"%c \"AI13\"%c \"AI14\"%c \"AI15\"%c \"AI16\"%c \"AI17\"%c \"AI18\"%c \"AI19\"%c \"AI20\"%c \"AI21\"%c \"AI22\"%c \"AI23\"%c \"AI24\"%c \"AI25\"%c \"AI26\"%c \"AI27\"%c \"AI28\"%c \"AI29\"%c \"AI30\"%c \"AI31\"%c \"AI32\"%c \"AI33\"%c \"AI34\"%c \"AI35\"%c \"AI36\"%c \"AI37\"%c \"AI38\"%c \"AI39\"%c \"AI40\"%c \"AI41\"%c \"AI42\"%c \"AI43\"%c \"AI44\"%c \"AI45\"%c \"AI46\"%c \"AI47\"%c \"AI48\"%c \"AI49\"%c \"AI50\"%c \"AI51\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	} else {
		ret=fprintf(fdlist[fdFASTDEGRADATION],"\"RX_WEEK\"%c\"RX_TOM\"%c\"PRN\"%c\"TLAT\"%c\"IODP\"%c\"AI01\"%c\"AI02\"%c\"AI03\"%c\"AI04\"%c\"AI05\"%c\"AI06\"%c\"AI07\"%c\"AI08\"%c\"AI09\"%c\"AI10\"%c\"AI11\"%c\"AI12\"%c\"AI13\"%c\"AI14\"%c\"AI15\"%c\"AI16\"%c\"AI17\"%c\"AI18\"%c\"AI19\"%c\"AI20\"%c\"AI21\"%c\"AI22\"%c\"AI23\"%c\"AI24\"%c\"AI25\"%c\"AI26\"%c\"AI27\"%c\"AI28\"%c\"AI29\"%c\"AI30\"%c\"AI31\"%c\"AI32\"%c\"AI33\"%c\"AI34\"%c\"AI35\"%c\"AI36\"%c\"AI37\"%c\"AI38\"%c\"AI39\"%c\"AI40\"%c\"AI41\"%c\"AI42\"%c\"AI43\"%c\"AI44\"%c\"AI45\"%c\"AI46\"%c\"AI47\"%c\"AI48\"%c\"AI49\"%c\"AI50\"%c\"AI51\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	}
	if(ret<0) {
		fclose(readfd);		//Close RINEXB or EMS file where the data has been read
		closefiles(fdlist,18);	//Close all other files before exiting
		sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdFASTDEGRADATION]);
		printError(aux,options);
	}

	//Write header of Degradation (Message type 10)
	if(options->pegasusaligned==1) {
		ret=fprintf(fdlist[fdDEGRADATION],"\"RX_WEEK\"%c   \"RX_TOM\"%c   \"PRN\"%c   \"BRRC\"%c   \"CLTC_LSB\"%c   \"CLTC_V1\"%c   \"ILTC_V1\"%c   \"CLTC_V0\"%c   \"ILTC_V0\"%c   \"CGEO_LSB\"%c   \"CGEO_V\"%c   \"IGEO\"%c   \"CER\"%c   \"CIONO_STEP\"%c   \"IIONO\"%c   \"CIONO_RAMP\"%c   \"RSS_UDRE\"%c   \"RSS_IONO\"%c   \"CCOV\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	} else {
		ret=fprintf(fdlist[fdDEGRADATION],"\"RX_WEEK\"%c\"RX_TOM\"%c\"PRN\"%c\"BRRC\"%c\"CLTC_LSB\"%c\"CLTC_V1\"%c\"ILTC_V1\"%c\"CLTC_V0\"%c\"ILTC_V0\"%c\"CGEO_LSB\"%c\"CGEO_V\"%c\"IGEO\"%c\"CER\"%c\"CIONO_STEP\"%c\"IIONO\"%c\"CIONO_RAMP\"%c\"RSS_UDRE\"%c\"RSS_IONO\"%c\"CCOV\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	}
	if(ret<0) {
		fclose(readfd);		//Close RINEXB or EMS file where the data has been read
		closefiles(fdlist,18);	//Close all other files before exiting
		sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdDEGRADATION]);
		printError(aux,options);
	}

	//Write header of Ionospheric Grid (Message type 18)
	if(options->pegasusaligned==1) {
		ret=fprintf(fdlist[fdIONOSPHERICGRID],"\"RX_WEEK\"%c   \"RX_TOM\"%c   \"PRN\"%c   \"NB\"%c   \"BN\"%c   \"IODI\"%c   \"IGP\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	} else {
		ret=fprintf(fdlist[fdIONOSPHERICGRID],"\"RX_WEEK\"%c\"RX_TOM\"%c\"PRN\"%c\"NB\"%c\"BN\"%c\"IODI\"%c\"IGP\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	}
	if(ret<0) {
		fclose(readfd);		//Close RINEXB or EMS file where the data has been read
		closefiles(fdlist,18);	//Close all other files before exiting
		sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdIONOSPHERICGRID]);
		printError(aux,options);
	}

	//Write header of SBAS Slow Corrections (Message type 24,25)
	if(options->pegasusaligned==1) {
		ret=fprintf(fdlist[fdSLOWCORRECTIONS],"\"RX_WEEK\"%c   \"RX_TOM\"%c   \"PRN\"%c   \"VC1\"%c \"MASK1\"%c \"IOD1\"%c \"DX1\"%c   \"DY1\"%c   \"DZ1\"%c   \"DAF01\"%c        \"DXROC1\"%c       \"DYROC1\"%c       \"DZROC1\"%c       \"DAF11\"%c        \"TOA1\"%c   \"IODP1\"%c \"VC2\"%c \"MASK2\"%c \"IOD2\"%c \"DX2\"%c   \"DY2\"%c   \"DZ2\"%c   \"DAF02\"%c        \"DXROC2\"%c       \"DYROC2\"%c       \"DZROC2\"%c       \"DAF12\"%c        \"TOA2\"%c   \"IODP2\"%c \"VC3\"%c \"MASK3\"%c \"IOD3\"%c \"DX3\"%c   \"DY3\"%c   \"DZ3\"%c   \"DAF03\"%c        \"DXROC3\"%c       \"DYROC3\"%c       \"DZROC3\"%c       \"DAF13\"%c        \"TOA3\"%c   \"IODP3\"%c \"VC4\"%c \"MASK4\"%c \"IOD4\"%c \"DX4\"%c   \"DY4\"%c   \"DZ4\"%c   \"DAF04\"%c        \"DXROC4\"%c       \"DYROC4\"%c       \"DZROC4\"%c       \"DAF14\"%c        \"TOA4\"%c   \"IODP4\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	} else {
		ret=fprintf(fdlist[fdSLOWCORRECTIONS],"\"RX_WEEK\"%c\"RX_TOM\"%c\"PRN\"%c\"VC1\"%c\"MASK1\"%c\"IOD1\"%c\"DX1\"%c\"DY1\"%c\"DZ1\"%c\"DAF01\"%c\"DXROC1\"%c\"DYROC1\"%c\"DZROC1\"%c\"DAF11\"%c\"TOA1\"%c\"IODP1\"%c\"VC2\"%c\"MASK2\"%c\"IOD2\"%c\"DX2\"%c\"DY2\"%c\"DZ2\"%c\"DAF02\"%c\"DXROC2\"%c\"DYROC2\"%c\"DZROC2\"%c\"DAF12\"%c\"TOA2\"%c\"IODP2\"%c\"VC3\"%c\"MASK3\"%c\"IOD3\"%c\"DX3\"%c\"DY3\"%c\"DZ3\"%c\"DAF03\"%c\"DXROC3\"%c\"DYROC3\"%c\"DZROC3\"%c\"DAF13\"%c\"TOA3\"%c\"IODP3\"%c\"VC4\"%c\"MASK4\"%c\"IOD4\"%c\"DX4\"%c\"DY4\"%c\"DZ4\"%c\"DAF04\"%c\"DXROC4\"%c\"DYROC4\"%c\"DZROC4\"%c\"DAF14\"%c\"TOA4\"%c\"IODP4\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	}
	if(ret<0) {
		fclose(readfd);		//Close RINEXB or EMS file where the data has been read
		closefiles(fdlist,18);	//Close all other files before exiting
		sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdSLOWCORRECTIONS]);
		printError(aux,options);
	}

	//Write header of Ionospheric Corrections (Message type 26)
	if(options->pegasusaligned==1) {
		ret=fprintf(fdlist[fdIONOSPHERICCORRECTIONS],"\"RX_WEEK\"%c   \"RX_TOM\"%c   \"PRN\"%c \"BN\"%c \"BI\"%c \"IODI\"%c \"DELAY01\"%c  \"GIVEI01\"%c \"DELAY02\"%c  \"GIVEI02\"%c \"DELAY03\"%c  \"GIVEI03\"%c \"DELAY04\"%c  \"GIVEI04\"%c \"DELAY05\"%c  \"GIVEI05\"%c \"DELAY06\"%c  \"GIVEI06\"%c \"DELAY07\"%c  \"GIVEI07\"%c \"DELAY08\"%c  \"GIVEI08\"%c \"DELAY09\"%c  \"GIVEI09\"%c \"DELAY10\"%c  \"GIVEI10\"%c \"DELAY11\"%c  \"GIVEI11\"%c \"DELAY12\"%c  \"GIVEI12\"%c \"DELAY13\"%c  \"GIVEI13\"%c \"DELAY14\"%c  \"GIVEI14\"%c \"DELAY15\"%c  \"GIVEI15\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	} else {
		ret=fprintf(fdlist[fdIONOSPHERICCORRECTIONS],"\"RX_WEEK\"%c\"RX_TOM\"%c\"PRN\"%c\"BN\"%c\"BI\"%c\"IODI\"%c\"DELAY01\"%c\"GIVEI01\"%c\"DELAY02\"%c\"GIVEI02\"%c\"DELAY03\"%c\"GIVEI03\"%c\"DELAY04\"%c\"GIVEI04\"%c\"DELAY05\"%c\"GIVEI05\"%c\"DELAY06\"%c\"GIVEI06\"%c\"DELAY07\"%c\"GIVEI07\"%c\"DELAY08\"%c\"GIVEI08\"%c\"DELAY09\"%c\"GIVEI09\"%c\"DELAY10\"%c\"GIVEI10\"%c\"DELAY11\"%c\"GIVEI11\"%c\"DELAY12\"%c\"GIVEI12\"%c\"DELAY13\"%c\"GIVEI13\"%c\"DELAY14\"%c\"GIVEI14\"%c\"DELAY15\"%c\"GIVEI15\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	}
	if(ret<0) {
		fclose(readfd);		//Close RINEXB or EMS file where the data has been read
		closefiles(fdlist,18);	//Close all other files before exiting
		sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdIONOSPHERICCORRECTIONS]);
		printError(aux,options);
	}

	//Write header of Service Region (Message type 27)
	if(options->pegasusaligned==1) {
		ret=fprintf(fdlist[fdSERVICEREGION27],"\"RX_WEEK\"%c   \"RX_TOM\"%c   \"PRN\"%c \"IODS\"%c \"NS\"%c \"MSG\"%c \"NR\"%c \"PC\"%c \"DELTAI\"%c \"DELTAO\"%c \"LAT11\"%c \"LON11\"%c \"LAT21\"%c \"LON21\"%c \"RS1\"%c \"LAT12\"%c \"LON12\"%c \"LAT22\"%c \"LON22\"%c \"RS2\"%c \"LAT13\"%c \"LON13\"%c \"LAT23\"%c \"LON23\"%c \"RS3\"%c \"LAT14\"%c \"LON14\"%c \"LAT24\"%c \"LON24\"%c \"RS4\"%c \"LAT15\"%c \"LON15\"%c \"LAT25\"%c \"LON25\"%c \"RS5\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	} else {
		ret=fprintf(fdlist[fdSERVICEREGION27],"\"RX_WEEK\"%c\"RX_TOM\"%c\"PRN\"%c\"IODS\"%c\"NS\"%c\"MSG\"%c\"NR\"%c\"PC\"%c\"DELTAI\"%c\"DELTAO\"%c\"LAT11\"%c\"LON11\"%c\"LAT21\"%c\"LON21\"%c\"RS1\"%c\"LAT12\"%c\"LON12\"%c\"LAT22\"%c\"LON22\"%c\"RS2\"%c\"LAT13\"%c\"LON13\"%c\"LAT23\"%c\"LON23\"%c\"RS3\"%c\"LAT14\"%c\"LON14\"%c\"LAT24\"%c\"LON24\"%c\"RS4\"%c\"LAT15\"%c\"LON15\"%c\"LAT25\"%c\"LON25\"%c\"RS5\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	}
	if(ret<0) {
		fclose(readfd);		//Close RINEXB or EMS file where the data has been read
		closefiles(fdlist,18);	//Close all other files before exiting
		sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdSERVICEREGION27]);
		printError(aux,options);
	}

	//Write header of Service Region (Message type 28)
	if(options->pegasusaligned==1) {
		ret=fprintf(fdlist[fdSERVICEREGION28],"\"RX_WEEK\"%c   \"RX_TOM\"%c   \"PRN\"%c   \"IODP\"%c \"MASK1\"%c \"SE1\"%c \"E111\"%c \"E221\"%c \"E331\"%c \"E441\"%c \"E121\"%c \"E131\"%c \"E141\"%c \"E231\"%c \"E241\"%c \"E341\"%c \"MASK2\"%c \"SE2\"%c \"E112\"%c \"E222\"%c \"E332\"%c \"E442\"%c \"E122\"%c \"E132\"%c \"E142\"%c \"E232\"%c \"E242\"%c \"E342\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	} else {
		ret=fprintf(fdlist[fdSERVICEREGION28],"\"RX_WEEK\"%c\"RX_TOM\"%c\"PRN\"%c\"IODP\"%c\"MASK1\"%c\"SE1\"%c\"E111\"%c\"E221\"%c\"E331\"%c\"E441\"%c\"E121\"%c\"E131\"%c\"E141\"%c\"E231\"%c\"E241\"%c\"E341\"%c\"MASK2\"%c\"SE2\"%c\"E112\"%c\"E222\"%c\"E332\"%c\"E442\"%c\"E122\"%c\"E132\"%c\"E142\"%c\"E232\"%c\"E242\"%c\"E342\"\n",options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs,options->pegasusfs);
	}
	if(ret<0) {
		fclose(readfd);		//Close RINEXB or EMS file where the data has been read
		closefiles(fdlist,18);	//Close all other files before exiting
		sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdSERVICEREGION28]);
		printError(aux,options);
	}

	//Write header of Ephemeris (Message type 9) (In Rinex 2.10 format)
		//Write first line of header (RINEX VERSION / TYPE)
		ret=fprintf(fdlist[fdEPHEMERIS],"     2.10           H: GEO NAV MSG DATA                     RINEX VERSION / TYPE\n");
		if(ret<0) {
			fclose(readfd);         //Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);  //Close all other files before exiting
			sprintf(aux,"When writing SBAS data to RINEX H file [%s]\n",filelist[fdEPHEMERIS]);
			printError(aux,options);
		}
		ret=fprintf(fdlist[fdEPHEMERISPREVDAY],"     2.10           H: GEO NAV MSG DATA                     RINEX VERSION / TYPE\n");
		if(ret<0) {
			fclose(readfd);         //Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);  //Close all other files before exiting
			sprintf(aux,"When writing SBAS data to RINEX H file [%s]\n",filelist[fdEPHEMERISPREVDAY]);
			printError(aux,options);
		}

		//Write second line of header (PGM / RUN BY / DATE)
		getcurrentsystemtime(&tm);      //Get current system time
		MonthName(tm.tm_mon+1,monthname);
		ret=fprintf(fdlist[fdEPHEMERIS],"gLAB                gAGE (www.gage.es)  %02d-%.3s-%02d %02d:%02d     PGM / RUN BY / DATE\n",tm.tm_mday,monthname,tm.tm_year%100,tm.tm_hour,tm.tm_min);
		if(ret<0) {
			fclose(readfd);         //Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);  //Close all other files before exiting
			sprintf(aux,"When writing SBAS data to RINEX H file [%s]\n",filelist[fdEPHEMERIS]);
			printError(aux,options);
		}
		ret=fprintf(fdlist[fdEPHEMERISPREVDAY],"gLAB                gAGE (www.gage.es)  %02d-%.3s-%02d %02d:%02d     PGM / RUN BY / DATE\n",tm.tm_mday,monthname,tm.tm_year%100,tm.tm_hour,tm.tm_min);
		if(ret<0) {
			fclose(readfd);         //Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);  //Close all other files before exiting
			sprintf(aux,"When writing SBAS data to RINEX H file [%s]\n",filelist[fdEPHEMERISPREVDAY]);
			printError(aux,options);
		}

		//Write third line of header (END OF HEADER)
		ret=fprintf(fdlist[fdEPHEMERIS],"                                                            END OF HEADER\n");
		if(ret<0) {
			fclose(readfd);         //Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);  //Close all other files before exiting
			sprintf(aux,"When writing SBAS data to RINEX H file [%s]\n",filelist[fdEPHEMERIS]);
			printError(aux,options);
		}
		ret=fprintf(fdlist[fdEPHEMERISPREVDAY],"                                                            END OF HEADER\n");
		if(ret<0) {
			fclose(readfd);         //Close RINEXB or EMS file where the data has been read
			closefiles(fdlist,18);  //Close all other files before exiting
			sprintf(aux,"When writing SBAS data to RINEX H file [%s]\n",filelist[fdEPHEMERISPREVDAY]);
			printError(aux,options);
		}
}

/*****************************************************************************
 * Name        : writePegasusfiles
 * Description : Writes data in Pegasus file format in their corresponding files
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  **fdlist                  I  N/A  Vector of file descriptors for writing data
 * FILE  *readfd                   I  N/A  File descriptor of the data that is being read
 * char **filelist		   I  N/A  List of filenames for writing data
 * char **line 		           I  N/A  Data lines from the data source file
 * int numlines                    I  N/A  Number of lines given in string "line"
 * int sourcefile                  I  N/A  Indicates from what type file the
 *                                         SBAS data has been read.
 *                                         0 => RINEXB
 *                                         1 => EMS
 * int header                      I  N/A  Indicates if the current lines
 *                                         are from a header
 *                                         0 => Not a header line
 *                                         1 => Header line
 * int decoderesult		   I  N/A  Result of the message decodification
 * int decodedmessagetype          I  N/A  Message type decoded from hexadecimal message 					   
 * int *prevday			   O  N/A  Indicates if a RINEX H file from the
 *                                         previous day has been written
 * TSBASblock  *sbasblock          I  N/A  TSBASblock struct                                      
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void writePegasusfiles (FILE **fdlist, FILE *readfd, char **filelist, char **line, int numlines, int sourcefile, int header, int decoderesult, int decodedmessagetype, int *prevday, TSBASblock  *sbasblock, TOptions  *options) {

	static int	headerwritten=0;
	int		i,j;
	int		ret;
	int		number;
	int		type;
	int		numberofcolumns;
	int		crcpass=1,valid=1;
	int		GPSweek,GPSweektrans,GPSweektoe;
	int		previousday=0,filetoprint;
	static int	firstbrdcday=-1;
	double		SoW,SoWtrans,SoWtoe;
	double		seconds;
	double		uravalue;
	double		health;
	double		transtimesecadjusted;
	double		igpverticaldelay;
	char		aux[1000];
	char		aux2[1000];
	char            hexstring[300];
	char            binarystring[2400];
	struct tm 	tm;
	TTime		transtime,toe;
	

	//This variable is to save the last health status read. We initialize all to 0, so if we don't have any health information (from message 18) put 31 as default (bits 4,3,2,1,0 to 1)
	//See below (on GEO Navigation message output) the Rinex specfications
	//The size of 90 is because there are 90 possible GEO PRN
	static double	lasthealthlist[91]={31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31};

	//If we have header lines from RINEX B don't do nothing
	if(header) return;

	//Write the header for all file types
	if(headerwritten==0) {
		writePegasusheaders(fdlist,readfd,filelist,options);
		headerwritten=1;
	}

	//Calculate GPS Week and Seconds of Week
	ttime2gpswsnoroll(&sbasblock->t, &GPSweek, &SoW);

	//Print Message Transmission Sequence
	if(decoderesult<0) {
		crcpass=0;
		valid=0;	//Valid variable should contain the same info as CRCPASS. but it can be manually turned to 0 to simulate errors
	}

	//If type 0 are decoded as type2, change the decodedmessagetype value
	if(options->alarmmessageastype2==1 && decodedmessagetype==DONTUSE) {
		decodedmessagetype=2;
	}
	if(options->pegasusaligned==1) {
		ret=fprintf(fdlist[fdMESSAGESEQUENCE],"%-4d%c        %-9.2f%c  %-3d%c    %-d%c           %-d%c         %-2d%c      %-2d\n",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,crcpass,options->pegasusfs,valid,options->pegasusfs,sbasblock->messagetype,options->pegasusfs,decodedmessagetype);
	} else {
		ret=fprintf(fdlist[fdMESSAGESEQUENCE],"%d%c%.2f%c%d%c%d%c%d%c%d%c%d\n",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,crcpass,options->pegasusfs,valid,options->pegasusfs,sbasblock->messagetype,options->pegasusfs,decodedmessagetype);
	}
	if(ret<0) {
		fclose(readfd);         //Close RINEXB or EMS file where the data has been read
		closefiles(fdlist,18);  //Close all other files before exiting
		sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdMESSAGESEQUENCE]);
		printError(aux,options);
	}

	if(decoderesult<0) return;	//If message had errors it must not be written to Pegasus message type files


	switch(sbasblock->messagetype) {
		case(PRNMASKASSIGNMENTS):
			//First we need to extract the hexadecimal string corresponding to the SVMASK
			if(sourcefile==0) {
				//The origin file is RINEX B
				//First join the hexadecimal message in one string
				for(i=1;i<numlines;i++) {
					//Compute the number of records (columns) in the current line (normally the last one is not full)
					if(i==numlines-1) {
						numberofcolumns=sbasblock->datalength%18;
					} else {
						numberofcolumns=18;
					}

					//Read lines
					for(j=0;j<numberofcolumns;j++) {
						getstr(&hexstring[18*2*(i-1)+2*j],line[i],7+3*j,2);
					}
				}	

			} else {
				//The origin file is EMS
				if(sbasblock->messagetype<10) {	
					getstr(hexstring,line[0],24,64);	
				} else {
					getstr(hexstring,line[0],25,64);
				}
			}

			////Now we have the hexadecimal string. We need to convert to a binary string, then 
			////extract the corresponding bits, then add 6 trailing zeros so it is a multiple of 8 and then convert to a hexadecimal string
			
			//Convert the hexadecimal string to a binary string (1 bit per byte)
			strhextostrbin(hexstring,binarystring);
			
			//Extract the PRN Mask in binary string
			getstr(aux2,binarystring,14,210);

			//Add the 6 trailing zeros to the PRN Mask
			sprintf(aux,"%s000000",aux2);

			//Convert from binary to hexadecimal
			strbintostrhex(aux,hexstring,1);


			//Print PRN Mask (Message type 1)
			if(options->pegasusaligned==1) {
				ret=fprintf(fdlist[fdPRNMASK],"%-4d%c        %-9.2f%c  %-3d%c     \"%s\"%c %-d\n",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,hexstring,options->pegasusfs,sbasblock->IODP);
			} else {
				ret=fprintf(fdlist[fdPRNMASK],"%d%c%.2f%c%d%c\"%s\"%c%d\n",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,hexstring,options->pegasusfs,sbasblock->IODP);
			}
			if(ret<0) {
				fclose(readfd);         //Close RINEXB or EMS file where the data has been read
				closefiles(fdlist,18);  //Close all other files before exiting
				sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdPRNMASK]);
				printError(aux,options);
			}
			break;
		case(DONTUSE):
			//Print Alarm message (Message type 0)
			if(options->pegasusaligned==1) {
				ret=fprintf(fdlist[fdALARMMESSAGE],"%-4d%c        %-9.2f%c  %-3d\n",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN);
			} else {
				ret=fprintf(fdlist[fdALARMMESSAGE],"%d%c%.2f%c%d\n",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN);
			}
			if(ret<0) {
				fclose(readfd);         //Close RINEXB or EMS file where the data has been read
				closefiles(fdlist,18);  //Close all other files before exiting
				sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdALARMMESSAGE]);
				printError(aux,options);
			}
			if(options->alarmmessageastype2==0) break;
		case(FASTCORRECTIONS2):case(FASTCORRECTIONS3):case(FASTCORRECTIONS4):case(FASTCORRECTIONS5):case(MIXEDFASTLONGTERMCORRECTIONS):
			//Print Fast Corrections (Message type 2,3,4,5,24)

			if(sbasblock->messagetype==MIXEDFASTLONGTERMCORRECTIONS) {
				type=sbasblock->BlockID;
				number=6;
			} else {
				type=sbasblock->messagetype;
				number=13;
			}


			if(options->pegasusaligned==1) {
				ret=fprintf(fdlist[fdFASTCORRECTIONS],"%-4d%c        %-9.2f%c  %-3d%c     %-d%c        %-d%c        %-d%c        ",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,type,options->pegasusfs,sbasblock->IODF[0],options->pegasusfs,sbasblock->IODP,options->pegasusfs);
			} else {
				ret=fprintf(fdlist[fdFASTCORRECTIONS],"%d%c%.2f%c%d%c%d%c%d%c%d%c",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,type,options->pegasusfs,sbasblock->IODF[0],options->pegasusfs,sbasblock->IODP,options->pegasusfs);
			}
			if(ret<0) {
				fclose(readfd);         //Close RINEXB or EMS file where the data has been read
				closefiles(fdlist,18);  //Close all other files before exiting
				sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdFASTCORRECTIONS]);
				printError(aux,options);
			}

			//Write PRC
			aux[0]='\0';		//Initialize aux to empty
			for(i=0;i<number;i++) {
				if(options->pegasusaligned==1) {
					sprintf(aux2,"%s%-8.3f%c",aux,sbasblock->PRC[i],options->pegasusfs);	//Write PRC
				} else {
					sprintf(aux2,"%s%.3f%c",aux,sbasblock->PRC[i],options->pegasusfs);	//Write PRC
				}
				sprintf(aux,"%s",aux2);					//Copy to aux
			}

			if(sbasblock->messagetype==MIXEDFASTLONGTERMCORRECTIONS) {
				//In this case we only have 6 Fast Corrections, so the unavailable data (PRC 7-13) must have empty fields
				if(options->pegasusaligned==1) {
					if(options->pegasusfs==';') sprintf(aux2,"%s;;;;;;;                                                        ",aux);  //Add empty fields
					else sprintf(aux2,"%s                                                               ",aux);	//Add empty fields
				} else {
					if(options->pegasusfs==';') sprintf(aux2,"%s;;;;;;;",aux);  //Add empty fields
					else sprintf(aux2,"%s       ",aux);	//Add empty fields
				}
				sprintf(aux,"%s",aux2);					//Copy to aux
			}

			//Write UDREI
			for(i=0;i<number;i++) {
				if(i==number-1) {
					if(options->pegasusaligned==1) {
						sprintf(aux2,"%s%-2d",aux,sbasblock->UDREI[i]);      //Write UDREI
					} else {
						sprintf(aux2,"%s%d",aux,sbasblock->UDREI[i]);      //Write UDREI
					}
				} else {
					if(options->pegasusaligned==1) {
						sprintf(aux2,"%s%-2d%c        ",aux,sbasblock->UDREI[i],options->pegasusfs);	//Write UDREI
					} else {
						sprintf(aux2,"%s%d%c",aux,sbasblock->UDREI[i],options->pegasusfs);	//Write UDREI
					}

				}
				sprintf(aux,"%s",aux2);						//Copy to aux
			}

			if(sbasblock->messagetype==MIXEDFASTLONGTERMCORRECTIONS) {
				//In this case we only have 6 UDREI, so the unavailable data (UDREI 7-13) must have empty fields
				if(options->pegasusaligned==1) {
					if(options->pegasusfs==';') sprintf(aux2,"%s;;;;;;;",aux);      //Add empty fields
					else sprintf(aux2,"%s                                                                           ",aux);	//Add empty fields
				} else {
					if(options->pegasusfs==';') sprintf(aux2,"%s;;;;;;;",aux);      //Add empty fields
					else sprintf(aux2,"%s       ",aux);	//Add empty fields
				}
				sprintf(aux,"%s",aux2);					//Copy to aux
			}

			ret=fprintf(fdlist[fdFASTCORRECTIONS],"%s\n",aux);
				if(ret<0) {
						fclose(readfd);         //Close RINEXB or EMS file where the data has been read
						closefiles(fdlist,18);  //Close all other files before exiting
						sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdFASTCORRECTIONS]);
						printError(aux,options);
				}
			
			//If it is a Mixed Corrections message, we have to print the long term data. Skipping the break will result in executing the next case
			if(sbasblock->messagetype!=MIXEDFASTLONGTERMCORRECTIONS) break;

		case(LONGTERMSATELLITECORRECTIONS):
			//Print Slow Corrections (Message type 24,25)
			//In this case, the manual says that if there are less data sets, then they should be filled with zeros

			//Print first half of message
			if(sbasblock->longtermsaterrcorrections[0][VELOCITYCODE]==0) {
				i=0;
				if(options->pegasusaligned==1) {
					ret=fprintf(fdlist[fdSLOWCORRECTIONS],"%-4d%c        %-9.2f%c  %-3d%c     %-d%c     %-2d%c      %-3d%c    %-8.3f%c%-8.3f%c%-8.3f%c%-15E%c%-15E%c%-15E%c%-15E%c%-15E%c%-5d%c    %-d%c       %-d%c     %-2d%c      %-3d%c    %-8.3f%c%-8.3f%c%-8.3f%c%-15E%c%-15E%c%-15E%c%-15E%c%-15E%c%-5d%c    %-d%c",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][VELOCITYCODE],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][PRNMASKNUMBER],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][ISSUEOFDATA],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAX],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAY],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAZ],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAAF0],options->pegasusfs,
						0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][IODPPOSITION],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[1][VELOCITYCODE],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[1][PRNMASKNUMBER],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[1][ISSUEOFDATA],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[1][DELTAX],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[1][DELTAY],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[1][DELTAZ],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[1][DELTAAF0],options->pegasusfs,
						0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[1][IODPPOSITION],options->pegasusfs);
				} else {
					ret=fprintf(fdlist[fdSLOWCORRECTIONS],"%d%c%.2f%c%-3d%c%-d%c%d%c%d%c%.3f%c%.3f%c%.3f%c%E%c%E%c%E%c%E%c%E%c%d%c%d%c%d%c%d%c%d%c%.3f%c%.3f%c%.3f%c%E%c%E%c%E%c%E%c%E%c%d%c%d%c",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][VELOCITYCODE],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][PRNMASKNUMBER],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][ISSUEOFDATA],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAX],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAY],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAZ],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAAF0],options->pegasusfs,
						0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][IODPPOSITION],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[1][VELOCITYCODE],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[1][PRNMASKNUMBER],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[1][ISSUEOFDATA],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[1][DELTAX],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[1][DELTAY],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[1][DELTAZ],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[1][DELTAAF0],options->pegasusfs,
						0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[1][IODPPOSITION],options->pegasusfs);
				}
			} else { //sbasblock->longtermsaterrcorrections[0][VELOCITYCODE]==1
				i=1;
				if(options->pegasusaligned==1) {
					ret=fprintf(fdlist[fdSLOWCORRECTIONS],"%-4d%c        %-9.2f%c  %-3d%c     %-d%c     %-2d%c      %-3d%c    %-8.3f%c%-8.3f%c%-8.3f%c%-15E%c%-15.7f%c%-15.7f%c%-15.7f%c%-15E%c%-5d%c    %-d%c       %-d%c     %-2d%c      %-3d%c    %-8.3f%c%-8.3f%c%-8.3f%c%-15E%c%-15E%c%-15E%c%-15E%c%-15E%c%-5d%c    %-d%c",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][VELOCITYCODE],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][PRNMASKNUMBER],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][ISSUEOFDATA],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAX],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAY],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAZ],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAAF0],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAXRATEOFCHANGE],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAYRATEOFCHANGE],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAZRATEOFCHANGE],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAAF1],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][TIMEOFDAYAPPLICABILITY],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][IODPPOSITION],options->pegasusfs,
						0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,
						0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs);
				} else {
					ret=fprintf(fdlist[fdSLOWCORRECTIONS],"%d%c%.2f%c%d%c%d%c%d%c%d%c%.3f%c%.3f%c%.3f%c%E%c%.7f%c%.7f%c%.7f%c%E%c%d%c%d%c%d%c%d%c%d%c%.3f%c%.3f%c%.3f%c%E%c%E%c%E%c%E%c%E%c%d%c%d%c",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][VELOCITYCODE],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][PRNMASKNUMBER],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][ISSUEOFDATA],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAX],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAY],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAZ],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAAF0],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAXRATEOFCHANGE],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAYRATEOFCHANGE],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAZRATEOFCHANGE],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[0][DELTAAF1],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][TIMEOFDAYAPPLICABILITY],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[0][IODPPOSITION],options->pegasusfs,
						0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,
						0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs);
				}
			}
				if(ret<0) {
						fclose(readfd);         //Close RINEXB or EMS file where the data has been read
						closefiles(fdlist,18);  //Close all other files before exiting
						sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdSLOWCORRECTIONS]);
						printError(aux,options);
				}

			//Print second half of message
			if(sbasblock->messagetype==MIXEDFASTLONGTERMCORRECTIONS) {
				//We have the mixed message (type 24), so there is no second half of message. Print zeros
				if(options->pegasusaligned==1) {
					ret=fprintf(fdlist[fdSLOWCORRECTIONS],"       %-d%c     %-2d%c      %-3d%c    %-8.3f%c%-8.3f%c%-8.3f%c%-15E%c%-15E%c%-15E%c%-15E%c%-15E%c%-5d%c    %-d%c       %-d%c     %-2d%c      %-3d%c    %-8.3f%c%-8.3f%c%-8.3f%c%-15E%c%-15E%c%-15E%c%-15E%c%-15E%c%-5d%c    %-d\n",0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,0);
				} else {
					ret=fprintf(fdlist[fdSLOWCORRECTIONS],"%d%c%d%c%d%c%.3f%c%.3f%c%.3f%c%E%c%E%c%E%c%E%c%E%c%d%c%d%c%d%c%d%c%d%c%.3f%c%.3f%c%.3f%c%E%c%E%c%E%c%E%c%E%c%d%c%d\n",0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,0);
				}
			} else if(sbasblock->longtermsaterrcorrections[2-i][VELOCITYCODE]==0) {
				if(options->pegasusaligned==1) {
					ret=fprintf(fdlist[fdSLOWCORRECTIONS],"       %-d%c     %-2d%c      %-3d%c    %-8.3f%c%-8.3f%c%-8.3f%c%-15E%c%-15E%c%-15E%c%-15E%c%-15E%c%-5d%c    %-d%c       %-d%c     %-2d%c      %-3d%c    %-8.3f%c%-8.3f%c%-8.3f%c%-15E%c%-15E%c%-15E%c%-15E%c%-15E%c%-5d%c    %-d\n",
						(int)sbasblock->longtermsaterrcorrections[2-i][VELOCITYCODE],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[2-i][PRNMASKNUMBER],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[2-i][ISSUEOFDATA],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAX],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAY],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAZ],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAAF0],options->pegasusfs,
						0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[2-i][IODPPOSITION],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[3-i][VELOCITYCODE],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[3-i][PRNMASKNUMBER],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[3-i][ISSUEOFDATA],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[3-i][DELTAX],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[3-i][DELTAY],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[3-i][DELTAZ],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[3-i][DELTAAF0],options->pegasusfs,
						0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[3-i][IODPPOSITION]);
				} else {
					ret=fprintf(fdlist[fdSLOWCORRECTIONS],"%d%c%d%c%d%c%.3f%c%.3f%c%.3f%c%E%c%E%c%E%c%E%c%E%c%d%c%d%c%d%c%d%c%d%c%.3f%c%.3f%c%.3f%c%E%c%E%c%E%c%E%c%E%c%d%c%d\n",
						(int)sbasblock->longtermsaterrcorrections[2-i][VELOCITYCODE],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[2-i][PRNMASKNUMBER],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[2-i][ISSUEOFDATA],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAX],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAY],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAZ],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAAF0],options->pegasusfs,
						0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[2-i][IODPPOSITION],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[3-i][VELOCITYCODE],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[3-i][PRNMASKNUMBER],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[3-i][ISSUEOFDATA],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[3-i][DELTAX],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[3-i][DELTAY],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[3-i][DELTAZ],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[3-i][DELTAAF0],options->pegasusfs,
						0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[3-i][IODPPOSITION]);
				}
			} else { //sbasblock->longtermsaterrcorrections[2][VELOCITYCODE]==1) {
				if(options->pegasusaligned==1) {
					ret=fprintf(fdlist[fdSLOWCORRECTIONS],"       %-d%c     %-2d%c      %-3d%c    %-8.3f%c%-8.3f%c%-8.3f%c%-15E%c%-15.7f%c%-15.7f%c%-15.7f%c%-15E%c%-5d%c    %-d%c       %-d%c     %-2d%c      %-3d%c    %-8.3f%c%-8.3f%c%-8.3f%c%-15E%c%-15E%c%-15E%c%-15E%c%-15E%c%-5d%c    %-d\n",
						(int)sbasblock->longtermsaterrcorrections[2-i][VELOCITYCODE],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[2-i][PRNMASKNUMBER],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[2-i][ISSUEOFDATA],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAX],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAY],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAZ],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAAF0],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAXRATEOFCHANGE],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAYRATEOFCHANGE],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAZRATEOFCHANGE],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAAF1],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[2-i][TIMEOFDAYAPPLICABILITY],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[2-i][IODPPOSITION],options->pegasusfs,
						0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,0);
				} else {
					ret=fprintf(fdlist[fdSLOWCORRECTIONS],"%d%c%d%c%d%c%.3f%c%.3f%c%.3f%c%E%c%.7f%c%.7f%c%.7f%c%E%c%d%c%d%c%d%c%d%c%d%c%.3f%c%.3f%c%.3f%c%E%c%E%c%E%c%E%c%E%c%d%c%d\n",
						(int)sbasblock->longtermsaterrcorrections[2-i][VELOCITYCODE],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[2-i][PRNMASKNUMBER],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[2-i][ISSUEOFDATA],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAX],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAY],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAZ],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAAF0],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAXRATEOFCHANGE],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAYRATEOFCHANGE],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAZRATEOFCHANGE],options->pegasusfs,
						sbasblock->longtermsaterrcorrections[2-i][DELTAAF1],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[2-i][TIMEOFDAYAPPLICABILITY],options->pegasusfs,
						(int)sbasblock->longtermsaterrcorrections[2-i][IODPPOSITION],options->pegasusfs,
						0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0.0,options->pegasusfs,0,options->pegasusfs,0);
				}
			}
			if(ret<0) {
				fclose(readfd);         //Close RINEXB or EMS file where the data has been read
				closefiles(fdlist,18);  //Close all other files before exiting
				sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdSLOWCORRECTIONS]);
				printError(aux,options);
			}
			break;
		case(INTEGRITYINFO):
			//Print Integrity (Message type 6)
			if(options->pegasusaligned==1) {
				sprintf(aux,"%-4d%c        %-9.2f%c  %-3d%c   %-d%c       %-d%c       %-d%c       %-d%c       ",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,sbasblock->IODF[0],options->pegasusfs,sbasblock->IODF[1],options->pegasusfs,sbasblock->IODF[2],options->pegasusfs,sbasblock->IODF[3],options->pegasusfs);
			} else {
				sprintf(aux,"%d%c%.2f%c%d%c%d%c%d%c%d%c%d%c",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,sbasblock->IODF[0],options->pegasusfs,sbasblock->IODF[1],options->pegasusfs,sbasblock->IODF[2],options->pegasusfs,sbasblock->IODF[3],options->pegasusfs);
			}
			for(i=0;i<51;i++) {
				if(i==50) {
					if(options->pegasusaligned==1) {
						sprintf(aux2,"%s%-2d",aux,sbasblock->UDREIacu[i]);
					} else {
						sprintf(aux2,"%s%d",aux,sbasblock->UDREIacu[i]);
					}
				} else {
					if(options->pegasusaligned==1) {
						sprintf(aux2,"%s%-2d%c        ",aux,sbasblock->UDREIacu[i],options->pegasusfs);
					} else {
						sprintf(aux2,"%s%d%c",aux,sbasblock->UDREIacu[i],options->pegasusfs);
					}
				}
				sprintf(aux,"%s",aux2);
			}
			ret=fprintf(fdlist[fdINTEGRITY],"%s\n",aux);
			if(ret<0) {
				fclose(readfd);         //Close RINEXB or EMS file where the data has been read
				closefiles(fdlist,18);  //Close all other files before exiting
				sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdINTEGRITY]);
				printError(aux,options);
			}
			break;
		case(FASTCORRECTIONSDEGRADATIONFACTOR):
			//Print Integrity (Message type 7)
			if(options->pegasusaligned==1) {
				sprintf(aux,"%-4d%c        %-9.2f%c  %-3d%c     %-2d%c       %-2d%c       ",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,sbasblock->tlat,options->pegasusfs,sbasblock->IODP,options->pegasusfs);
			} else {
				sprintf(aux,"%d%c%.2f%c%d%c%d%c%d%c",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,sbasblock->tlat,options->pegasusfs,sbasblock->IODP,options->pegasusfs);
			}
			for(i=0;i<51;i++) {
				if(i==50) {
					if(options->pegasusaligned==1) {
						sprintf(aux2,"%s%-2d",aux,sbasblock->aiind[i]);
					} else {
						sprintf(aux2,"%s%d",aux,sbasblock->aiind[i]);
					}
				} else {
					if(options->pegasusaligned==1) {
						sprintf(aux2,"%s%-2d%c     ",aux,sbasblock->aiind[i],options->pegasusfs);
					} else {
						sprintf(aux2,"%s%d%c",aux,sbasblock->aiind[i],options->pegasusfs);
					}
				}
				sprintf(aux,"%s",aux2);
			}
			ret=fprintf(fdlist[fdFASTDEGRADATION],"%s\n",aux);
			if(ret<0) {
				fclose(readfd);         //Close RINEXB or EMS file where the data has been read
				closefiles(fdlist,18);  //Close all other files before exiting
				sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdFASTDEGRADATION]);
				printError(aux,options);
			}
			break;
		case(DEGRADATIONPARAMETERS):
			//Print Degradation parameters (Message type 10)
			if(options->pegasusaligned==1) {
				ret=fprintf(fdlist[fdDEGRADATION],"%-4d%c        %-9.2f%c  %-3d%c     %-5.3f%c    %-5.3f%c        %-7.5f%c     %-3d%c         %-5.3f%c       %-3d%c         %-6.4f%c       %-7.5f%c    %-3d%c      %-4.1f%c    %-5.3f%c          %-3d%c       %-8.6f%c       %-d%c            %-d%c            %-4.1f\n",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,
					 sbasblock->degradationfactors[BRRC],options->pegasusfs,
					 sbasblock->degradationfactors[CLTCLSB],options->pegasusfs,
					 sbasblock->degradationfactors[CLTCV1],options->pegasusfs,
					 (int)sbasblock->degradationfactors[ILTCV1],options->pegasusfs,
					 sbasblock->degradationfactors[CLTCV0],options->pegasusfs,
					 (int)sbasblock->degradationfactors[ILTCV0],options->pegasusfs,
					 sbasblock->degradationfactors[CGEOLSB],options->pegasusfs,
					 sbasblock->degradationfactors[CGEOV],options->pegasusfs,
					 (int)sbasblock->degradationfactors[IGEO],options->pegasusfs,
					 sbasblock->degradationfactors[CER],options->pegasusfs,
					 sbasblock->degradationfactors[CIONOSTEP],options->pegasusfs,
					 (int)sbasblock->degradationfactors[IIONO],options->pegasusfs,
					 sbasblock->degradationfactors[CIONORAMP],options->pegasusfs,
					 (int)sbasblock->degradationfactors[RSSUDRE],options->pegasusfs,
					 (int)sbasblock->degradationfactors[RSSIONO],options->pegasusfs,
					 sbasblock->degradationfactors[CCOVARIANCE]);
			} else {
				ret=fprintf(fdlist[fdDEGRADATION],"%d%c%.2f%c%d%c%.3f%c%.3f%c%.5f%c%d%c%.3f%c%d%c%.4f%c%.5f%c%d%c%.1f%c%.3f%c%d%c%.6f%c%-d%c%-d%c%.1f\n",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,
					 sbasblock->degradationfactors[BRRC],options->pegasusfs,
					 sbasblock->degradationfactors[CLTCLSB],options->pegasusfs,
					 sbasblock->degradationfactors[CLTCV1],options->pegasusfs,
					 (int)sbasblock->degradationfactors[ILTCV1],options->pegasusfs,
					 sbasblock->degradationfactors[CLTCV0],options->pegasusfs,
					 (int)sbasblock->degradationfactors[ILTCV0],options->pegasusfs,
					 sbasblock->degradationfactors[CGEOLSB],options->pegasusfs,
					 sbasblock->degradationfactors[CGEOV],options->pegasusfs,
					 (int)sbasblock->degradationfactors[IGEO],options->pegasusfs,
					 sbasblock->degradationfactors[CER],options->pegasusfs,
					 sbasblock->degradationfactors[CIONOSTEP],options->pegasusfs,
					 (int)sbasblock->degradationfactors[IIONO],options->pegasusfs,
					 sbasblock->degradationfactors[CIONORAMP],options->pegasusfs,
					 (int)sbasblock->degradationfactors[RSSUDRE],options->pegasusfs,
					 (int)sbasblock->degradationfactors[RSSIONO],options->pegasusfs,
					 sbasblock->degradationfactors[CCOVARIANCE]);
			}
				if(ret<0) {
						fclose(readfd);         //Close RINEXB or EMS file where the data has been read
						closefiles(fdlist,18);  //Close all other files before exiting
						sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdDEGRADATION]);
						printError(aux,options);
				}
			break;
		case(IONOSPHERICGRIDPOINTSMASKS):
			//First we need to extract the hexadecimal string corresponding to the IGP Mask
			if(sourcefile==0) {
				//The origin file is RINEX B
				//First join the hexadecimal message in one string
				for(i=1;i<numlines;i++) {
					//Compute the number of records (columns) in the current line (normally the last one is not full)
					if(i==numlines-1) {
						numberofcolumns=sbasblock->datalength%18;
					} else {
						numberofcolumns=18;
					}

					//Read lines
					for(j=0;j<numberofcolumns;j++) {
						getstr(&hexstring[18*2*(i-1)+2*j],line[i],7+3*j,2);
					}
				}	

			} else {
				//The origin file is EMS
				if(sbasblock->messagetype<10) {	
					getstr(hexstring,line[0],24,64);	
				} else {
					getstr(hexstring,line[0],25,64);
				}
			}

			
			//Convert the hexadecimal string to a binary string (1 bit per byte)
			strhextostrbin(hexstring,binarystring);
			
			//Extract the IGP Mask in binary string
			getstr(aux2,binarystring,24,201);

			//Add the 6 trailing zeros to the IGP Mask
			sprintf(aux,"%s0000000",aux2);

			//Convert from binary to hexadecimal
			strbintostrhex(aux,hexstring,1);

			//Print Ionospheric grid (Message type 18)
			if(options->pegasusaligned==1) {
				ret=fprintf(fdlist[fdIONOSPHERICGRID],"%-4d%c        %-9.2f%c  %-3d%c     %-2d%c     %-2d%c     %-d%c        \"%s\"\n",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,
					(int)sbasblock->igpmaskmessage[NUMBEROFBANDS],options->pegasusfs,
					(int)sbasblock->igpmaskmessage[BANDNUMBER],options->pegasusfs,
					(int)sbasblock->igpmaskmessage[IODIPOS],options->pegasusfs,
					hexstring);
			} else {
				ret=fprintf(fdlist[fdIONOSPHERICGRID],"%d%c%.2f%c%d%c%d%c%d%c%d%c\"%s\"\n",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,
					(int)sbasblock->igpmaskmessage[NUMBEROFBANDS],options->pegasusfs,
					(int)sbasblock->igpmaskmessage[BANDNUMBER],options->pegasusfs,
					(int)sbasblock->igpmaskmessage[IODIPOS],options->pegasusfs,
					hexstring);
			}
			if(ret<0) {
				fclose(readfd);         //Close RINEXB or EMS file where the data has been read
				closefiles(fdlist,18);  //Close all other files before exiting
				sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdIONOSPHERICGRID]);
				printError(aux,options);
			}
			
			break;
		case(IONOSPHERICDELAYCORRECTIONS):
			//Print Ionospheric corrections (message type 26)
			if(options->pegasusaligned==1) {
				ret=fprintf(fdlist[fdIONOSPHERICCORRECTIONS],"%-4d%c        %-9.2f%c  %-3d%c   %-2d%c   %-2d%c   %-d%c      ",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,
					(int)sbasblock->ionodelayparameters[0][BANDNUMBER],options->pegasusfs,
					(int)sbasblock->ionodelayparameters[0][BLOCKID],options->pegasusfs,
					(int)sbasblock->ionodelayparameters[0][IODIPOS],options->pegasusfs);
			} else {
				ret=fprintf(fdlist[fdIONOSPHERICCORRECTIONS],"%d%c%.2f%c%d%c%d%c%d%c%d%c",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,
					(int)sbasblock->ionodelayparameters[0][BANDNUMBER],options->pegasusfs,
					(int)sbasblock->ionodelayparameters[0][BLOCKID],options->pegasusfs,
					(int)sbasblock->ionodelayparameters[0][IODIPOS],options->pegasusfs);
			}
			if(ret<0) {
				fclose(readfd);         //Close RINEXB or EMS file where the data has been read
				closefiles(fdlist,18);  //Close all other files before exiting
				sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdIONOSPHERICCORRECTIONS]);
				printError(aux,options);
			}

			//Write IGP Vertical Delay and GIVEI
			aux[0]='\0';		//Initialize aux to empty
			for(i=0;i<15;i++) {
				if(sbasblock->ionodelayparameters[i][IGPVERTICALDELAY]==-1)  igpverticaldelay=63.875;
				else igpverticaldelay=sbasblock->ionodelayparameters[i][IGPVERTICALDELAY];
				if(i==14) {
					if(options->pegasusaligned==1) {
						sprintf(aux2,"%s%-6.3f%c     %-2d",aux,igpverticaldelay,options->pegasusfs,(int)sbasblock->ionodelayparameters[i][GIVEI]); //Write IGP and GIVEI
					} else {
						sprintf(aux2,"%s%.3f%c%d",aux,igpverticaldelay,options->pegasusfs,(int)sbasblock->ionodelayparameters[i][GIVEI]); //Write IGP and GIVEI
					}
				} else {
					if(options->pegasusaligned==1) {
						sprintf(aux2,"%s%-6.3f%c     %-2d%c        ",aux,igpverticaldelay,options->pegasusfs,(int)sbasblock->ionodelayparameters[i][GIVEI],options->pegasusfs); //Write IGP and GIVEI
					} else {
						sprintf(aux2,"%s%.3f%c%d%c",aux,igpverticaldelay,options->pegasusfs,(int)sbasblock->ionodelayparameters[i][GIVEI],options->pegasusfs); //Write IGP and GIVEI
					}

				}
				sprintf(aux,"%s",aux2);					//Copy to aux
			}

			ret=fprintf(fdlist[fdIONOSPHERICCORRECTIONS],"%s\n",aux);
			if(ret<0) {
				fclose(readfd);         //Close RINEXB or EMS file where the data has been read
				closefiles(fdlist,18);  //Close all other files before exiting
				sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdIONOSPHERICCORRECTIONS]);
				printError(aux,options);
			}

			break;
		case(SBASSERVICEMESSAGE):
			//Print SBAS Service message (message type 27)
			if(options->pegasusaligned==1) {
				ret=fprintf(fdlist[fdSERVICEREGION27],"%-4d%c        %-9.2f%c  %-3d%c   %-d%c      %-d%c    %-d%c     %-d%c    %-d%c    %-2d%c       %-2d%c       ",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,
					(int)sbasblock->servicemessage[IODSPOS],options->pegasusfs,
					(int)sbasblock->servicemessage[NUMBERSERVICEMESSAGES],options->pegasusfs,
					(int)sbasblock->servicemessage[SERVICEMESSAGENUMBER],options->pegasusfs,
					(int)sbasblock->servicemessage[NUMBEROFREGIONS],options->pegasusfs,
					(int)sbasblock->servicemessage[PRIORITYCODE],options->pegasusfs,
					(int)sbasblock->servicemessage[UDREINSIDEIND],options->pegasusfs,
					(int)sbasblock->servicemessage[UDREOUTSIDEIND],options->pegasusfs);
			} else {
				ret=fprintf(fdlist[fdSERVICEREGION27],"%d%c%.2f%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,
					(int)sbasblock->servicemessage[IODSPOS],options->pegasusfs,
					(int)sbasblock->servicemessage[NUMBERSERVICEMESSAGES],options->pegasusfs,
					(int)sbasblock->servicemessage[SERVICEMESSAGENUMBER],options->pegasusfs,
					(int)sbasblock->servicemessage[NUMBEROFREGIONS],options->pegasusfs,
					(int)sbasblock->servicemessage[PRIORITYCODE],options->pegasusfs,
					(int)sbasblock->servicemessage[UDREINSIDEIND],options->pegasusfs,
					(int)sbasblock->servicemessage[UDREOUTSIDEIND],options->pegasusfs);
			}
				if(ret<0) {
						fclose(readfd);         //Close RINEXB or EMS file where the data has been read
						closefiles(fdlist,18);  //Close all other files before exiting
						sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdSERVICEREGION27]);
						printError(aux,options);
				}

			//Write LAT1 LON1 LAT2 LON2 RS 
			aux[0]='\0';		//Initialize aux to empty
			for(i=0;i<5;i++) {
				if(i>=(int)sbasblock->servicemessage[NUMBEROFREGIONS]) {
					//If there are not five regions fill with zeros the missing regions
					if(i==4) {
						if(options->pegasusaligned==1) {
							sprintf(aux2,"%s%-3d%c     %-4d%c    %-3d%c     %-4d%c    %-d",aux,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0);
						} else {
							sprintf(aux2,"%s%d%c%d%c%d%c%d%c%d",aux,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0);
						}
					} else {
						if(options->pegasusaligned==1) {
							sprintf(aux2,"%s%-3d%c     %-4d%c    %-3d%c     %-4d%c    %-d%c     ",aux,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs);
						} else {
							sprintf(aux2,"%s%d%c%d%c%d%c%d%c%d%c",aux,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs);
						}
					}
				} else {
					if(i==4) {
						if(options->pegasusaligned==1) {
							sprintf(aux2,"%s%-3d%c     %-4d%c    %-3d%c     %-4d%c    %-d",aux,
								(int)sbasblock->regioncoordinates[i][COORD1LAT],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][COORD1LON],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][COORD2LAT],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][COORD2LON],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][REGIONSHAPE]); 
						} else {
							sprintf(aux2,"%s%d%c%d%c%d%c%d%c%d",aux,
								(int)sbasblock->regioncoordinates[i][COORD1LAT],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][COORD1LON],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][COORD2LAT],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][COORD2LON],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][REGIONSHAPE]); 
						}
					} else {
						if(options->pegasusaligned==1) {
							sprintf(aux2,"%s%-3d%c     %-4d%c    %-3d%c     %-4d%c    %-d%c     ",aux,
								(int)sbasblock->regioncoordinates[i][COORD1LAT],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][COORD1LON],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][COORD2LAT],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][COORD2LON],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][REGIONSHAPE],options->pegasusfs); 
						} else {
							sprintf(aux2,"%s%d%c%d%c%d%c%d%c%d%c",aux,
								(int)sbasblock->regioncoordinates[i][COORD1LAT],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][COORD1LON],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][COORD2LAT],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][COORD2LON],options->pegasusfs,
								(int)sbasblock->regioncoordinates[i][REGIONSHAPE],options->pegasusfs); 
						}

					}
				}
				sprintf(aux,"%s",aux2);					//Copy to aux
			}

			ret=fprintf(fdlist[fdSERVICEREGION27],"%s\n",aux);
			if(ret<0) {
				fclose(readfd);         //Close RINEXB or EMS file where the data has been read
				closefiles(fdlist,18);  //Close all other files before exiting
				sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdSERVICEREGION27]);
				printError(aux,options);
			}

			break;
		case(CLOCKEPHEMERISCOVARIANCEMATRIX):
			//Print SBAS Service Region (message type 28)
			if(sbasblock->clockephemeriscovariance[0][IODPPOSITION]==-1) {
				if(options->pegasusaligned==1) {
					ret=fprintf(fdlist[fdSERVICEREGION28],"%-4d%c        %-9.2f%c  %-3d%c     %-d%c      ",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,0,options->pegasusfs);
				} else {
					ret=fprintf(fdlist[fdSERVICEREGION28],"%d%c%.2f%c%d%c%d%c",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,0,options->pegasusfs);
				}
			} else {
				if(options->pegasusaligned==1) {
					ret=fprintf(fdlist[fdSERVICEREGION28],"%-4d%c        %-9.2f%c  %-3d%c     %-d%c      ",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,sbasblock->clockephemeriscovariance[0][IODPPOSITION],options->pegasusfs);
				} else {
					ret=fprintf(fdlist[fdSERVICEREGION28],"%d%c%.2f%c%d%c%d%c",GPSweek,options->pegasusfs,SoW,options->pegasusfs,sbasblock->PRN,options->pegasusfs,sbasblock->clockephemeriscovariance[0][IODPPOSITION],options->pegasusfs);
				}
			}
				if(ret<0) {
						fclose(readfd);         //Close RINEXB or EMS file where the data has been read
						closefiles(fdlist,18);  //Close all other files before exiting
						sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdSERVICEREGION28]);
						printError(aux,options);
				}

			aux[0]='\0';		//Initialize aux to empty
			for(i=0;i<2;i++) {
				if(i>=sbasblock->numclockephemeriscovariance || sbasblock->clockephemeriscovariance[0][IODPPOSITION]==-1) {
					//If we exceed the number of satellites fill with zeros
					if(i==0) {
						if(options->pegasusaligned==1) {
							sprintf(aux2,"%s%-2d%c      %-d%c     %-3d%c    %-3d%c    %-3d%c    %-3d%c    %-4d%c   %-4d%c   %-4d%c   %-4d%c   %-4d%c   %-4d%c   ",aux,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs);
						} else {
							sprintf(aux2,"%s%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c",aux,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs);
						}
					} else {
						if(options->pegasusaligned==1) {
							sprintf(aux2,"%s%-2d%c      %-d%c     %-3d%c    %-3d%c    %-3d%c    %-3d%c    %-4d%c   %-4d%c   %-4d%c   %-4d%c   %-4d%c   %-4d",aux,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0);
						} else {
							sprintf(aux2,"%s%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d",aux,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0,options->pegasusfs,0);
						}

					}
				} else {
					if(i==0) {
						if(options->pegasusaligned==1) {
							sprintf(aux2,"%s%-2d%c      %-d%c     %-3d%c    %-3d%c    %-3d%c    %-3d%c    %-4d%c   %-4d%c   %-4d%c   %-4d%c   %-4d%c   %-4d%c   ",aux,
								sbasblock->clockephemeriscovariance[i][PRNMASKNUMBER],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][SCALEEXPONENT],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E11],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E22],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E33],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E44],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E12],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E13],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E14],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E23],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E24],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E34],options->pegasusfs);
						} else {
							sprintf(aux2,"%s%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c",aux,
								sbasblock->clockephemeriscovariance[i][PRNMASKNUMBER],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][SCALEEXPONENT],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E11],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E22],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E33],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E44],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E12],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E13],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E14],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E23],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E24],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E34],options->pegasusfs);
						}
					} else {
						if(options->pegasusaligned==1) {
							sprintf(aux2,"%s%-2d%c      %-d%c     %-3d%c    %-3d%c    %-3d%c    %-3d%c    %-4d%c   %-4d%c   %-4d%c   %-4d%c   %-4d%c   %-4d",aux,
								sbasblock->clockephemeriscovariance[i][PRNMASKNUMBER],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][SCALEEXPONENT],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E11],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E22],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E33],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E44],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E12],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E13],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E14],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E23],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E24],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E34]);
						} else {
							sprintf(aux2,"%s%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d%c%d",aux,
								sbasblock->clockephemeriscovariance[i][PRNMASKNUMBER],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][SCALEEXPONENT],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E11],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E22],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E33],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E44],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E12],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E13],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E14],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E23],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E24],options->pegasusfs,
								sbasblock->clockephemeriscovariance[i][E34]);
						}
					}
				}
				sprintf(aux,"%s",aux2);					//Copy to aux
			}

			ret=fprintf(fdlist[fdSERVICEREGION28],"%s\n",aux);
				if(ret<0) {
						fclose(readfd);         //Close RINEXB or EMS file where the data has been read
						closefiles(fdlist,18);  //Close all other files before exiting
						sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdSERVICEREGION28]);
						printError(aux,options);
				}

			break;

		case(GEOSATELLITEALMANACS):
			//Print SBAS Almanac (message type 17)
			for(i=0;i<sbasblock->numgeoalmanacs;i++) {
				aux[0]='\0';		//Initialize aux to empty
				sprintf(aux2,"%s*** Week %4d almanac for PRN-%3d *******\n",aux,GPSweek,(int)sbasblock->geoalmanacsmessage[i][PRNNUMBER]);
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sRX_WEEK               : %-4d\n",aux,GPSweek);
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sRX_TOM                : %-9.2f\n",aux,SoW);
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sPRN                   : %-3d\n",aux,sbasblock->PRN);
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sDATA_PRN              : %-3d\n",aux,(int)sbasblock->geoalmanacsmessage[i][PRNNUMBER]);
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sDATA_ID               : %-d\n",aux,(int)sbasblock->geoalmanacsmessage[i][DATAID]);
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sHEALTH                : %-3d\n",aux,(int)sbasblock->geoalmanacsmessage[i][HEALTHPROVIDER]);
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sDAYTIME_APPLICABILITY : %-5d\n",aux,(int)sbasblock->geoalmanacsmessage[i][T0ALMANAC]);
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sDAY_OF_WEEK           : %-d\n",aux,1+(int)SoW/86400);	//Day of week goes from 1 to 7
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sWEEK_APPLICABILITY    : %-4d\n",aux,GPSweek);
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sXGEO                  : %19.12E\n",aux,sbasblock->geoalmanacsmessage[i][XGALMANAC]);
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sYGEO                  : %19.12E\n",aux,sbasblock->geoalmanacsmessage[i][YGALMANAC]);
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sZGEO                  : %19.12E\n",aux,sbasblock->geoalmanacsmessage[i][ZGALMANAC]);
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sXGEO_RATE             : %19.12E\n",aux,sbasblock->geoalmanacsmessage[i][XGALMANACRATEOFCHANGE]);
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sYGEO_RATE             : %19.12E\n",aux,sbasblock->geoalmanacsmessage[i][YGALMANACRATEOFCHANGE]);
				sprintf(aux,"%s",aux2);                                 //Copy to aux
				sprintf(aux2,"%sZGEO_RATE             : %19.12E\n",aux,sbasblock->geoalmanacsmessage[i][ZGALMANACRATEOFCHANGE]);
				sprintf(aux,"%s",aux2);                                 //Copy to aux

				ret=fprintf(fdlist[fdALMANAC],"%s\n",aux);
					if(ret<0) {
							fclose(readfd);         //Close RINEXB or EMS file where the data has been read
							closefiles(fdlist,18);  //Close all other files before exiting
							sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[fdALMANAC]);
							printError(aux,options);
					}

				//Save the last Health status read
				lasthealthlist[(int)sbasblock->geoalmanacsmessage[i][PRNNUMBER]-120]=sbasblock->geoalmanacsmessage[i][HEALTHSTATUS];
			}

			break;

		case(GEONAVIGATIONMESSAGE):
			//Print Navigation data in Rinex 2.1 format (message type 9)
			/*Rinex v2.1 specifications (from http://igscb.jpl.nasa.gov/igscb/data/format/rinex211.txt)

			  The "Transmission Time of Message" (PRN / EPOCH / SV CLK header record) is
			  expressed in GPS seconds of the week. It marks the beginning of the message
			  transmission. It has to refer to the same GPS week as the "Epoch of
			  Ephemerides". It has to be adjusted by - or + 604800 seconds, if necessary
			  (which would make it lower than zero or larger than 604800, respectively).
			  It is a redefinition of the Version 2.10 "Message frame time".

			 "Health" shall be defined as follows:

			  - Bits 0 to 3 equal to Health in Message Type 17 (MT17)
			  - bit 4 is set to 1 if MT17 health is unavailable
			  - bit 5 is set to 1 if the URA index is equal to 15

			  In the SBAS message definitions bit 3 of the health is currently marked as
			  'reserved'.  
			  In case of bit 4 set to 1, it is recommended to set bits 0,1,2,3 to 1, too.

			"User Range Accuracy" (URA):

			  The same convention for converting the URA index to meters is used as
			  with GPS. Set URA = 32767 meters if URA index = 15.

			"IODN" (Issue Of Data Navigation)

			  The IODN is defined as the 8 first bits after the message type 9,
			  called IODN in RTCA DO229, Annex A and Annex B and called "spare" in Annex C.*/




			//Get the calendar time in transmission time (we have to take out the travelling time and the time to receive all the message)
			transtime.MJDN=sbasblock->t.MJDN;
			transtime.SoD=sbasblock->t.SoD-1.1;

			if(transtime.SoD<0) {
				transtime.SoD+=86400;
				transtime.MJDN--;
			}

			//Save the first day of ephemerides read in the file
			if(firstbrdcday==-1) firstbrdcday=sbasblock->t.MJDN;

			//Get the time of ephemerides
			toe.MJDN=transtime.MJDN;
			toe.SoD=sbasblock->geonavigationmessage[T0NAV];
			

			if(transtime.MJDN<sbasblock->t.MJDN && sbasblock->t.MJDN<=firstbrdcday && (int)transtime.SoD==86399 && sourcefile==1) {
				//This is the special case when in an EMS file, a navigation message is given in SoD 0, so the transmission time and toe are from previous day
				previousday=1;
			//Check if time given if toe (in seconds of day) is greater than the seconds of day of the transmission time
			} else if((int)(toe.SoD-transtime.SoD)>40000) {
				//SoD of toe if greater than transmission time SoD, so the toe refers to the previous day
				toe.MJDN--;
				//We also have to set previousday=1 to indicate that we have to write data to the file of previous day,
				//If the file contains more than one day of data, the following days will be written in the current day file, not the previous one.
				if(sbasblock->t.MJDN<=firstbrdcday) {
					previousday=1;	
				}
			} else if (transtime.SoD>80000 && toe.SoD<3600) {
				//If transmission time is close to the end of the day and toe is very small, this means that toe refers to the next day
				//The limits imposed in this if statement are arbitrary
				toe.MJDN++;
			}

			
			//Get the calendar time for toe (We use t2tm without localtime due to sometimes there was an hour shift)
			t2tmnolocal(&toe,&tm,&seconds);

			//Calculate GPS Week and Seconds of Week for time of ephemeris
			ttime2gpswsnoroll(&toe, &GPSweektoe, &SoWtoe);

			//Calculate GPS Week and Seconds of Week for transmission time
			ttime2gpswsnoroll(&transtime, &GPSweektrans, &SoWtrans);

			//Check if GPS Week of toe is the same for GPS Week of transmission time
			if(GPSweektoe<GPSweektrans) {
				//GPS Week of toe is smaller than GPS Week of transmission time (due to toe referred to the previous day)
				//We have to substract 604800 seconds so the transmission time refers to the same week as the toe
				transtimesecadjusted=SoWtrans-604800;
			} else {
				transtimesecadjusted=SoWtrans;
			}

			//Set health
			health=lasthealthlist[sbasblock->PRN-120];

			//Get URA value. If URA Index is 15, set to 32767 meters (according to RINEX v2.11) and set bit 5 of health to 1 (add 32)
			uravalue=sbasblock->geonavigationmessage[URANOMINALVALUE];
			if(uravalue==-1) {
				uravalue=32767;
				health+=32;
			}

			//In B2A, if we don't have the health status put zero. Else just put the last four bits (LSB) of the health status bits
			if(options->pegasuspurerinex==0) {
				if(lasthealthlist[sbasblock->PRN-120]==31) {
					//health=0;
					health=1;
				} else {
					health=lasthealthlist[sbasblock->PRN-120];
				}
			}

			//Check in what file do we have to print
			if(previousday) {		
				filetoprint=fdEPHEMERISPREVDAY;
				*prevday=1;	//Set to 1 to indicate that we have written a RINEX H file from the previous day
			} else {
				filetoprint=fdEPHEMERIS;
			}

			//Print data
			if(options->pegasuspurerinex==0) {
				//Write the file the same way as Pegasus converter
				ret=fprintf(fdlist[filetoprint],"%02d %02d %2d %2d %2d %2d%5.1f%19.12E%19.12E%19.12E\n   %19.12E%19.12E%19.12E%19.12E\n   %19.12E%19.12E%19.12E%19.12E\n   %19.12E%19.12E%19.12E%19.12E\n", 				 	 sbasblock->PRN%100,tm.tm_year%100,tm.tm_mon+1,tm.tm_mday,tm.tm_hour,tm.tm_min,seconds,
						sbasblock->geonavigationmessage[AGF0],
						sbasblock->geonavigationmessage[AGF1],
							SoWtrans, /*Time in Transmission time*/ 
							//transtimesecadjusted, /*Time in Transmission time, referring to the same week as toe*/ 
						sbasblock->geonavigationmessage[XG]/1000,
						sbasblock->geonavigationmessage[XGRATEOFCHANGE]/1000,
						sbasblock->geonavigationmessage[XGACCELERATION]/1000,
						health, 
						sbasblock->geonavigationmessage[YG]/1000,
						sbasblock->geonavigationmessage[YGRATEOFCHANGE]/1000,
						sbasblock->geonavigationmessage[YGACCELERATION]/1000,
						uravalue,	/*URA in meters*/
						sbasblock->geonavigationmessage[ZG]/1000,
						sbasblock->geonavigationmessage[ZGRATEOFCHANGE]/1000,
						sbasblock->geonavigationmessage[ZGACCELERATION]/1000,
						sbasblock->geonavigationmessage[IODN]);
			} else {
				ret=fprintf(fdlist[filetoprint],"%02d %02d %2d %2d %2d %2d%5.1f%19.12E%19.12E%19.12E\n   %19.12E%19.12E%19.12E%19.12E\n   %19.12E%19.12E%19.12E%19.12E\n   %19.12E%19.12E%19.12E%19.12E\n", 				 	 sbasblock->PRN%100,tm.tm_year%100,tm.tm_mon+1,tm.tm_mday,tm.tm_hour,tm.tm_min,seconds,
						sbasblock->geonavigationmessage[AGF0],
						sbasblock->geonavigationmessage[AGF1],
							transtimesecadjusted, /*Time in Transmission time, referring to the same week as toe*/ 
						sbasblock->geonavigationmessage[XG]/1000,
						sbasblock->geonavigationmessage[XGRATEOFCHANGE]/1000,
						sbasblock->geonavigationmessage[XGACCELERATION]/1000,
						health, 
						sbasblock->geonavigationmessage[YG]/1000,
						sbasblock->geonavigationmessage[YGRATEOFCHANGE]/1000,
						sbasblock->geonavigationmessage[YGACCELERATION]/1000,
						uravalue,	/*URA in meters*/
						sbasblock->geonavigationmessage[ZG]/1000,
						sbasblock->geonavigationmessage[ZGRATEOFCHANGE]/1000,
						sbasblock->geonavigationmessage[ZGACCELERATION]/1000,
						sbasblock->geonavigationmessage[IODN]);
			}

				if(ret<0) {
						fclose(readfd);         //Close RINEXB or EMS file where the data has been read
						closefiles(fdlist,18);  //Close all other files before exiting
						sprintf(aux,"When writing SBAS data to Pegasus file [%s]\n",filelist[filetoprint]);
						printError(aux,options);
				}
			break;

		default:
			//Messages with no data or that do no have any effect. No need to print anything
			//Message WAAS Network Time/UTC/GLONASS is not read by Pegasus
			break;
	}
}

/*****************************************************************************
 * Name        : printSBASCORR
 * Description : Prints the SBASCORR data line for SBAS
 * Parameters  :
 * Name                                    |Da|Unit|Description
 * TEpoch  *epoch                           I  N/A  Reference epoch
 * TSBASdata *SBASdata                      I  N/A  Struct with SBAS data
 * TSBAScorr *SBAScorr                      I  N/A  Struct with the SBAS corrections to be applied
 * int  satIndex                            I  N/A  Selected satellite
 * int  measIndex                           I  N/A  Selected measurement
 * double  flightTime                       I    s  Signal flight time
 * double  geometricDistance                I    m  Geometric distance between satellite and receiver
 * double  satelliteClockModel              I    m  Satellite clock error
 * double  satellitePhaseCenterProjection   I    m  Projection of the satellite phase center into the LoS
 * double  receiverPhaseCenterProjection    I    m  Projection of the receiver phase center into the LoS 
 * double  receiverARPProjection            I    m  Projection of the receiver ARP into the LoS
 * double  relativityModel                  I    m  Relativity model
 * double  windUpMetModel                   I    m  Wind up model (for carrier phase only)
 * double  troposphereModel                 I    m  Troposphere model
 * double  ionosphereModel                  I    m  Ionosphere model
 * double  gravitationalDelayModel          I    m  Gravitational delay model
 * double  TGDModel                         I    m  TGD model
 * double  solidTidesModel                  I    m  Projection of the effect of solid tides into the LoS
 * TOptions  *options                       I  N/A  TOptions structure
 *****************************************************************************/
void printSBASCORR (TEpoch *epoch, TSBASdata *SBASdata, TSBAScorr *SBAScorr, int satIndex, int measIndex, double flightTime, double geometricDistance, double satelliteClockModel, double satellitePhaseCenterProjection, double receiverPhaseCenterProjection, double receiverARPProjection, double relativityModel, double windUpMetModel, double troposphereModel, double ionosphereModel, double gravitationalDelayModel, double TGDModel, double solidTidesModel, TOptions *options) {
	int			i,j;
	//int		ind;
	int			ret;
	int			GPSweek;
	int			IODE;
	double		SoW;
	double		measurement;
	double		model,prefit;
	double		smoothWith,smoothedVal;
	static int	initialized=0;
	static char	receiverid[5];

	//Calculate GPS Week and Seconds of Week
	ttime2gpswsnoroll(&epoch->t, &GPSweek, &SoW);

	j = satIs(epoch,epoch->sat[satIndex].PRN,epoch->sat[satIndex].GNSS);

	if(initialized==0) {
		initialized=1;
		//Convert receiver name to lower case
		for(i=0;i<4;i++) {
			receiverid[i]=tolower(epoch->receiver.name[i]);
		}
		receiverid[4]='\0';
	}


	ret = getMeasModelValue(epoch,epoch->sat[satIndex].GNSS,epoch->sat[satIndex].PRN,epoch->measOrder[epoch->sat[satIndex].GNSS].ind2Meas[measIndex],&measurement,&model);
	if (ret!=1) {
		return;
	}

	if(SBAScorr->IOD!=-1) {
		IODE=SBAScorr->IOD;
	} else if (SBAScorr->IODnoSBAS!=-1) {
		IODE=-SBAScorr->IODnoSBAS;
	} else {
		IODE=999;
	}
		


	//ind = epoch->measOrder[GPS].meas2Ind[C1C];

	// Smoothing
	if (options->smoothEpochs!=0 && options->smoothMeas[0]!=NA) {
		ret = getMeasModelValue(epoch,epoch->sat[satIndex].GNSS,epoch->sat[satIndex].PRN,options->smoothMeas[0],&smoothWith,NULL);
		if (ret!=1) {
			return;
		}
		prefit = epoch->cycleslip.smoothedMeas[j][0] + smoothWith - model;
		smoothedVal = epoch->cycleslip.smoothedMeas[j][0] + smoothWith;
	} else {
		prefit = measurement - model;
		//smoothedVal=epoch->sat[satIndex].meas[ind].value;
		smoothedVal=measurement;
	}

	sprintf(printbuffer[satIndex][linesstored[satIndex]],"SBASCORR    %4s %3s %3s %3d %17s %4d %9.2f %3d %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %7.5f %14.4f %14.4f %7.1f %7.1f %7.1f %14.4f %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %2d %2d %2d %2d %2d %4d %2d %2d %2d %2d %8.3f %8.3f\n",
	/*2*/		receiverid,
	/*3*/		SBASnavmode2SBASnavstr(options->precisionapproach),
	/*4*/		gnsstype2gnssstr(epoch->sat[satIndex].GNSS),
	/*5*/		epoch->sat[satIndex].PRN,
	/*6*/		t2doystr(&epoch->t), 		//Year
	/*7*/						//DoY
	/*8*/						//Seconds of day
	/*9*/		GPSweek,
	/*10*/		SoW,
	/*11*/		SBASdata[options->GEOindex].PRN,
	/*12*/		prefit,
	/*13*/		measurement,			//C1 raw + user added noise (in meters)
	/*14*/		smoothedVal,			//C1 smooothed + user added noise (if smoothing is disabled, then is C1 raw + user added noise) (in meters)
	/*15*/		geometricDistance, 		//With SBAS corrections (in meters)
	/*16*/		relativityModel,		//With SBAS corrections (in meters)
	/*17*/		satelliteClockModel, 		//With SBAS corrections (in meters)
	/*18*/		TGDModel,			//From navigation file (in meters)
	/*19*/		epoch->IPPlat,			//Ionospheric pierce point latitude (-90..90º)
	/*20*/		epoch->IPPlon,			//Ionospheric pierce point longitude (0..360º)
	/*21*/		ionosphereModel,		//In meters
	/*22*/		troposphereModel,		//In meters
	/*23*/		SBAScorr->PRC,	 		//PRC Fast Correction applied to the satellite (in meters)
	/*24*/		SBAScorr->RRCtime,		//RRC Fast Correction applied to the satellite (in meters)
	/*25*/		SBAScorr->fastcorrdegfactor,	//Fast correction degradation factor (in meters/seconds^2)
	/*26*/		SBAScorr->fastcorrtmout,	//PRC time out (in seconds)
	/*27*/		SBAScorr->fastcorrsmallertmout,	//RRC time out (smallest time out for all satellites) (in seconds)
	/*28*/		SBAScorr->tPRC,			//Time (seconds of day) used for computing PRC timeout
	/*29*/		SBAScorr->tudre,		//Time (seconds of day) used for computing sigma UDRE timeout
	/*30*/		SBAScorr->tu,			//Time (seconds of day) used for computing fast correction degradation
	/*31*/		SBAScorr->SatCoordbrdc[0]+SBAScorr->SatCoordCor[0],	//X component of the satellite position in WGS84 system at emission time with SBAS corrections (in meters) 
	/*32*/		SBAScorr->SatCoordbrdc[1]+SBAScorr->SatCoordCor[1],	//Y component of the satellite position in WGS84 system at emission time with SBAS corrections (in meters)
	/*33*/		SBAScorr->SatCoordbrdc[2]+SBAScorr->SatCoordCor[2],	//Z component of the satellite position in WGS84 system at emission time with SBAS corrections (in meters)
	/*34*/		SBAScorr->SatCoordCor[0],	//Long term correction to be applied to the X component of the satellite (in meters)
	/*35*/		SBAScorr->SatCoordCor[1],	//Long term correction to be applied to the Y component of the satellite (in meters)
	/*36*/		SBAScorr->SatCoordCor[2],	//Long term correction to be applied to the Z component of the satellite (in meters)
	/*37*/		SBAScorr->ClockCor,		//Long term correction to be applied to the satellite clock (in meters)
	/*38*/		SBAScorr->fastcorIODP,		//IODP used for fast corrections
	/*39*/		SBAScorr->fastcorIODF,		//IODF in messages type 2-5,24 for fast corrections
	/*40*/		SBAScorr->fastcorPRNpos+1,	//Satellite position in the fast correction mask (1..51)
	/*41*/		SBAScorr->longtermIODP,		//IODP used for long term corrections
	/*42*/		SBAScorr->longtermPRNpos+1,	//Satellite position in the long term correction mask (1..51)
	/*43*/		IODE,				//IODE used for broadcast ephemeris
	/*44*/		SBAScorr->serviceIODS,		//IODS used for service message
	/*45*/		SBAScorr->covarianceIODP,	//IODP used for Covariance matrix
	/*46*/		SBAScorr->covariancePRNpos+1,	//Satellite position in the covariance mask (1..51)
	/*47*/		SBAScorr->ionomodelflag,        //Flag to indicate which iono model is used (-1->no model, 0->SBASiono, 1->Klobuchar, 2->Other model)	
	/*48*/		epoch->sat[satIndex].elevation/d2r, //Elevation (in degrees)
	/*49*/		epoch->sat[satIndex].azimuth/d2r);  //Azimuth (in degrees)

	//Check if satellite has reached steady-state operation. If not, put an '*' to indicate that it will not be used
	if(options->excludeSmoothingConvergence>0 && epoch->cycleslip.arcLength[j] < options->excludeSmoothingConvergence) {
		//Steady state not reached
		printbuffer[satIndex][linesstored[satIndex]][8]='*';
	}

	if(options->printInBuffer==1) {
		linesstored[satIndex]++;
	} else {
		printf("%s",printbuffer[satIndex][linesstored[satIndex]]);
		printbuffer[satIndex][linesstored[satIndex]][0]='\0';
	}
}

/*****************************************************************************
 * Name        : printSBASVAR
 * Description : Prints the SBASVAR data line for SBAS
 * Parameters  :
 * Name                                    |Da|Unit|Description
 * TEpoch  *epoch                           I  N/A  Reference epoch
 * TSBASdata *SBASdata                      I  N/A  Struct with SBAS data
 * TSBAScorr *SBAScorr                      I  N/A  Struct with the SBAS corrections to be applied
 * int  satIndex                            I  N/A  Selected satellite
 * int  measIndex                           I  N/A  Selected measurement
 * TOptions  *options                       I  N/A  TOptions structure
 *****************************************************************************/
void printSBASVAR (TEpoch *epoch, TSBASdata *SBASdata, TSBAScorr *SBAScorr, int satIndex, int measIndex, TOptions *options) {
	int     	i,j;
	int     	ind;
	int     	ret;
	int     	GPSweek;
	double  	SoW;
	double  	measurement;
	double  	model;
	double 		smoothWith;
	static int	initialized=0;
	static char	receiverid[5];

	//Calculate GPS Week and Seconds of Week
	ttime2gpswsnoroll(&epoch->t, &GPSweek, &SoW);

	j = satIs(epoch,epoch->sat[satIndex].PRN,epoch->sat[satIndex].GNSS);

	if(initialized==0) {
		initialized=1;
		//Convert receiver name to lower case
		for(i=0;i<4;i++) {
			receiverid[i]=tolower(epoch->receiver.name[i]);
		}
		receiverid[4]='\0';
	}

	ret = getMeasModelValue(epoch,epoch->sat[satIndex].GNSS,epoch->sat[satIndex].PRN,epoch->measOrder[epoch->sat[satIndex].GNSS].ind2Meas[measIndex],&measurement,&model);
	if (ret!=1) {
			return;
	}

	ind = epoch->measOrder[GPS].meas2Ind[C1C];

	// Smoothing
	if (options->smoothEpochs!=0 && options->smoothMeas[ind]!=NA) {
			ret = getMeasModelValue(epoch,epoch->sat[satIndex].GNSS,epoch->sat[satIndex].PRN,options->smoothMeas[ind],&smoothWith,NULL);
			if (ret!=1) {
					return;
			}
	}

	sprintf(printbuffer[satIndex][linesstored[satIndex]],"SBASVAR     %4s %3s %3s %3d %17s %4d %9.2f %3d %14.4f %14.4f %14.4f %8.3f %3d %14.4f %14.4f %14.4f %14.4f %1d %14.4f %14.4f %14.4f %8.3f %8.3f %8.3f %8.3f %8.3f\n",
	/*2*/		receiverid,
	/*3*/		SBASnavmode2SBASnavstr(options->precisionapproach),
	/*4*/		gnsstype2gnssstr(epoch->sat[satIndex].GNSS),
	/*5*/		epoch->sat[satIndex].PRN,
	/*6*/		t2doystr(&epoch->t), 		//Year
	/*7*/						//DoY
	/*8*/						//Seconds of day
	/*9*/		GPSweek,
	/*10*/           SoW,
	/*11*/          SBASdata[options->GEOindex].PRN,
	/*12*/		sqrt(SBAScorr->SBASsatsigma2),					//Sigma of the total residual error associated to the satellite (in meters)
	/*13*/		sqrt(SBAScorr->sigma2flt),					//Sigma of the residual error associated to the fast and long-term corrections (in meters)
	/*14*/		sqrt(SBAScorr->UDREsigma),					//Sigma of the UDRE (in meters)
	/*15*/		SBAScorr->deltaudre,						//Delta UDRE factor
	/*16*/		SBAScorr->deltaudresource,					//Indicates the data source for Delta UDRE(27 or 28 for their respective message type,-27 or -28 if received any of these message types but there was missing data or was timed out, 0 if no message type received or both received)
	/*17*/		SBAScorr->fastcorrdeg,						//Degradation parameter for fast correction data (in meters)
	/*18*/		SBAScorr->RRCdeg,						//Degradation parameter for range rate correction data (in meters)
	/*19*/		SBAScorr->longtermdegradation+SBAScorr->geonavdegradation,	//Degradation parameter for long term correction data or GEO navigation message data (in meters)
	/*20*/		SBAScorr->enroutedegradation,					//Degradation parameter for en route through NPA applications (in meters)
	/*21*/	       	(int)SBASdata[options->GEOindex].degradationfactors[RSSUDRE],	//Root-sum-square flag
	/*22*/		sqrt(SBAScorr->sigma2iono),					//Sigma of the residual error associated to the ionosheric corrections (in meters)
	/*23*/		sqrt(SBAScorr->sigma2tropo),					//Sigma of the residual error associated to the tropospheric corrections (in meters)
	/*24*/		sqrt(SBAScorr->sigma2air),					//Sigma of the total receiver error (in meters)
	/*25*/		sqrt(SBAScorr->sigma2noise),					//Sigma of the receiver noise (in meters)
	/*26*/		sqrt(SBAScorr->sigma2multipath),				//Sigma of the receiver multipath (in meters)
	/*27*/		sqrt(SBAScorr->sigma2divergence),				//Sigma of the receiver divergence (in meters)
	/*28*/		epoch->sat[satIndex].elevation/d2r,				//Elevation (in degrees)
	/*29*/		epoch->sat[satIndex].azimuth/d2r);				//Azimuth (in degrees)

	//Check if satellite has reached steady-state operation. If not, put an '*' to indicate that it will not be used
	if(options->excludeSmoothingConvergence>0 && epoch->cycleslip.arcLength[j]<options->excludeSmoothingConvergence) {
		//Steady state not reached
		printbuffer[satIndex][linesstored[satIndex]][7]='*';
	}

	if(options->printInBuffer==1) {
		linesstored[satIndex]++;
	} else {
		printf("%s",printbuffer[satIndex][linesstored[satIndex]]);
		printbuffer[satIndex][linesstored[satIndex]][0]='\0';
	}
}

/*****************************************************************************
 * Name        : printSBASIONO
 * Description : Prints the SBASIONO data line for SBAS
 * Parameters  :
 * Name                                    |Da|Unit|Description
 * TEpoch  *epoch                           I  N/A  Reference epoch
 * TSBASdata *SBASdata                      I  N/A  Struct with SBAS data
 * TSBAScorr *SBAScorr                      I  N/A  Struct with the SBAS corrections to be applied
 * int  satIndex                            I  N/A  Selected satellite
 * int  InterpolationMode                   I  N/A  Interpolation mode
 *                                                  0 -> Square
 *                                                  [1-4] -> Vertex not used in interpolation
 * int  IGP1                                I  N/A  IGP number for upper right corner 
 * int  BandNumber1                         I  N/A  Band number for upper right corner
 * double IGPlat1                           I    º  Latitude for upper right corner
 * double IGPlon1                           I    º  Longitude for upper right corner
 * int  IGP2                                I  N/A  IGP number for upper left corner 
 * int  BandNumber2                         I  N/A  Band number for upper left corner 
 * double IGPlat2                           I    º  Latitude for upper left corner
 * double IGPlon2                           I    º  Longitude for upper left corner
 * int  IGP3                                I  N/A  IGP number for lower right corner 
 * int  BandNumber3                         I  N/A  Band number for lower right corner 
 * double IGPlat3                           I    º  Latitude for lower right corner
 * double IGPlon3                           I    º  Longitude for lower right corner
 * int  IGP4                                I  N/A  IGP number for lower left corner 
 * int  BandNumber4                         I  N/A  Band number for lower left corner 
 * double IGPlat4                           I    º  Latitude for lower left corner
 * double IGPlon4                           I    º  Longitude for lower left corner
 * double  eps_iono[5]                      I    m  Degradation number for the four IGPs
 * double  IGPDelays[5]                     I    m  IGP delays used for interpolation for the four IGPs
 * double  IGPSigma2[5]                     I  m^2  IGP sigma used for interpolation for the four IGPs
 * double  IPPWeight[5]                     I    m  IGP weight for the four IGPs
 * double  Fpp				    I  N/A  Mapping function 
 * double  SlantDelay                       I    m  Slant Delay
 * double  SlantSigma                       I    m  Slant Sigma
 * double  elevation			    I  rad  Satellite elevation
 * double  azimuth                          I  rad  Satelite azimuth
 * TOptions  *options                       I  N/A  TOptions structure
 *****************************************************************************/
void printSBASIONO (TEpoch  *epoch, TSBASdata *SBASdata, TSBAScorr *SBAScorr, int  satIndex, int  InterpolationMode, int  IGP1, int  BandNumber1, double IGPlat1, double IGPlon1, int  IGP2, int  BandNumber2, double IGPlat2, double IGPlon2, int  IGP3, int  BandNumber3, double IGPlat3, double IGPlon3, int  IGP4, int  BandNumber4, double IGPlat4, double IGPlon4, double  eps_iono[5], double  IGPDelays[5], double  IGPSigma2[5], double  IPPWeight[5], double  Fpp, double  SlantDelay, double  SlantSigma, double elevation, double azimuth, TOptions  *options) {
	int		i,j;
	int     	GPSweek;
	double		SoW;
	static int	initialized=0;
	static char	receiverid[5];

		//Calculate GPS Week and Seconds of Week
		ttime2gpswsnoroll(&epoch->t, &GPSweek, &SoW);

		j = satIs(epoch,epoch->sat[satIndex].PRN,epoch->sat[satIndex].GNSS);

	if(initialized==0) {
		initialized=1;
		//Convert receiver name to lower case
		for(i=0;i<4;i++) {
			receiverid[i]=tolower(epoch->receiver.name[i]);
		}
		receiverid[4]='\0';
	}

		/* Four points:
	
				  NORTH	   
		(IGPlat2,IGPlon1)   (IGPlat2,IGPlon2)
		  W2=(1-x)*y          W1=x*y
			  v2..............v1
			  .               .
			  .               .
			  .               .
	 WEST     .   IPP         .  EAST
			  .     *         .
			  .     |         .
			  .     |         .
			  v3____..........v4
		  W3=(1-x)*(1-y)      W4=x*(1-y)
		(IGPlat1,IGPlon1)   (IGPlat1,IGPlon2)
				  SOUTH	   
	*/

	sprintf(printbuffer[satIndex][linesstored[satIndex]],"SBASIONO    %4s %3s %3s %3d %17s %4d %9.2f %3d %8.3f %8.3f %d %d %2d %3d %7.1f %3d %3d %6.3f %8.4f %6.3f %8.4f %8.4f %6.4f %d %2d %3d %7.1f %3d %3d %6.3f %8.4f %6.3f %8.4f %8.4f %6.4f %d %2d %3d %7.1f %3d %3d %6.3f %8.4f %6.3f %8.4f %8.4f %6.4f %d %2d %3d %7.1f %3d %3d %6.3f %8.4f %6.3f %8.4f %8.4f %6.4f %8.4f %8.4f %8.4f %5.2f %5.2f\n",
	/*2*/		receiverid,
	/*3*/		SBASnavmode2SBASnavstr(options->precisionapproach),
	/*4*/		gnsstype2gnssstr(epoch->sat[satIndex].GNSS),
	/*5*/		epoch->sat[satIndex].PRN,
	/*6*/		t2doystr(&epoch->t), 		//Year
	/*7*/						//DoY
	/*8*/						//Seconds of day
	/*9*/		GPSweek,
	/*10*/		SoW,
	/*11*/		SBASdata[options->GEOindex].PRN,
	/*12*/		epoch->IPPlat,								//Ionospheric pierce point latitude (-90..90º)	
	/*13*/		epoch->IPPlon,								//Ionospheric pierce point longitude (0..360º)
	/*14*/		InterpolationMode,							//If 0, square interpolation. If [1-4] indicates the vertex not used
	/*15*/		SBASdata[options->GEOindex].IGP2Mask[BandNumber1][IGP1],		//IODI for vertex 1
	/*16*/		BandNumber1,								//Band Number for vertex 1
	/*17*/		IGP1,									//IGP Number for vertex 1
	/*18*/		SBASdata[options->GEOindex].lastiono[BandNumber1][IGP1].SoD,		//Time of reception of last bit of vertex 1 IGP (in seconds of day)
	/*19*/		(int)IGPlat1,								//Latitude of the IGP for vertex 1 (-90..90º)
	/*20*/		(int)IGPlon1,								//Longitude of the IGP for vertex 1 (0..360º)
	/*21*/		SBASdata[options->GEOindex].Ionodelay[BandNumber1][IGP1],		//Iono delay (raw value from MT26) for vertex 1 (meters)
	/*22*/		SBASdata[options->GEOindex].Ionosigma[BandNumber1][IGP1],		//Iono sigma (raw value from MT26) for vertex 1 (meters^2)
	/*23*/		eps_iono[1],								//Degradation term for vertex 1 (meters)
	/*24*/		IGPDelays[1],								//Iono delay after interpolation (if required) for vertex 1 (meters)
	/*25*/		IGPSigma2[1],								//Iono sigma after applying degradation and interpolation (if lat=85º) for vertex 1 (meters^2)
	/*26*/		IPPWeight[1],								//Interpolation weight for vertex 1
	/*27*/		SBASdata[options->GEOindex].IGP2Mask[BandNumber2][IGP2],
	/*28*/		BandNumber2,
	/*29*/		IGP2,
	/*30*/		SBASdata[options->GEOindex].lastiono[BandNumber2][IGP2].SoD,
	/*31*/		(int)IGPlat2,
	/*32*/		(int)IGPlon2,
	/*33*/		SBASdata[options->GEOindex].Ionodelay[BandNumber2][IGP2],
	/*34*/		SBASdata[options->GEOindex].Ionosigma[BandNumber2][IGP2],
	/*35*/		eps_iono[2],
	/*36*/		IGPDelays[2],
	/*37*/		IGPSigma2[2],
	/*38*/		IPPWeight[2],
	/*39*/		SBASdata[options->GEOindex].IGP2Mask[BandNumber3][IGP3],
	/*40*/		BandNumber3,
	/*41*/		IGP3,
	/*42*/		SBASdata[options->GEOindex].lastiono[BandNumber3][IGP3].SoD,
	/*43*/		(int)IGPlat3,
	/*44*/		(int)IGPlon3,
	/*45*/		SBASdata[options->GEOindex].Ionodelay[BandNumber1][IGP3],
	/*46*/		SBASdata[options->GEOindex].Ionosigma[BandNumber1][IGP3],
	/*47*/		eps_iono[3],
	/*48*/		IGPDelays[3],
	/*49*/		IGPSigma2[3],
	/*50*/		IPPWeight[3],
	/*51*/		SBASdata[options->GEOindex].IGP2Mask[BandNumber4][IGP4],
	/*52*/		BandNumber4,
	/*53*/		IGP4,
	/*54*/		SBASdata[options->GEOindex].lastiono[BandNumber4][IGP4].SoD,
	/*55*/		(int)IGPlat4,
	/*56*/		(int)IGPlon4,
	/*57*/		SBASdata[options->GEOindex].Ionodelay[BandNumber4][IGP4],
	/*58*/		SBASdata[options->GEOindex].Ionosigma[BandNumber4][IGP4],
	/*59*/		eps_iono[4],
	/*60*/		IGPDelays[4],
	/*61*/		IGPSigma2[4],
	/*62*/		IPPWeight[4],
	/*63*/		Fpp,									//Mapping function
	/*64*/		SlantDelay,								//Slant delay (meters of L1)
	/*65*/		SlantSigma,								//Slant Sigma (meters of L1)
	/*66*/		elevation/d2r,								//Elevation (in degrees)
	/*67*/		azimuth/d2r);								//Azimuth (in degrees)

	//Check if satellite has reached steady-state operation. If not, put an '*' to indicate that it will not be used
	if(options->excludeSmoothingConvergence>0 && epoch->cycleslip.arcLength[j] < options->excludeSmoothingConvergence) {
		//Steady state not reached
		printbuffer[satIndex][linesstored[satIndex]][8]='*';
	}

	if(options->printInBuffer==1) {
		linesstored[satIndex]++;
	} else {
		printf("%s",printbuffer[satIndex][linesstored[satIndex]]);
		printbuffer[satIndex][linesstored[satIndex]][0]='\0';
	}
}

/*****************************************************************************
 * Name        : printSBASUNSEL
 * Description : Prints the SBASUNSEL data line for SBAS (printed only when a satellite is discarded)
 * Parameters  :
 * Name                                    |Da|Unit|Description
 * TEpoch  *epoch                           I  N/A  Reference epoch
 * int satIndex                             I  N/A  Selected satellite
 * int errorcode			    I  N/A  Error code indicating the reason for discarding the satellite
 * TSBASdata *SBASdata                      I  N/A  Struct with SBAS data
 * TSBAScorr *SBAScorr                      I  N/A  Struct with the SBAS corrections to be applied
 * TOptions  *options                       I  N/A  TOptions structure
 *****************************************************************************/
void printSBASUNSEL (TEpoch  *epoch, int  satIndex, int errorcode, TSBASdata *SBASdata, TSBAScorr *SBAScorr, TOptions  *options) {

	int				i,j;
	int				GPSweek;
	int				GEOPRN;
	double			SoW;
	char			receiverid[5];
	static int		initialized=0;
	static char 	errorstr[46][MAX_INPUT_LINE];

	//Calculate GPS Week and Seconds of Week
	ttime2gpswsnoroll(&epoch->t, &GPSweek, &SoW);

	j = satIs(epoch,epoch->sat[satIndex].PRN,epoch->sat[satIndex].GNSS);

	//Convert receiver name to lower case
	for(i=0;i<4;i++) {
		receiverid[i]=tolower(epoch->receiver.name[i]);
	}
	receiverid[4]='\0';

	errorcode=abs(errorcode);

	//Check for options->GEOindex value
	if(options->GEOindex==-1) GEOPRN=-1;
	else GEOPRN=SBASdata[options->GEOindex].PRN;

	if(initialized==0) {
		initialized=1;

		sprintf(errorstr[0],"Satellite selected");
		sprintf(errorstr[1],"No GEO satellites available");
		sprintf(errorstr[2],"No data for user selected GEO");
		sprintf(errorstr[3],"Not enough almanac or GEO navigation message to determine the GEO with highest elevation");
		sprintf(errorstr[5],"Received 4 or more consecutive messages with errors");
		sprintf(errorstr[6],"Missed 4 or more consecutive messages");
		sprintf(errorstr[7],"No PRN mask");
		sprintf(errorstr[8],"PRN mask timed out");
		sprintf(errorstr[9],"Satellite is not monitored in any of the PRN mask available");
		sprintf(errorstr[10],"No message type 10 available [PA only]");
		sprintf(errorstr[11],"Message type 10 timed out [PA only]");
		sprintf(errorstr[12],"No fast correction data received for current PRN [PA only]");
		sprintf(errorstr[13],"Sigma UDRE timed out [PA only]");
		sprintf(errorstr[14],"Satellite flagged as 'Not monitored' (UDREI=14)");
		sprintf(errorstr[15],"Satellite flagged as 'Do not use' (UDREI=15)");
		sprintf(errorstr[17],"No fast correction degradation data [PA only]");
		sprintf(errorstr[18],"Fast correction degradation data timed out [PA only]");
		sprintf(errorstr[19],"Last PRC received timed out [PA only]");
		sprintf(errorstr[20],"Only one PRC received. RRC calculation not possible [PA only]");
		sprintf(errorstr[21],"RRC timed out (under alarm condition) due to time difference between PRC used [PA only]");
		sprintf(errorstr[22],"RRC timed out (under alarm condition) due to excessive PRC propagation in time [PA only]");
		sprintf(errorstr[23],"RRC timed out due to time difference between PRC used [PA only]");
		sprintf(errorstr[24],"RRC timed out due to excessive PRC propagation in time [PA only]");
		sprintf(errorstr[25],"Service message timed out [PA only]");
		sprintf(errorstr[26],"Not received a full set of service messages with the same IODS [PA only]");
		sprintf(errorstr[27],"No clock-ephemeris covariance matrix data for current satellite [PA only]");
		sprintf(errorstr[28],"Clock-ephemeris covariance matrix data timed out [PA only]");
		sprintf(errorstr[29],"No navigation data for ranging GEO");
		sprintf(errorstr[30],"Ranging GEO navigation data timed out");
		sprintf(errorstr[32],"No long term correction data for current satellite [PA only]");
		sprintf(errorstr[33],"Long term correction data timed out [PA only]");
		sprintf(errorstr[35],"No broadcast block available for current satellite (regardless of SBAS IOD) [NPA only]");
		sprintf(errorstr[36],"Could not compute transmission time for current PRN measurement");
		sprintf(errorstr[37],"No ionospheric grid mask [PA only]");
		sprintf(errorstr[38],"Ionospheric grid mask timed out [PA only]");
		sprintf(errorstr[39],"IGPs around ionospheric pierce point not found in MOPS grid [PA only]");
		sprintf(errorstr[40],"Not enough IGPs available in ionospheric grid mask [PA only]");
		sprintf(errorstr[41],"One IGP is set as don't use [PA only]");
		sprintf(errorstr[42],"One or more IGPs is set as not monitored or has timed out [PA only]");
		sprintf(errorstr[43],"Data not available for one or more IGPs [PA only]");
		sprintf(errorstr[44],"Ionospheric pierce point is outside triangle [PA only]");
		sprintf(errorstr[45],"External ionosphere model not available");
	}

	if(errorcode==4) {
		if(SBASdata[0].alarmTimeRemaining==1) {
			sprintf(errorstr[4],"Received alarm message for current GEO at epoch %17s. Time remaining to finish alarm: %2d second",t2doystr(&SBASdata[0].alarmtime),SBASdata[0].alarmTimeRemaining);
		} else {
			sprintf(errorstr[4],"Received alarm message for current GEO at epoch %17s. Time remaining to finish alarm: %2d seconds",t2doystr(&SBASdata[0].alarmtime),SBASdata[0].alarmTimeRemaining);
		}
	} else if(errorcode==16) {
		sprintf(errorstr[16],"Satellite has an UDREI value of %2d [PA only]",SBAScorr->UDREI);
	} else if (errorcode==31) {
		sprintf(errorstr[31],"URA index value of %2d for ranging GEO satellite",SBAScorr->uraindex);
	} else if (errorcode==34) {
		sprintf(errorstr[34],"No broadcast block with IOD %d [PA only]",SBAScorr->IOD);
	}

	sprintf(printbuffer[satIndex][linesstored[satIndex]],"SBASUNSEL   %4s %3s %3s %3d %17s %4d %9.2f %3d %2d \"%s\"\n",
	/*2*/		receiverid,
	/*3*/		SBASnavmode2SBASnavstr(options->precisionapproach),
	/*4*/		gnsstype2gnssstr(epoch->sat[satIndex].GNSS),
	/*5*/		epoch->sat[satIndex].PRN,
	/*6*/		t2doystr(&epoch->t),		//Year
	/*7*/						//DoY
	/*8*/						//Seconds of day
	/*9*/		GPSweek,
	/*10*/		SoW,
	/*11*/		GEOPRN,
	/*12*/		errorcode,
	/*13*/		errorstr[errorcode]);

	//Check if satellite has reached steady-state operation. If not, put an '*' to indicate that it will not be used
	if(options->excludeSmoothingConvergence > 0 && epoch->cycleslip.arcLength[j] < options->excludeSmoothingConvergence) {
		//Steady state not reached
		printbuffer[satIndex][linesstored[satIndex]][9]='*';
	}

	if(options->printInBuffer==1) {
		linesstored[satIndex]++;
	} else {
		printf("%s",printbuffer[satIndex][linesstored[satIndex]]);
		printbuffer[satIndex][linesstored[satIndex]][0]='\0';
	}
}
 
/*****************************************************************************
 * Name        : printSBASOUT
 * Description : Prints the solution information for SBAS
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TFilterSolution  *solution      I  N/A  Solution of the filter
 * TSBASdata *SBASdata             I  N/A  Struct with SBAS data
 * TOptions  *options              I  N/A  TOptions structure
 * TUnkinfo  *unkinfo              I  N/A  Information on the unknowns of the filter
 *****************************************************************************/
void printSBASOUT (TEpoch *epoch, TFilterSolution *solution, TSBASdata *SBASdata, TOptions *options, TUnkinfo *unkinfo) {

	int		i,j,k;
	int     	GPSweek;
	double  	SoW;
	static int	initialized=0;
	static char	receiverid[5];

	double		positionNEU[3];
	double		dNEU[3];
	double		sigmaEN=0,dmajor;
	int 		indx;
	
	if(options->printSBASOUT==0) return;

	//Calculate GPS Week and Seconds of Week
	ttime2gpswsnoroll(&epoch->t, &GPSweek, &SoW);


	if(initialized==0) {
		initialized=1;
		//Convert receiver name to lower case
		for(i=0;i<4;i++) {
			receiverid[i]=tolower(epoch->receiver.name[i]);
		}
		receiverid[4]='\0';
	}

	// First index of XYZ
	indx = unkinfo->par2unk[DR_UNK];
	
	// Obtaing NEU coordinates
	XYZ2NEU(&solution->x[indx],positionNEU);
	
	// Compute:
	//		dNEU differences in relation to apriori position
	for (i=0;i<3;i++) {
		dNEU[i] = 0;
		for (j=0;j<3;j++) {
			dNEU[i] += (solution->x[indx+j] - epoch->receiver.aproxPosition[j])*epoch->receiver.orientation[i][j];
		}
	}
	
	if (options->receiverPositionSource >= rpCALCULATE) {
		for (i=0;i<3;i++) {
			dNEU[i] = 0;
		}
	}

	printf("SBASOUT     %4s %3s %17s %4d %9.2f %3d %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %14.4f %2d %2d", 
	/*2*/		receiverid,
	/*3*/		SBASnavmode2SBASnavstr(options->precisionapproach),
	/*4*/		t2doystr(&epoch->t),
	/*5*/
	/*6*/
	/*7*/		GPSweek,
	/*8*/		SoW,
	/*9*/		SBASdata[options->GEOindex].PRN,
	/*10*/		dNEU[0],	//North error (Meters)
	/*11*/		dNEU[1],	//East error (Meters)
	/*12*/		dNEU[2],	//Up error (Meters)
	/*13*/		sqrt(dNEU[0]*dNEU[0]+dNEU[1]*dNEU[1]),	//Horizontal error (Meters)
	/*14*/		solution->HPL,		//Horizontal protection level (Meters)
	/*15*/		solution->VPL,		//Vertical protection level (Meters)
	/*16*/		solution->x[unkinfo->par2unk[DT_UNK]],	//Receiver clock offset (Meters)
	/*17*/		epoch->numSatSolutionSBAS+epoch->numSatnotUsedSBAS,	//Number of satellites in view that can be used for SBAS
	/*18*/		epoch->numSatSolutionSBAS);	//Number of satellites used to compute the solution

	//Print list of PRN used to compute the solution in SBAS
	for (i=0;i<epoch->numSatSolutionSBAS;i++) {	
		printf(" +%c%02d",epoch->listConsSatSolutionSBAS[i],epoch->listPRNSatSolutionSBAS[i]);
	}

	//Print list of PRN not used to compute the solution but are suitable for SBAS
	for (i=0;i<epoch->numSatnotUsedSBAS;i++) {	
		printf(" -%c%02d",epoch->listConsSatnotUsedSBAS[i],epoch->listPRNSatnotUsedSBAS[i]);
	}
	printf("\n");
}

/*****************************************************************************
 * Name        : printBuffers
 * Description : Prints all the information saved in the buffers
 * Parameters  :
 * Name                                    |Da|Unit|Description
 * TEpoch  *epoch                           I  N/A  Reference epoch
 * TOptions  *options                       I  N/A  TOptions structure
 *****************************************************************************/
void printBuffers (TEpoch  *epoch, TOptions  *options) {

	int i,j;

	for(i=0;i<epoch->numSatellites;i++) {
		for(j=0;j<linesstored[i];j++) {
			if(options->printInBuffer==3 && options->printSBASUNUSED==1) {
				if(printbuffer[i][j][7]=='R') printbuffer[i][j][8]='*';
				else if(printbuffer[i][j][8]=='L') printbuffer[i][j][9]='*';
				else if(printbuffer[i][j][7]=='O') printbuffer[i][j][8]='*';
				else if(printbuffer[i][j][4]=='L') printbuffer[i][j][5]='*';
				else if(printbuffer[i][j][6]=='R') printbuffer[i][j][7]='*';
				else if(printbuffer[i][j][5]=='L') printbuffer[i][j][6]='*';
				printf("%s",printbuffer[i][j]);
			} else if (options->printInBuffer==2) {
				printf("%s",printbuffer[i][j]);
			}
			printbuffer[i][j][0]='\0';
		}
		linesstored[i]=0;
	}
}

/*****************************************************************************
 * Name        : printUserError
 * Description : Prints the user added error values
 * Parameters  :
 * Name                                    |Da|Unit|Description
 * TEpoch  *epoch                           I  N/A  Reference epoch
 * double  **SatMeasError                   I  N/A  User added errors vector
 * int numpos                               I  N/A  Number of positions in SatMeasError vector
 * TOptions  *options                       I  N/A  TOptions structure
 *****************************************************************************/
void printUserError (TEpoch *epoch, double **SatMeasError, int numpos, TOptions *options) {
	int             i;
		int             GPSweek;
		double          SoW;
		char            receiverid[5];

		//Calculate GPS Week and Seconds of Week
		ttime2gpswsnoroll(&epoch->t, &GPSweek, &SoW);

		//Convert receiver name to lower case
		for(i=0;i<4;i++) {
				receiverid[i]=tolower(epoch->receiver.name[i]);
		}
		receiverid[4]='\0';

	for(i=0;i<numpos;i++) {
		printf("USERADDEDERROR   %17s %4d %9.2f %3s %3d %3s %14.4f %14.4f %2d %14.4f %2d %14.4f %2d %14.4f %2d %14.4f %2d %14.4f\n",
		/*2*/			t2doystr(&epoch->t),
		/*3*/
		/*4*/
		/*5*/			GPSweek,
		/*6*/			SoW,
		/*7*/			gnsstype2gnssstr((int)SatMeasError[i][USRGNSSPOS]),	//GNSS system
		/*8*/			(int)SatMeasError[i][USRPRNPOS],			//PRN
		/*9*/			meastype2measstr((int)SatMeasError[i][USRMEASPOS]),	//Meas type
		/*10*/			SatMeasError[i][USRMEASRAWPOS],				//Raw measurement from RINEX
		/*11*/			SatMeasError[i][USRMEASTOTPOS],				//Raw measurement + total error value
		/*12*/			(int)SatMeasError[i][USRNUMTOTERRPOS],			//Total number of user defined error added
		/*13*/			SatMeasError[i][USRTOTERRPOS],				//Total user defined error value added
		/*14*/			(int)SatMeasError[i][USRNUMSTEPPOS],			//Number of Step function error added
		/*15*/			SatMeasError[i][USRSTEPERRPOS],				//Step function error value added
		/*16*/			(int)SatMeasError[i][USRNUMRAMPPOS],			//Number of Ramp function error added
		/*17*/			SatMeasError[i][USRRAMPERRPOS],				//Ramp function error value added
		/*18*/			(int)SatMeasError[i][USRNUMSINUPOS],			//Number of Sinusoidal function error added
		/*19*/			SatMeasError[i][USRSINUERRPOS],				//Sinusoidal function error value added
		/*20*/			(int)SatMeasError[i][USRNUMAWGNPOS],			//Number of AWGN function error added
		/*21*/			SatMeasError[i][USRAWGNERRPOS]);			//AWGN function error value added
	}
}

/*****************************************************************************
 * Name        : printStdESA
 * Description : Prints the Stanford-ESA plot values
 * Parameters  :
 * Name                                    |Da|Unit|Description
 * FILE *fd		         	    I  N/A  File descriptor where data is written
 * TStdESA *StdESA                          I  N/A  TStdESA struct                                        
 * TOptions  *options                       I  N/A  TOptions structure
 *****************************************************************************/
void printStdESA (FILE *fd, TStdESA *StdESA, TOptions *options) {

	int 		i,j;
	double		posx,posy;


	//Print resolution for each pixel in horizontal and vertical, and the maximum value for horizontal and vertical axis (in this order)
	fprintf(fd,"%5.3f %5.3f %4.1f %4.1f\n",StdESA->xstep,StdESA->ystep,StdESA->xmax,StdESA->ymax);

	//Print values for horizontal and vertical errors
	for(i=0;i<StdESA->numpixelver;i++) {
		for(j=0;j<StdESA->numpixelhor;j++) {
			posx=(double)(j)*StdESA->xstep;
			posy=(double)(i)*StdESA->ystep;
			//Only plot ocurrences if they are greater than zero
			if(StdESA->counthor[i][j]>0 || StdESA->countver[i][j]>0) {
				fprintf(fd,"%5.3f %5.3f %d %d\n",posx,posy,StdESA->counthor[i][j],StdESA->countver[i][j]);
			}
		}
	}
}
